/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Source file for NIC routines to access the Phantom hardware
 *
 * $Id: unm_nic_hw.c,v 1.105.2.3 2007/03/17 08:30:52 sanjeev Exp $
 *
 */
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/ethtool.h>
#include <linux/version.h>
#include <linux/pci.h>
#include "queue.h"
#include "unm_nic.h"
#include "unm_nic_hw.h"
#include "nic_cmn.h"
#include "unm_version.h"
#include "unm_brdcfg.h"

#ifdef __VMKERNEL_MODULE__
#define dump_stack(...)
#endif

/* static unsigned long unm_nic_pci_set_crbwindow(unm_adapter *adapter,
                                                    u64 off); */
static unsigned long unm_nic_pci_set_window (struct unm_adapter_s *adapter,
                                                    unsigned long long addr);


/*
 * Changes the CRB window to the specified window.
 */
void
unm_nic_pci_change_crbwindow(unm_adapter *adapter, uint32_t wndw)
{
        unm_pcix_crb_window_t        window;
        unsigned long                offset;
        uint32_t                     tmp;

        if (adapter->curr_window == wndw) {
                return;
        }

        /*
         * Move the CRB window.
         * We need to write to the "direct access" region of PCI
         * to avoid a race condition where the window register has
         * not been successfully written across CRB before the target
         * register address is received by PCI. The direct region bypasses
         * the CRB bus.
         */
        /*
         * NOTE: When B0 arrives, we can use direct access registers for
         * better performance
         */
	switch(adapter->ahw.pci_func) {
		case 0:
		        offset = PCI_OFFSET_SECOND_RANGE(adapter, 
				UNM_PCIX_PH_REG(PCIX_CRB_WINDOW));
			break;
		case 1:
		        offset = PCI_OFFSET_SECOND_RANGE(adapter, 
				UNM_PCIX_PH_REG(PCIX_CRB_WINDOW_F1));
			break;
		case 2:
		        offset = PCI_OFFSET_SECOND_RANGE(adapter, 
				UNM_PCIX_PH_REG(PCIX_CRB_WINDOW_F2));
			break;
		case 3:
		        offset = PCI_OFFSET_SECOND_RANGE(adapter, 
				UNM_PCIX_PH_REG(PCIX_CRB_WINDOW_F3));
			break;
		default:
			printk("Changing the window for PCI function %d\n",adapter->ahw.pci_func);
		        offset = PCI_OFFSET_SECOND_RANGE(adapter, 
				UNM_PCIX_PH_REG(PCIX_CRB_WINDOW));
			break;
	}

        *(unm_crbword_t *)&window = 0;
        window.addrbit = wndw;
        UNM_NIC_PCI_WRITE_32(*(unsigned int *)&window, (void*)(offset));

        //printk("set crbw : 0x%08x\n",off);
        /* MUST make sure window is set before we forge on... */
        while ((tmp = UNM_NIC_PCI_READ_32((void*)offset)) !=
            *(uint32_t *)&window) {
                printk(KERN_WARNING "%s: %s WARNING: CRB window value not "
                                    "registered properly: 0x%08x.\n",
                                    unm_nic_driver_name, __FUNCTION__, tmp);
        }

        adapter->curr_window = wndw;
        return;
}

uint32_t unm_nic_get_crbwindow(unm_adapter *adapter)
{
        return adapter->curr_window;
}

/*
 * Set the CRB window based on the offset.
 */
static inline unsigned long
unm_nic_pci_set_crbwindow(unm_adapter *adapter, u64 off)
{
        /*
         * See if we are currently pointing to the region we want to use next.
         */
        if ((off >= UNM_CRB_PCIX_HOST) && (off < UNM_CRB_DDR_NET)) {
                /*
                 * No need to change window. PCIX and PCIE regs are in both
                 * windows.
                 */
                return (off);
        }

        if ((off >= UNM_CRB_PCIX_HOST) && (off < UNM_CRB_PCIX_HOST2)) {
                /* We are in first CRB window */
                if (adapter->curr_window != 0) {
                        unm_nic_pci_change_crbwindow(adapter, 0);
                }
                return (off);
        }

        if ((off > UNM_CRB_PCIX_HOST2) && (off < UNM_CRB_MAX)) {
                /* We are in second CRB window */
                off = off - UNM_CRB_PCIX_HOST2 + UNM_CRB_PCIX_HOST;

                if (adapter->curr_window != 1) {
                        unm_nic_pci_change_crbwindow(adapter, 1);
                }
                return (off);
        }

        if ((off >= UNM_PCI_DIRECT_CRB) && (off < UNM_PCI_CAMQM_MAX)) {
                /*
                 * We are in the QM or direct access register region - do
                 * nothing
                 */
                return (off);
        }

        /* strange address given */
        dump_stack();
        printk(KERN_WARNING"%s: Warning: unm_nic_pci_set_crbwindow called with"
                " an unknown address(%llx)\n", unm_nic_driver_name, off);

        return (off);
}

int
unm_nic_hw_write_ioctl(unm_adapter *adapter, u64 off, void *data, int len)
{
	void		*addr;
	unsigned long	flags=0;
	u64		offset = off;
	uint8_t		*mem_ptr = NULL;
	unsigned long	mem_base;
	unsigned long	mem_page;

	if(ADDR_IN_WINDOW1(off)) {//Window 1
		addr = CRB_NORMALIZE(adapter, off);
		if(!addr) {
			mem_base = pci_resource_start(adapter->ahw.pdev, 0);
			offset = CRB_NORMAL(off);
			mem_page = offset & PAGE_MASK;
			if(mem_page != ((offset + len - 1) & PAGE_MASK))
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE * 2);
			else
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
			if(mem_ptr == 0UL) {
				return 1;
			}
			addr = mem_ptr;
			addr += offset & (PAGE_SIZE - 1);
		}
		read_lock(&adapter->adapter_lock);
	} else {//Window 0
		addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
		if(!addr) {
			mem_base = pci_resource_start(adapter->ahw.pdev, 0);
			mem_page = off & PAGE_MASK;
			if(mem_page != ((off + len - 1) & PAGE_MASK))
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE * 2);
			else
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
			if(mem_ptr == 0UL) {
				return 1;
			}
			addr = mem_ptr;
			addr += off & (PAGE_SIZE - 1);
		}
		write_lock_irqsave(&adapter->adapter_lock, flags);
		unm_nic_pci_change_crbwindow(adapter, 0);
	}

	DPRINTK(1, INFO, "writing to base %lx offset %llx addr %p"
			" data %llx len %d\n",
			pci_base(adapter, off), off, addr,
			*(unsigned long long *)data, len);
	switch (len) {
       	case 1:
		UNM_NIC_PCI_WRITE_8 (*(__uint8_t *)data, addr);
		break;
	case 2:
		UNM_NIC_PCI_WRITE_16 (*(__uint16_t *)data, addr);
		break;
	case 4:
		UNM_NIC_PCI_WRITE_32 (*(__uint32_t *)data, addr);
		break;
	case 8:
		UNM_NIC_PCI_WRITE_64 (*(__uint64_t *)data, addr);
		break;
	default:
#if !defined(NDEBUG)
		if ((len & 0x7) != 0)
			printk("%s: %s len(%d) not multiple of 8.\n",
				unm_nic_driver_name, __FUNCTION__, len);
#endif
		DPRINTK(1, INFO, "writing data %lx to offset %llx, num words=%d\n",
			*(unsigned long *)data, off, (len>>3));

		UNM_NIC_HW_BLOCK_WRITE_64(data, addr, (len>>3));
		break;
	}
	if(ADDR_IN_WINDOW1(off)) {//Window 1
		read_unlock(&adapter->adapter_lock);
	} else {//Window 0
		unm_nic_pci_change_crbwindow(adapter, 1);
       		write_unlock_irqrestore(&adapter->adapter_lock, flags);
	}

	if(mem_ptr)
		iounmap(mem_ptr);
	return 0;
}

/*
 * Note : 'len' argument should be either 1, 2, 4, or a multiple of 8.
 */
int
unm_nic_hw_write_wx(unm_adapter *adapter, u64 off, void *data, int len)
{//This is modified from _unm_nic_hw_write()._unm_nic_hw_write does not exist now.
        void *addr;
        unsigned long        flags = 0;

        if(ADDR_IN_WINDOW1(off))
        {//Window 1
                addr = CRB_NORMALIZE(adapter, off);
                read_lock(&adapter->adapter_lock);
        }
        else{//Window 0
                addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
                write_lock_irqsave(&adapter->adapter_lock, flags);
                unm_nic_pci_change_crbwindow(adapter, 0);
        }


        DPRINTK(1, INFO, "writing to base %lx offset %llx addr %p"
                            " data %llx len %d\n",
                            pci_base(adapter, off), off, addr,
                            *(unsigned long long *)data, len);
	if(!addr) {
		if(ADDR_IN_WINDOW1(off)) {//Window 1
			read_unlock(&adapter->adapter_lock);
		} else {//Window 0
			unm_nic_pci_change_crbwindow(adapter, 1);
			write_unlock_irqrestore(&adapter->adapter_lock, flags);
		}
		return 1;
	}
        switch (len) {
        case 1:
                UNM_NIC_PCI_WRITE_8 (*(__uint8_t *)data, addr);
                break;
        case 2:
                UNM_NIC_PCI_WRITE_16 (*(__uint16_t *)data, addr);
                break;
        case 4:
                UNM_NIC_PCI_WRITE_32 (*(__uint32_t *)data, addr);
                break;
        case 8:
                UNM_NIC_PCI_WRITE_64 (*(__uint64_t *)data, addr);
                break;
        default:
#if !defined(NDEBUG)
                if ((len & 0x7) != 0)
                        printk("%s: %s  len(%d) not multiple of 8.\n",
                                unm_nic_driver_name, __FUNCTION__, len);
#endif
                DPRINTK(1, INFO, "writing data %lx to offset %llx, num words=%d\n",
                            *(unsigned long *)data, off, (len>>3));

                UNM_NIC_HW_BLOCK_WRITE_64(data, addr, (len>>3));
                break;
        }
        if(ADDR_IN_WINDOW1(off))
        {//Window 1
                read_unlock(&adapter->adapter_lock);
        }
        else{//Window 0
                unm_nic_pci_change_crbwindow(adapter, 1);
                write_unlock_irqrestore(&adapter->adapter_lock, flags);
        }

        return 0;
}

 int
unm_nic_hw_read_ioctl (unm_adapter *adapter, u64 off, void *data, int len)
{
	void		*addr;
	unsigned long	flags=0;
	u64		offset;
	uint8_t		*mem_ptr = NULL;
	unsigned long	mem_base;
	unsigned long	mem_page;

	if(ADDR_IN_WINDOW1(off)) {//Window 1
		addr = CRB_NORMALIZE(adapter, off);
		if(!addr) {
			mem_base = pci_resource_start(adapter->ahw.pdev, 0);
			offset = CRB_NORMAL(off);
			mem_page = offset & PAGE_MASK;
			if(mem_page != ((offset + len - 1) & PAGE_MASK))
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE * 2);
			else
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
			if(mem_ptr == 0UL) {
				*(__uint8_t  *)data = 0;
				return 1;
			}
			addr = mem_ptr;
			addr += offset & (PAGE_SIZE - 1);
		}
		read_lock(&adapter->adapter_lock);
	} else {//Window 0
		addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
		if(!addr) {
			mem_base = pci_resource_start(adapter->ahw.pdev, 0);
			mem_page = off & PAGE_MASK;
			if(mem_page != ((off + len - 1) & PAGE_MASK))
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE * 2);
			else
				mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
			if(mem_ptr == 0UL) {
				return 1;
			}
			addr = mem_ptr;
			addr += off & (PAGE_SIZE - 1);
		}
		write_lock_irqsave(&adapter->adapter_lock, flags);
		unm_nic_pci_change_crbwindow(adapter, 0);
	}

	DPRINTK(1, INFO, "reading from base %lx offset %llx addr %p\n",
			pci_base(adapter, off), off, addr);

	switch (len) {
	case 1:
		*(__uint8_t  *)data = UNM_NIC_PCI_READ_8(addr);
		break;
	case 2:
		*(__uint16_t *)data = UNM_NIC_PCI_READ_16(addr);
		break;
	case 4:
		*(__uint32_t *)data = UNM_NIC_PCI_READ_32(addr);
		break;
	case 8:
		*(__uint64_t *)data = UNM_NIC_PCI_READ_64(addr);
		break;
	default:
#if !defined(NDEBUG)
		if ((len & 0x7) != 0)
			printk("%s: %s len(%d) not multiple of 8.\n",
				unm_nic_driver_name, __FUNCTION__, len);
#endif
		UNM_NIC_HW_BLOCK_READ_64(data, addr, (len>>3));
		break;
	}
	DPRINTK(1, INFO, "read %lx\n", *(unsigned long *)data);

	if(ADDR_IN_WINDOW1(off)) {//Window 1
		read_unlock(&adapter->adapter_lock);
	} else {//Window 0
		unm_nic_pci_change_crbwindow(adapter, 1);
		write_unlock_irqrestore(&adapter->adapter_lock, flags);
	}

	if(mem_ptr)
		iounmap(mem_ptr);
	return 0;
}

int
unm_nic_hw_read_wx (unm_adapter *adapter, u64 off, void *data, int len)
{
        void *addr;
        unsigned long        flags = 0;

        if(ADDR_IN_WINDOW1(off))
        {//Window 1
                addr = CRB_NORMALIZE(adapter, off);
		read_lock(&adapter->adapter_lock);
        }
        else{//Window 0
                addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
		write_lock_irqsave(&adapter->adapter_lock, flags);
                unm_nic_pci_change_crbwindow(adapter, 0);
        }

        DPRINTK(1, INFO, "reading from base %lx offset %llx addr %p\n",
                            pci_base(adapter, off), off, addr);

	if(!addr) {
		if(ADDR_IN_WINDOW1(off)) {//Window 1
			read_unlock(&adapter->adapter_lock);
		} else {//Window 0
			unm_nic_pci_change_crbwindow(adapter, 1);
			write_unlock_irqrestore(&adapter->adapter_lock, flags);
		}
		return 1;
	}

        switch (len) {
        case 1:
                *(__uint8_t  *)data = UNM_NIC_PCI_READ_8(addr);
                break;
        case 2:
                *(__uint16_t *)data = UNM_NIC_PCI_READ_16(addr);
                break;
        case 4:
                *(__uint32_t *)data = UNM_NIC_PCI_READ_32(addr);
                break;
        case 8:
                *(__uint64_t *)data = UNM_NIC_PCI_READ_64(addr);
                break;
        default:
#if !defined(NDEBUG)
                if ((len & 0x7) != 0)
                        printk("%s: %s len(%d) not multiple of 8.\n",
                                unm_nic_driver_name, __FUNCTION__, len);
#endif
                UNM_NIC_HW_BLOCK_READ_64(data, addr, (len>>3));
                break;
        }

        DPRINTK(1, INFO, "read %lx\n", *(unsigned long *)data);

        if(ADDR_IN_WINDOW1(off))
        {//Window 1
		read_unlock(&adapter->adapter_lock);
        }
        else{//Window 0
                unm_nic_pci_change_crbwindow(adapter, 1);
		write_unlock_irqrestore(&adapter->adapter_lock, flags);
        }

        return 0;
}

/*  PCI Windowing for DDR regions.  */

#define ADDR_IN_RANGE(addr, low, high)            \
        (((addr) <= (high)) && ((addr) >= (low)))

int unm_pci_set_window_warning_count = 0;

static unsigned long
unm_nic_pci_set_window (struct unm_adapter_s *adapter, unsigned long long addr)
{
        static int ddr_mn_window=-1;
        static int qdr_sn_window=-1;
        int window;

        /* printk("addr: %lx\n", addr); */
        if (ADDR_IN_RANGE(addr, UNM_ADDR_DDR_NET, UNM_ADDR_DDR_NET_MAX)) {
                /* DDR network side */
                /* printk("MN DDR\n"); */
                addr -= UNM_ADDR_DDR_NET;
                window = (addr >> 25 )  & 0x3ff;
                if (ddr_mn_window != window) {
                        /* printk("Change the MN window!!! \n"); */
                        ddr_mn_window = window;
                        UNM_NIC_PCI_WRITE_32(window, (void *)(uptr_t)(PCI_OFFSET_SECOND_RANGE(adapter,
                                            UNM_PCIX_PH_REG(PCIX_MN_WINDOW))));
                        /* MUST make sure window is set before we forge on... */
                        UNM_NIC_PCI_READ_32((void *)(uptr_t)(PCI_OFFSET_SECOND_RANGE(adapter,
                                            UNM_PCIX_PH_REG(PCIX_MN_WINDOW))));
                }
                addr -= (window * 0x2000000);
                addr+= UNM_PCI_DDR_NET;
        } else if (ADDR_IN_RANGE(addr, UNM_ADDR_OCM0, UNM_ADDR_OCM0_MAX)) {
                addr -= UNM_ADDR_OCM0;
                addr += UNM_PCI_OCM0;
        } else if (ADDR_IN_RANGE(addr, UNM_ADDR_OCM1, UNM_ADDR_OCM1_MAX)) {
                addr -= UNM_ADDR_OCM1;
                addr += UNM_PCI_OCM1;
        } else if (ADDR_IN_RANGE(addr, UNM_ADDR_QDR_NET, UNM_ADDR_QDR_NET_MAX)) {
                /* printk("SN QDR\n"); */
                /* QDR network side */
                addr -= UNM_ADDR_QDR_NET;
                window = (addr >> 22 )  & 0x3f;
                if (qdr_sn_window != window) {
                        /* printk("Change the SN window!!! \n"); */
                        qdr_sn_window = window;
                        UNM_NIC_PCI_WRITE_32((window << 22), 
                                            (void *)(uptr_t)(PCI_OFFSET_SECOND_RANGE(adapter,
                                            UNM_PCIX_PH_REG(PCIX_SN_WINDOW))));
                        /* MUST make sure window is set before we forge on... */
                        UNM_NIC_PCI_READ_32((void *)(uptr_t)(PCI_OFFSET_SECOND_RANGE(adapter,
                                            UNM_PCIX_PH_REG(PCIX_SN_WINDOW))));
                }
                addr -= (window * 0x400000);
                addr+= UNM_PCI_QDR_NET;
        } else {
                /*
                 * peg gdb frequently accesses memory that doesn't exist,
                 * this limits the chit chat so debugging isn't slowed down.
                 */
                if((unm_pci_set_window_warning_count++ < 8)
                    || (unm_pci_set_window_warning_count%64 == 0)) {
                        printk("%s: Warning:unm_nic_pci_set_window() "
                                "Unknown address range!\n",unm_nic_driver_name);
                }
		addr = -1UL;
        }
        /* printk("New address: 0x%08lx\n",addr); */
        return addr;
}

int
unm_nic_pci_mem_write_ioctl(struct unm_adapter_s *adapter, u64 off,
                       void *data, int size)
{
       unsigned long   flags;
       void            *addr;
       int             ret = 0;
       uint8_t         *mem_ptr = NULL;
       unsigned long   mem_base;
       unsigned long   mem_page;

       if (data == NULL || off > (128*1024*1024)) {
               printk(KERN_ERR "data: %p off:%llx\n", data, off);
               return 1;
       }
       write_lock_irqsave(&adapter->adapter_lock, flags);
       off = unm_nic_pci_set_window(adapter, off);
       /* Corner case : Malicious user tried to break the driver by reading
          last few bytes in ranges and tries to read further addresses.
        */
       if(!pci_base(adapter, off + size - 1) && pci_base(adapter, off)) {
               write_unlock_irqrestore(&adapter->adapter_lock, flags);
               printk(KERN_ERR "Invalid access to memory address range 0x%llx - 0x%llx\n", off,
 off + size);
               return 1;
       }
       addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
       DPRINTK(1, INFO, "writing data %llx to offset %llx\n",
                       *(unsigned long long *)data, off);
       if(!addr) {
               write_unlock_irqrestore(&adapter->adapter_lock, flags);
               mem_base = pci_resource_start(adapter->ahw.pdev, 0);
               mem_page = off & PAGE_MASK;
               /* Map two pages whenever user tries to access addresses in two
                  consecutive pages.
                */
               if(mem_page != ((off + size - 1) & PAGE_MASK))
                       mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE*2);
               else
                       mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
               if(mem_ptr == 0UL) {
                       return 1;
               }
               addr = mem_ptr;
               addr += off & (PAGE_SIZE - 1);
               write_lock_irqsave(&adapter->adapter_lock, flags);
       }

       switch (size) {
       case 1:
               UNM_NIC_PCI_WRITE_8( *(__uint8_t  *)data, addr);
               break;
       case 2:
               UNM_NIC_PCI_WRITE_16(*(__uint16_t *)data, addr);
               break;
       case 4:
               UNM_NIC_PCI_WRITE_32(*(__uint32_t *)data, addr);
               break;
       case 8:
               UNM_NIC_PCI_WRITE_64(*(__uint64_t *)data, addr);
               break;
       default:
               ret = 1;
               break;
       }
       write_unlock_irqrestore(&adapter->adapter_lock, flags);
       if(mem_ptr)
               iounmap(mem_ptr);
       DPRINTK(1, INFO, "wrote %llx\n", *(unsigned long long*)data);

       return ret;
}

int
unm_nic_pci_mem_read_ioctl(struct unm_adapter_s *adapter,
                       u64 off, void *data, int size)
{
       unsigned long   flags;
       void            *addr;
       int             ret = 0;
       uint8_t         *mem_ptr = NULL;
       unsigned long   mem_base;
       unsigned long   mem_page;

       if (data == NULL || off > UNM_ADDR_QDR_NET_MAX) {
               printk(KERN_ERR "data: %p off:%llx\n", data, off);
               return 1;
       }
       write_lock_irqsave(&adapter->adapter_lock, flags);
       off = unm_nic_pci_set_window(adapter, off);
       /* Corner case : Malicious user tried to break the driver by reading
          last few bytes in ranges and tries to read further addresses.
        */
       if(!pci_base(adapter, off + size - 1) && pci_base(adapter, off)) {
               write_unlock_irqrestore(&adapter->adapter_lock, flags);
               printk(KERN_ERR "Invalid access to memory address range 0x%llx - 0x%llx\n", off, off + size);
               return 1;
       }
       addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
       if(!addr) {
               write_unlock_irqrestore(&adapter->adapter_lock, flags);
               mem_base = pci_resource_start(adapter->ahw.pdev, 0);
               mem_page = off & PAGE_MASK;
               /* Map two pages whenever user tries to access addresses in two
                  consecutive pages.
                */
               if(mem_page != ((off + size - 1) & PAGE_MASK))
                       mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE * 2);
               else
                       mem_ptr = ioremap(mem_base + mem_page, PAGE_SIZE);
               if(mem_ptr == 0UL) {
                       *(__uint8_t  *)data = 0;
                       return 1;
               }
               addr = mem_ptr;
               addr += off & (PAGE_SIZE - 1);
               write_lock_irqsave(&adapter->adapter_lock, flags);
       }

       switch (size) {
       case 1:
               *(__uint8_t  *)data = UNM_NIC_PCI_READ_8(addr);
               break;
       case 2:
               *(__uint16_t *)data = UNM_NIC_PCI_READ_16(addr);
               break;
       case 4:
               *(__uint32_t *)data = UNM_NIC_PCI_READ_32(addr);
               break;
       case 8:
               *(__uint64_t *)data = UNM_NIC_PCI_READ_64(addr);
               break;
       default:
               ret = 1;
               break;
      }
       write_unlock_irqrestore(&adapter->adapter_lock, flags);
       if(mem_ptr)
               iounmap(mem_ptr);
       DPRINTK(1, INFO, "read %llx\n", *(unsigned long long*)data);

      return ret;
}
#if 0

/* read num_words from hardware (num_words in 64 bits    */
void
unm_nic_mem_block_read(struct unm_adapter_s *adapter, u64 off,
                        void *data, int num_words)
{
        void *addr;
        unsigned long flags;

        write_lock_irqsave(&adapter->adapter_lock, flags);
	/*
	 * TODO: Check address range validity and alignment.
	 */
        off  = unm_nic_pci_set_window(adapter->ahw.pci_base, off);
        addr = (void *)(uptr_t)(adapter->ahw.pci_base + off);
        DPRINTK(1, INFO, "reading from base %lx offset %lx addr %p\n",
                        adapter->ahw.pci_base, (unsigned long)off, addr);
        UNM_NIC_HW_BLOCK_READ_64(data, addr, num_words);
        write_unlock_irqrestore(&adapter->adapter_lock, flags);

        return;
}

/* Write num_words (in 64 bits) in the phantom memory    */
void
unm_nic_mem_block_write(struct unm_adapter_s *adapter, u64 off,
                            void *data, int num_words)
{
        void *addr;
        unsigned long    flags;

        write_lock_irqsave(&adapter->adapter_lock, flags);
	/*
	 * TODO: Check address range validity and alignment.
	 */
        off  = unm_nic_pci_set_window(adapter->ahw.pci_base, off);
        addr = (void *)(uptr_t)(off + adapter->ahw.pci_base);
        DPRINTK(1, INFO, "writing data %lx to offset %lx\n",
                        *(unsigned long *)data, (unsigned long)off);
        UNM_NIC_HW_BLOCK_WRITE_64(data, addr, num_words);
        write_unlock_irqrestore(&adapter->adapter_lock, flags);

        return;
}

#endif

int
unm_nic_pci_mem_write(struct unm_adapter_s *adapter, u64 off,
                        void *data, int size)
{
        unsigned long   flags;
        void           *addr;
        int             ret = 0;

#if 0
	/*
	 * This check can not be currently executed, since firmware load
	 * breaks this check whereby 8 byte writes are being attempted on
	 * "aligned-by-4" addresses on x86. Reason this works is our version
	 * of "readq"/"writeq" breaks up the access into 2 consecutive 4
	 * byte writes; on other architectures, readq/writeq might require
	 * "aligned-by-8" addresses and we will run into trouble.
	 *
	 * Check alignment for expected sizes of 1, 2, 4, 8. Other size
	 * values will not trigger access.
	 */
	if ((off & (size - 1)) != 0)
		return(-1);
#endif

        write_lock_irqsave(&adapter->adapter_lock, flags);

	/*
	 * If attempting to access unknown address range, do not access.
	 */
        if ((off = unm_nic_pci_set_window(adapter, off)) == -1UL){
                size = -1;
        }


        addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
        if(!addr) {
                write_unlock_irqrestore(&adapter->adapter_lock, flags);
                printk(KERN_ERR"%s Tried to access memory outside "
                        "the mapped pci address range: offset is 0x%llx\n", unm_nic_driver_name, off);
                return 1;
        }

        DPRINTK(1, INFO, "writing data %llx to offset %llx\n",
                            *(unsigned long long *)data, off);
        switch (size) {
        case 1:
                UNM_NIC_PCI_WRITE_8( *(__uint8_t  *)data, addr);
                break;
        case 2:
                UNM_NIC_PCI_WRITE_16(*(__uint16_t *)data, addr);
                break;
        case 4:
                UNM_NIC_PCI_WRITE_32(*(__uint32_t *)data, addr);
                break;
        case 8:
                UNM_NIC_PCI_WRITE_64(*(__uint64_t *)data, addr);
                break;
        default:
                ret = -1;
                break;
        }
        write_unlock_irqrestore(&adapter->adapter_lock, flags);
        DPRINTK(1, INFO, "wrote %llx\n", *(unsigned long long*)data);

        return ret;
}

int
unm_nic_pci_mem_read(struct unm_adapter_s *adapter,
                        u64 off, void *data, int size)
{
        unsigned long   flags;
        void           *addr;
        int             ret = 0;

#if 0
	/*
	 * This check can not be currently executed, since phanmon findq
	 * command breaks this check whereby 8 byte reads are being attempted
	 * on "aligned-by-4" addresses on x86. Reason this works is our version
	 * of "readq"/"writeq" breaks up the access into 2 consecutive 4
	 * byte writes; on other architectures, readq/writeq might require
	 * "aligned-by-8" addresses and we will run into trouble.
	 *
	 * Check alignment for expected sizes of 1, 2, 4, 8. Other size
	 * values will not trigger access.
	 */
	if ((off & (size - 1)) != 0)
		return(-1);
#endif

        write_lock_irqsave(&adapter->adapter_lock, flags);

	/*
	 * If attempting to access unknown address range, do not access.
	 */
        if ((off = unm_nic_pci_set_window(adapter, off)) == -1UL){
                size = -1;
        }

        addr = (void *)(uptr_t)(pci_base_offset(adapter, off));
        if(!addr) {
                write_unlock_irqrestore(&adapter->adapter_lock, flags);
                printk(KERN_ERR"%s Tried to access memory outside "
                        "the mapped pci address range: offset is 0x%llx\n", unm_nic_driver_name, off);

                return 1;
        }

        switch (size) {
        case 1: 
                *(__uint8_t  *)data = UNM_NIC_PCI_READ_8(addr);
                break;
        case 2:
                *(__uint16_t *)data = UNM_NIC_PCI_READ_16(addr);
                break;
        case 4:
                *(__uint32_t *)data = UNM_NIC_PCI_READ_32(addr);
                break;
        case 8:
                *(__uint64_t *)data = UNM_NIC_PCI_READ_64(addr);
                break;
        default:
                ret = -1;
                break;
        }
        write_unlock_irqrestore(&adapter->adapter_lock, flags);
        DPRINTK(1, INFO, "read %llx\n", *(unsigned long long*)data);

        return ret;
}

/*static void
 *unm_load_cam_ram(struct unm_adapter_s *adapter)
 *{
 *       DPRINTK(1, INFO, "writing to  LOAD CAM RAM.....START \n");
 *       unm_nic_reg_write(adapter, UNM_CRB_CAM + 0x2000, 0);
 *       // may be we need to have some value for bootpeg - IND
 *
 *       // Hw bug - needs an extra write after reset
 *       unm_nic_reg_write(adapter, UNM_CRB_CAM + 0x2000, 0);
 *       // one more for good measure
 *       unm_nic_reg_write(adapter, UNM_CRB_CAM + 0x2000, 0);
 *
 *       return;
 *}
*/
int
unm_nic_erase_pxe(struct unm_adapter_s *adapter)
{
        if (rom_fast_write(adapter, PXE_START, 0) == -1) { 
                printk("erase pxe failed\n");
                return -1;
        }
        return 0;
}

int
unm_nic_get_board_info(struct unm_adapter_s *adapter)
{
        int                rv=0;
        unm_board_info_t  *boardinfo;
        int                i;
        int                addr = BRDCFG_START;
        uint32_t          *ptr32;

        boardinfo = &adapter->ahw.boardcfg;
        ptr32 = (uint32_t *)boardinfo;

        for (i=0;i<sizeof(unm_board_info_t)/sizeof(uint32_t);i++) {
                if (rom_fast_read(adapter, addr, ptr32) == -1) {
printk("Rom read failed after %d \n",i);
                        return -1;
                }
                ptr32++;
                //printk("ROM: %x\n",*ptr32);
                addr+= sizeof(uint32_t);
        }
        if (boardinfo->magic != UNM_BDINFO_MAGIC) {
                printk("%s: ERROR reading %s board config."
                        " Read %x, expected %x\n", unm_nic_driver_name,
                        unm_nic_driver_name, boardinfo->magic, UNM_BDINFO_MAGIC);
                rv=-1;
        }
        if (boardinfo->header_version != UNM_BDINFO_VERSION) {
                printk("%s: Unknown board config version."
                        " Read %x, expected %x\n", unm_nic_driver_name,
                        boardinfo->header_version, UNM_BDINFO_VERSION);
                rv=-1;
        }

        DPRINTK(1, INFO, "Discovered board type:0x%x  ",boardinfo->board_type);
        switch((unm_brdtype_t)boardinfo->board_type){
        case UNM_BRDTYPE_P2_SB35_4G:
                adapter->ahw.board_type = UNM_NIC_GBE;
                break;
        case UNM_BRDTYPE_P2_SB31_10G:
        case UNM_BRDTYPE_P2_SB31_10G_IMEZ:
        case UNM_BRDTYPE_P2_SB31_10G_HMEZ:
        case UNM_BRDTYPE_P2_SB31_10G_CX4:
                adapter->ahw.board_type = UNM_NIC_XGBE;
                break;
        case UNM_BRDTYPE_P1_BD:
        case UNM_BRDTYPE_P1_SB:
        case UNM_BRDTYPE_P1_SMAX:
        case UNM_BRDTYPE_P1_SOCK:
                adapter->ahw.board_type = UNM_NIC_GBE;
                break;
        default:
                printk("%s: Unknown(%x)\n",unm_nic_driver_name, 
			boardinfo->board_type);
                break;
        }

        return rv;
}

int
unm_nic_get_board_num(unm_adapter *adapter)
{
        return adapter->ahw.boardcfg.board_num;
}

/* NIU access sections */

/* Set the MAC address for a port */
int
unm_nic_macaddr_set(struct unm_adapter_s *adapter, __uint8_t *addr)
{
        int                     phy = physical_port[adapter->portnum];
        int ret = 0;

        if ((phy < 0) || (phy > 3))
                return -1;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                ret = unm_niu_macaddr_set(adapter, phy, addr);
                break;

        case UNM_NIC_XGBE:
                ret = unm_niu_xg_macaddr_set(adapter, phy, addr);
                break;

        default:
                dump_stack();
                printk(KERN_ERR"%s: Unknown board type encountered while "
			"setting the MAC address.\n", unm_nic_driver_name);
                return -1;
        }
        return ret;
}

#if 0
/* returns FALSE(0) if not acceptable speed else (TRUE)1 */
static int validate_clk_speed(struct unm_adapter_s *adapter)
{
        unm_crbword_t time1=0, time2=0;
        unsigned long jiffies1, jiffies2, jiffies_elapsed, time;

        /* Window 1 read */
        read_lock(&adapter->adapter_lock);
        time1 = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,UNM_TIMER_TIMESTAMP));
        jiffies1 = jiffies;
        mdelay(1000);
        time2 = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,UNM_TIMER_TIMESTAMP));
        jiffies2 = jiffies;
        read_unlock(&adapter->adapter_lock);

        if (time2 > time1)
            time = time2 - time1;
        else
            time = ~time1 + time2 + 1;
        /* Get actual speed */
        if(jiffies2 > jiffies1)
            jiffies_elapsed = jiffies2 - jiffies1;
        else
            jiffies_elapsed = ~jiffies1 + jiffies2 + 1;

        /* For P2 expected ticks per second is 2048 */
        time = (time * HZ)/jiffies_elapsed;
        return CHECK_TICKS_PER_SEC(time);
}

/* Update CRB register for reflecting correct core clock speed */
int update_core_clock(struct unm_port *port)
{
        struct unm_adapter_s *adapter = port->adapter;
        int                   phy = port->portnum;
        unsigned long jiffies1, jiffies2, jiffies_elapsed, time;
        unm_crbword_t time1=0, time2=0;
        unm_timer_tickctl_t tickctl;

        if ((phy < 0) || (phy > 3))
                return -1;

        if (validate_clk_speed(adapter)) {
            /* No need to worry */
            return 0;
        }

        /* Window 1 read/write */
        read_lock(&adapter->adapter_lock);
        time1 = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,UNM_TIMER_GLOBAL_TIMESTAMP_LO));
        jiffies1 = jiffies;
        mdelay(1000);
        time2 = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,UNM_TIMER_GLOBAL_TIMESTAMP_LO));
        jiffies2 = jiffies;
        read_unlock(&adapter->adapter_lock);
        if (time2 > time1)
            time = time2 - time1;
        else
            time = ~time1 + time2 + 1;

        /* Get actual speed */
        if(jiffies2 > jiffies1)
            jiffies_elapsed = jiffies2 - jiffies1;
        else
            jiffies_elapsed = ~jiffies1 + jiffies2 + 1;

        /* add 500K, so that sounding of in MHz is correct */
        time = time + 500000UL;
        time /= 1000000UL;
        time = (time * HZ)/jiffies_elapsed;
        if(!ACCEPTABLE_CORE_CLK_RANGE(time)) {
            printk(" Core clock   : %ld MHz, not Acceptable\n", time);
            return -1;
        }
        DPRINTK(1, INFO," Core clock   : %ld MHz\n", time);
        *(unm_crbword_t *)&tickctl = 0;
        tickctl.enable = 1;
        tickctl.count = ((1000000ULL/(4*2048))* time);
        read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(*(unm_crbword_t *)&tickctl, CRB_NORMALIZE(adapter,UNM_TIMER_GT_TICKCTL));
        read_unlock(&adapter->adapter_lock);

        return 0;
}
#endif


int
unm_nic_set_mtu(struct unm_adapter_s *adapter, int new_mtu)
{
        int                   ret = 0;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                unm_nic_write_w0 (
                        adapter,UNM_NIU_GB_MAX_FRAME_SIZE(physical_port[adapter->portnum]),
                        new_mtu);

                break;

        case UNM_NIC_XGBE:
                new_mtu += 100; /* so that MAC accepts frames > MTU */
                if (physical_port[adapter->portnum] == 0) {
		    unm_nic_write_w0 (adapter, UNM_NIU_XGE_MAX_FRAME_SIZE, new_mtu);
                } else {
		    unm_nic_write_w0 (adapter, UNM_NIU_XG1_MAX_FRAME_SIZE, new_mtu);
                }
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n",unm_nic_driver_name);
        }

        return ret;
}

int
unm_nic_set_promisc_mode(struct unm_adapter_s *adapter)
{
        int                   phy = physical_port[adapter->portnum];
        int                   ret;

        if ((phy < 0) || (phy > 3))
                return -1;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                ret = unm_niu_set_promiscuous_mode(adapter, phy,
                                                    UNM_NIU_PROMISCOUS_MODE);
                break;

        case UNM_NIC_XGBE:
                ret = unm_niu_xg_set_promiscuous_mode(adapter, phy,
                                                    UNM_NIU_PROMISCOUS_MODE);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n",unm_nic_driver_name);
		ret = -1;
		break;
        }

        return ret;
}

int
unm_nic_unset_promisc_mode(struct unm_adapter_s *adapter)
{
        int                   phy = physical_port[adapter->portnum];
        int                   ret;

        if ((phy < 0) || (phy > 3))
                return -1;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                ret = unm_niu_set_promiscuous_mode(adapter, phy,
                                                UNM_NIU_NON_PROMISCOUS_MODE);
                break;

        case UNM_NIC_XGBE:
                ret = unm_niu_xg_set_promiscuous_mode(adapter, phy,
                                                UNM_NIU_NON_PROMISCOUS_MODE);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n", unm_nic_driver_name);
		ret = -1;
		break;
        }

        return ret;
}

long
unm_nic_phy_read (unm_adapter *adapter, long phy, long reg,
                    __uint32_t *readval)
{
        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                return unm_niu_gbe_phy_read (adapter, phy, reg, readval);

        case UNM_NIC_XGBE:
                printk(KERN_ERR "%s: Function %s is not implemented for XG\n",
                        unm_nic_driver_name, __FUNCTION__);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n", unm_nic_driver_name);
                break;
        }

        return -1;
}

long
unm_nic_phy_write (unm_adapter *adapter, long phy, long reg, __uint32_t val)
{
	/*
	 * Note that phy is already the phyical port number. The mapping is
	 * done by the caller.
	 */
        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                return unm_niu_gbe_phy_write (adapter, phy, reg, val);

        case UNM_NIC_XGBE:
                printk(KERN_ERR"%s: Function %s is not implemented for XG\n",
                           unm_nic_driver_name, __FUNCTION__);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n",unm_nic_driver_name);
                break;
        }

        return -1;
}

long
unm_nic_init_port (struct unm_adapter_s *adapter)
{
        long                  portnum = physical_port[adapter->portnum];
        long                  ret = 0;

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                break;

        case UNM_NIC_XGBE:
                unm_crb_writelit_adapter(adapter,
                         UNM_NIU_XGE_CONFIG_1+(0x10000*portnum), 0x1447);
                unm_crb_writelit_adapter(adapter,
                         UNM_NIU_XGE_CONFIG_0+(0x10000*portnum), 0x5);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n", unm_nic_driver_name);
        }

        return ret;
}

void
unm_nic_stop_port(struct unm_adapter_s *adapter)
{
        long                  portnum = physical_port[adapter->portnum];

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                unm_niu_disable_gbe_port(adapter, portnum);
                break;

        case UNM_NIC_XGBE:
                unm_niu_disable_xg_port(adapter, portnum);
                break;

        default:
                printk(KERN_ERR"%s: Unknown board type\n", unm_nic_driver_name);
        }

        return;
}

#if 0
void
unm_nic_stop_all_ports(struct unm_adapter_s *adapter)
{
        int              i;
        struct unm_port *port;

        for (i=0; i < adapter->ahw.max_ports; i++) {
                port = adapter->port[i];
                unm_nic_stop_port(port);
        }

         return;
}
#endif

/* Functions required by UNM Hal routines */
unsigned long
unm_xport_lock(void)
{
        /* No lock necessary */
        return 0;
}

void
unm_xport_unlock(unsigned long unused)
{
        /* No lock/unlock necessary */
        return;
}

int
unm_crb_writelit_adapter (struct unm_adapter_s *adapter,
			  unsigned long off, int data)
{
        unsigned long flags;
        void *addr;

        if(ADDR_IN_WINDOW1(off))
        {
                read_lock(&adapter->adapter_lock);
                UNM_NIC_PCI_WRITE_32(data, CRB_NORMALIZE(adapter, off));
                read_unlock(&adapter->adapter_lock);
        }
        else{
                //unm_nic_write_w0 (adapter, off, data);
                write_lock_irqsave(&adapter->adapter_lock, flags);
                unm_nic_pci_change_crbwindow(adapter, 0);
                addr = (void *)(pci_base_offset(adapter, off));
                UNM_NIC_PCI_WRITE_32(data, addr);
                unm_nic_pci_change_crbwindow(adapter, 1);
                write_unlock_irqrestore(&adapter->adapter_lock, flags);
        }

        return 0;
}

int
unm_crb_write_adapter (unsigned long off, void *data, struct unm_adapter_s *adapter)
{
        return  unm_nic_hw_write_wx (adapter, off, data, 4);
}

int
unm_crb_read_adapter (unsigned long off, void *data, struct unm_adapter_s *adapter)
{
        return unm_nic_hw_read_wx (adapter, off, data, 4);
}

int
unm_crb_read_val_adapter (unsigned long off, struct unm_adapter_s *adapter)
{
        int        data;

        unm_nic_hw_read_wx (adapter, off, &data, 4);
        return data;
}


void DELAY(A)
{
        unsigned long remainder;

        remainder = A/50000;
        do {
                if (remainder > 1000) {
                        udelay(1000);
                        remainder -= 1000;
                } else {
                        udelay(remainder + 1);
                        remainder = 0;
                }
        } while (remainder > 0);
}

void unm_nic_set_link_parameters(struct unm_adapter_s *adapter)
{
        unm_niu_phy_status_t status;
        uint16_t autoneg;
        unm_niu_control_t mode;

	unm_nic_read_w0(adapter, UNM_NIU_MODE, (uint32_t *)&mode);
        if(mode.enable_ge) { // Gb 10/100/1000 Mbps mode
                if (unm_nic_phy_read(adapter,physical_port[adapter->portnum],
                        UNM_NIU_GB_MII_MGMT_ADDR_PHY_STATUS,
                        (unm_crbword_t *)&status) == 0) {
                        if(status.link) {
                                adapter->state = 1;
                                switch(status.speed) {
                                case 0: adapter->link_speed = SPEED_10;
                                        break;
                                case 1: adapter->link_speed = SPEED_100;
                                        break;
                                case 2: adapter->link_speed = SPEED_1000;
                                        break;
                                default:
                                        adapter->link_speed = -1; // unknown speed
                                        break;
                                }
                                switch(status.duplex) {
                                case 0: adapter->link_duplex = DUPLEX_HALF;
                                        break;
                                case 1: adapter->link_duplex = DUPLEX_FULL;
                                        break;
                                default:
                                        adapter->link_duplex = -1; // unknown mode
                                        break;
                                }
                                if (unm_nic_phy_read(adapter,
                                    physical_port[adapter->portnum],
                                    UNM_NIU_GB_MII_MGMT_ADDR_AUTONEG, 
                                    (unm_crbword_t *)&autoneg)!= 0)
                                        adapter->link_autoneg = autoneg;
                        } else {
                                adapter->state = -1;
                                adapter->link_speed = -1;
                                adapter->link_duplex = -1;
                        }
                } else {
			adapter->state = -1;
			adapter->link_speed = -1;
			adapter->link_duplex = -1;
		}
        }
}

#define FLASH_SIZE      (0x100000)

int
unm_nic_get_flash_size(void)
{
        return FLASH_SIZE;
}

void unm_nic_get_serial_num(struct unm_adapter_s *adapter)
{
	unm_user_info_t user_info;
	int              i;
	int              addr = USER_START;
	uint32_t         *ptr32;

	ptr32 = (U32 *)&user_info;
	for (i=0; i<sizeof(unm_user_info_t)/sizeof(uint32_t); i++) {
		if (rom_fast_read(adapter, addr, ptr32) == -1) {
			printk("%s: ERROR reading %s board userarea.\n",
					unm_nic_driver_name, unm_nic_driver_name);
			return;
		}
		ptr32++;
		addr+= sizeof(uint32_t);
	}

	memcpy(adapter->id, user_info.serial_num,sizeof(user_info.serial_num));
#if 0
	printk("%s adapter_id=%x size=%x\n",__FUNCTION__,adapter->id,sizeof(user_info.serial_num));
	for(i = 0; i < 32;i++) {
		printk("%x ",(unsigned char)*(adapter->id+i));
	}
#endif
	//      printk("%s: adapter_id=%x\n",__FUNCTION__,adapter->id);
}

void unm_nic_flash_print(struct unm_adapter_s* adapter)
{
	int valid = 1;
	uint32_t fw_major = 0;
	uint32_t fw_minor = 0;
	uint32_t fw_build = 0;

	unm_board_info_t* board_info = &(adapter->ahw.boardcfg);
	if (board_info->magic != UNM_BDINFO_MAGIC) {
		printk("UNM Unknown board config, Read 0x%x expected as 0x%x\n", board_info->magic,
			UNM_BDINFO_MAGIC);
		valid = 0;
	}
	if (board_info->header_version != UNM_BDINFO_VERSION) {
		printk("UNM Unknown board config version."
			" Read %x, expected %x\n",
			board_info->header_version, UNM_BDINFO_VERSION);
		valid = 0;
	}
	if(valid) {
        char            *brd_name = NULL;
        unm_user_info_t  user_info;
        int              i;
        int              addr = USER_START;
        uint32_t         *ptr32;

        ptr32 = (U32 *)&user_info;
        for (i=0; i<sizeof(unm_user_info_t)/sizeof(uint32_t); i++) {
             if (rom_fast_read(adapter, addr, ptr32) == -1) {
                 printk("%s: ERROR reading %s board userarea.\n",
                         unm_nic_driver_name, unm_nic_driver_name);
                 return;
             }
             ptr32++;
             addr+= sizeof(uint32_t);
        }
        GET_BRD_NAME_BY_TYPE(board_info->board_type, brd_name);
        printk(KERN_INFO "%s %s Board S/N %s  Chip id 0x%x\n",
                        DRIVER_NAME, brd_name, user_info.serial_num,
                        board_info->chip_id);

		read_lock(&adapter->adapter_lock);
		fw_major = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_MAJOR));
		fw_minor = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_MINOR));
		fw_build = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
				UNM_FW_VERSION_SUB));
		read_unlock(&adapter->adapter_lock);
                printk(KERN_INFO "%s Firmware version %d.%d.%d\n",
                        DRIVER_NAME, fw_major, fw_minor,
                        fw_build);
	}
	if(fw_major != _UNM_NIC_LINUX_MAJOR) {
		printk(KERN_ERR"The mismatch in driver version and firmware "
			"version major number\n"
			"Driver version major number = %d \t"
			"Firmware version major number = %d \n",
			_UNM_NIC_LINUX_MAJOR, fw_major);
		adapter->driver_mismatch = 1;
	}
	if(fw_minor != _UNM_NIC_LINUX_MINOR) {
		printk(KERN_ERR"The mismatch in driver version and firmware "
			"version minor number\n"
			"Driver version minor number = %d \t"
			"Firmware version minor number = %d \n",
			_UNM_NIC_LINUX_MINOR, fw_minor);
		adapter->driver_mismatch = 1;
	}
	if(adapter->driver_mismatch) {
		printk(KERN_INFO"Use the driver with version no %d.%d.xxx\n"
			,fw_major,fw_minor);
	} else {
		DPRINTK(1, INFO,"The major and minor number of driver and "
			"firmware versions match\n");
	}
}
