/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Source file for NIC routines to initialize the Phantom Hardware
 *
 * $Id: unm_nic_init.c,v 1.55.2.1 2007/02/24 02:05:59 datton Exp $
 *
 */
#include <linux/netdevice.h>
#include <linux/delay.h>
#include "queue.h"
#include "unm_nic.h"
#include "unm_nic_hw.h"
#include "nic_cmn.h"
#include "nic_phan_reg.h"
#include "unm_nic_ioctl.h"
#include "nic_phan_reg.h"

struct crb_addr_pair { 
        long addr;
        long data;
};

#define MAX_CRB_XFORM 60
static unsigned int crb_addr_xform[MAX_CRB_XFORM];
#define ADDR_ERROR ((unsigned long ) 0xffffffff )

#define crb_addr_transform(name) \
        crb_addr_xform[UNM_HW_PX_MAP_CRB_##name] = \
        UNM_HW_CRB_HUB_AGT_ADR_##name << 20
static void
crb_addr_transform_setup(void)
{
        crb_addr_transform(XDMA);
        crb_addr_transform(TIMR);
        crb_addr_transform(SRE);
        crb_addr_transform(SQN3);
        crb_addr_transform(SQN2);
        crb_addr_transform(SQN1);
        crb_addr_transform(SQN0);
        crb_addr_transform(SQS3);
        crb_addr_transform(SQS2);
        crb_addr_transform(SQS1);
        crb_addr_transform(SQS0);
        crb_addr_transform(RPMX7);
        crb_addr_transform(RPMX6);
        crb_addr_transform(RPMX5);
        crb_addr_transform(RPMX4);
        crb_addr_transform(RPMX3);
        crb_addr_transform(RPMX2);
        crb_addr_transform(RPMX1);
        crb_addr_transform(RPMX0);
        crb_addr_transform(ROMUSB);
        crb_addr_transform(SN);
        crb_addr_transform(QMN);
        crb_addr_transform(QMS);
        crb_addr_transform(PGNI);
        crb_addr_transform(PGND);
        crb_addr_transform(PGN3);
        crb_addr_transform(PGN2);
        crb_addr_transform(PGN1);
        crb_addr_transform(PGN0);
        crb_addr_transform(PGSI);
        crb_addr_transform(PGSD);
        crb_addr_transform(PGS3);
        crb_addr_transform(PGS2);
        crb_addr_transform(PGS1);
        crb_addr_transform(PGS0);
        crb_addr_transform(PS);
        crb_addr_transform(PH);
        crb_addr_transform(NIU);
        crb_addr_transform(I2Q);
        crb_addr_transform(EG);
        crb_addr_transform(MN);
        crb_addr_transform(MS);
        crb_addr_transform(CAS2);
        crb_addr_transform(CAS1);
        crb_addr_transform(CAS0);
        crb_addr_transform(CAM);
        crb_addr_transform(C2C1);
        crb_addr_transform(C2C0);
        crb_addr_transform(SMB);
}

/*
 * decode_crb_addr(0 - utility to translate from internal Phantom CRB address
 * to external PCI CRB address.
 */
unsigned long
decode_crb_addr (unsigned long addr)
{
        int i;
        unsigned long base_addr, offset, pci_base;

        crb_addr_transform_setup();

        pci_base = ADDR_ERROR;
        base_addr = addr & 0xfff00000;
        offset = addr & 0x000fffff;

        for (i=0; i< MAX_CRB_XFORM; i++) {
                if (crb_addr_xform[i] == base_addr) {
                        pci_base = i << 20;
                        break;
                }
        }
        if (pci_base == ADDR_ERROR) {
                return pci_base;
        } else {
                return (pci_base + offset);
        }
}

static long rom_max_timeout= 100;
static long rom_lock_timeout= 10000;

int
rom_lock(unm_adapter *adapter)
{
    int i;
    int done = 0, timeout = 0;

    while (!done) {
        /* acquire semaphore2 from PCI HW block */
        unm_nic_read_w0(adapter, UNM_PCIE_REG(PCIE_SEM2_LOCK), &done);
        if (done == 1)
            break;
        if (timeout >= rom_lock_timeout) {
            return -1;
        }
        timeout++;
        /*
         * Yield CPU
         */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if(!in_atomic())
                schedule();
        else {
#endif
                for(i = 0; i < 20; i++)
                        cpu_relax();    /*This a nop instr on i386*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        }
#endif
    }
    unm_nic_reg_write(adapter, UNM_ROM_LOCK_ID, ROM_LOCK_DRIVER);
    return 0;
}

int
rom_unlock(unm_adapter *adapter)
{
    int val;

    /* release semaphore2 */
    unm_nic_read_w0(adapter, UNM_PCIE_REG(PCIE_SEM2_UNLOCK), &val);

    return 0;
}

int
wait_rom_done (unm_adapter *adapter)
{
        long timeout=0;
        long done=0 ;

        /*DPRINTK(1,INFO,"WAIT ROM DONE \n");*/
        /*printk(KERN_INFO "WAIT ROM DONE \n");*/

        while (done == 0) {
                done = unm_nic_reg_read(adapter, UNM_ROMUSB_GLB_STATUS);
                done &=2;
                timeout++;
                if (timeout >= rom_max_timeout) {
                        printk( "Timeout reached  waiting for rom done");
                        return -1;
                }
        }
        return 0;
}

int
do_rom_fast_read (unm_adapter *adapter, int addr, int *valp)
{
        /*DPRINTK(1,INFO,"ROM FAST READ \n");*/
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ADDRESS, addr);
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ABYTE_CNT, 3);
        udelay(100);   /* prevent bursting on CRB */
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_DUMMY_BYTE_CNT, 0);
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_INSTR_OPCODE, 0xb);
        if (wait_rom_done(adapter)) {
                printk("Error waiting for rom done\n");
                return -1;
        }
        //reset abyte_cnt and dummy_byte_cnt
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ABYTE_CNT, 0);
        udelay(100);   /* prevent bursting on CRB */
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_DUMMY_BYTE_CNT, 0);

        *valp = unm_nic_reg_read(adapter, UNM_ROMUSB_ROM_RDATA);
        return 0;
}

int
rom_fast_read (struct unm_adapter_s *adapter, int addr, int *valp)
{
    int ret;

    if (rom_lock(adapter) != 0) {
        printk("rom_lock failed\n");
        return -1;
    }
    ret = do_rom_fast_read(adapter, addr, valp);
    rom_unlock(adapter);
    return ret;
}

int 
rom_wren(unm_adapter *adapter)
{
        // set write enable latch in ROM status register
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ABYTE_CNT, 0);
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_INSTR_OPCODE, M25P_INSTR_WREN);
        if (wait_rom_done(adapter)) {
                printk("Error waiting for rom done\n");
                return -1;
        }
        return 0;
}

int 
do_rom_rdsr(unm_adapter *adapter)
{
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_INSTR_OPCODE, M25P_INSTR_RDSR);
        if (wait_rom_done(adapter)) {
                printk("Error waiting for rom done\n");
                return -1;
        }

        return unm_nic_reg_read(adapter, UNM_ROMUSB_ROM_RDATA);
}

int 
rom_wip_poll(unm_adapter *adapter)
{
        long timeout=0;
        long wip=1;
        int val;

        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ABYTE_CNT, 0);
        while (wip != 0) {
                val = do_rom_rdsr(adapter);
                wip = val & 1;
                timeout++;
                if (timeout > rom_max_timeout) {
                        printk( "Timeout reached  waiting for rom wip");
                        return -1;
                }
        }
        return 0;
}


int 
do_rom_fast_write (unm_adapter *adapter, int addr, int data)
{
        int ret;

        if (rom_wren(adapter)) {
                return -1;
        }

        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_WDATA, data);
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ADDRESS, addr);
        udelay(100);   /* prevent bursting on CRB */
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_ABYTE_CNT, 3);
        unm_nic_reg_write(adapter, UNM_ROMUSB_ROM_INSTR_OPCODE, M25P_INSTR_PP);
        if (wait_rom_done(adapter)) {
                printk("Error waiting for rom done\n");
                return -1;
        }

        ret = rom_wip_poll(adapter);
        return ret;
}

int
rom_fast_write (struct unm_adapter_s *adapter, int addr, int val)
{
        int ret;

        if (rom_lock(adapter) != 0) {
                printk("rom_lock failed\n");
                return -1;
        }
        ret = do_rom_fast_write(adapter, addr, val);
        rom_unlock(adapter);
        return ret;
}
/* Error codes */
#define FPGA_NO_ERROR 0
#define FPGA_FILE_ERROR 1 
#define FPGA_ILLEGAL_PARAMETER 2 
#define FPGA_MEMORY_ERROR 3

/* macros */
#ifdef VIRTEX
#define RESET_STATE  changeState(adapter,1,32)
#else
#define RESET_STATE  changeState(adapter,1,5)
#endif

#define GO_TO_RTI RESET_STATE;\
                  changeState(adapter,0,1)

/* 
 *  TCLK = GPIO0 = 0x60
 *  TRST = GPIO1 = 0x64
 *  TMS  = GPIO4 = 0x70
 *  TDO  = GPIO5 = 0xc
 *  TDI  = GPIO8 = 0x80
 *
 */
#define TCLK (UNM_CRB_ROMUSB + 0x60)
#define TRST (UNM_CRB_ROMUSB + 0x64)
#define TMS  (UNM_CRB_ROMUSB + 0x70)
#define TDI  (UNM_CRB_ROMUSB + 0x80)
#define TDO  (UNM_CRB_ROMUSB + 0xc)

#define TDO_SHIFT 5

#define ASSERT_TRST   \
        do { \
                unm_nic_reg_write(adapter, TRST, 2);\
                unm_nic_reg_write(adapter, TMS, 2); \
                unm_nic_reg_write(adapter, TDI, 2); \
                unm_nic_reg_write(adapter, TCLK, 2);\
                unm_nic_reg_write(adapter, TRST, 3);\
        } while (0)

#define CLOCK_IN_BIT(tdi,tms)  \
        do { \
                unm_nic_reg_write(adapter, TRST, 3); \
                unm_nic_reg_write(adapter, TMS, 2 | (tms)); \
                unm_nic_reg_write(adapter, TDI, 2 | (tdi)); \
                unm_nic_reg_write(adapter, TCLK, 2); \
                unm_nic_reg_write(adapter, TCLK, 3); \
                unm_nic_reg_write(adapter, TCLK, 2);\
        } while (0)

#define CLOCK_OUT_BIT(bit,tms) \
        do { \
                unm_nic_reg_write(adapter, TRST, 3); \
                unm_nic_reg_write(adapter, TMS, 2 | (tms)); \
                unm_nic_reg_write(adapter, TDI, 2); \
                unm_nic_reg_write(adapter, TCLK, 2); \
                unm_nic_reg_write(adapter, TCLK, 3); \
                unm_nic_reg_write(adapter, TCLK, 2); \
                (bit) = (unm_nic_reg_read(adapter, TDO) >> TDO_SHIFT) & 1; \
        } while (0)

/* boundary scan instructions */
#define CMD_EXTEST    0x0
#define CMD_CAPTURE   0x1
#define CMD_IDCODE    0x2
#define CMD_SAMPLE    0x3

//Memory BIST
#define CMD_MBSEL     0x4
#define CMD_MBRES     0x5

//Logic BIST
#define CMD_LBSEL     0x6
#define CMD_LBRES     0x7
#define CMD_LBRUN    0x18 

//Memory Interface
#define CMD_MEM_WDAT  0x8
#define CMD_MEM_ACTL  0x9
#define CMD_MEM_READ  0xa

//Memory Interface
#define CMD_CRB_WDAT  0xb
#define CMD_CRB_ACTL  0xc
#define CMD_CRB_READ  0xd

#define CMD_TRISTATE  0xe
#define CMD_CLAMP     0xf

#define CMD_STATUS    0x10
#define CMD_XG_SCAN   0x11
#define CMD_BYPASS    0x1f 

#ifdef VIRTEX
#define CMD_LENGTH_BITS 6
#else
#define CMD_LENGTH_BITS 5
#endif


/* This is the TDI bit that will be clocked out for don't care value */
#define TDI_DONT_CARE 0

#define TMS_LOW  0
#define TMS_HIGH 1

#define ID_REGISTER_SIZE_BITS 32

#define EXIT_IR 1
#define NO_EXIT_IR 0

#define TAP_DELAY() 

__inline__ void
changeState(unm_adapter *adapter,int tms,int tck) 
{
        int i;
        DPRINTK(1, INFO, "changing state tms: %d tck: %d\n", tms, tck);
        for (i = 0; i< tck; i++)
        {
                CLOCK_IN_BIT(TDI_DONT_CARE,tms);
        }
}

// Assumes we are in RTI, will return to RTI
__inline__ int 
loadInstr(unm_adapter *adapter,int instr,int exit_ir)
{
        int i,j;

        DPRINTK(1, INFO, "in loaderinstr instr: %d exit_ir %d\n", instr,
                                exit_ir);

        // go to Select-IR
        changeState(adapter,1,2);

        // go to Shift-IR
        changeState(adapter,0,2);

#ifdef VIRTEX 
        for (i = 0; i< (CMD_LENGTH_BITS * 7); i++)
        {
                CLOCK_IN_BIT(1,TMS_LOW);
        }
#endif
        for (i = 0; i< (CMD_LENGTH_BITS-1); i++)
        {
                j= (instr>>i) & 1;
                CLOCK_IN_BIT(j,TMS_LOW);
        }

        /* clock out last bit, and transition to next state */
        j= (instr >> (CMD_LENGTH_BITS-1)) & 1;
        // last bit, exit into Exit1-IR
        CLOCK_IN_BIT(j,1);
        // go to Update-IR
        changeState(adapter,1,1);
        // go to RTI 
        changeState(adapter,0,1);
        return FPGA_NO_ERROR;
}

int
getData(unm_adapter *adapter,u32 *data,int len, int more)
{
        u32 temp=0;
        int i, bit;
        DPRINTK(1, INFO, "doing getData data: %p len: %d more: %d\n", 
                                data, len,more);
        // go to Select-DR-Scan
        changeState(adapter,1,1);
        // go to shift-DR
        changeState(adapter,0,1);
#ifdef VIRTEX 
        // dummy reads
        for (i=0; i< 6; i++) {
                CLOCK_OUT_BIT(bit,0);
        }
#endif
        for (i=0; i< (len); i++) {
                CLOCK_OUT_BIT(bit,0);
                temp |= (bit << i);
        }
        if (!more) {
                // go to Exit1-DR
                changeState(adapter,1,1);
                // go to Update DR
                changeState(adapter,1,1);
                // go to RTI
                changeState(adapter,0,1);
        }

        *data = temp;

        return 0;
} 


int
get_status(unm_adapter *adapter)
{
        int status;

        DPRINTK(1, INFO, "doing get_status: %p\n", adapter);
        // tap_inst_wr(INST_STATUS)
        loadInstr(adapter,CMD_STATUS,EXIT_IR); 
        getData(adapter,&status,1,0);
        //printf("Status: 0x%02x\n",status);
        return status;
}
// assumes start in RTI, will return to RTI unless more is set
int
getData64(unm_adapter *adapter,u64 *data,int len, int more)
{
        u64 temp=0;
        u64 i, bit;
        DPRINTK(1, INFO, "getData64 data %p, len %d more %d\n",
                                data, len, more);
        // go to Select-DR-Scan
        changeState(adapter,1,1);
        // go to shift-DR
        changeState(adapter,0,1);
#ifdef VIRTEX 
        // dummy reads
        for (i=0; i< 6; i++) {
                CLOCK_OUT_BIT(bit,0);
        }
#endif
        for (i=0; i< (len); i++) {
                CLOCK_OUT_BIT(bit,0);
                temp |= (bit << i);
        }

        if (!more) {
                // go to Exit1-DR
                changeState(adapter,1,1);
                // go to Update DR
                changeState(adapter,1,1);
                // go to RTI
                changeState(adapter,0,1);
        }

        //temp |= (bit << (len - 1));
        //printf("Data read: 0x%016llx\n",temp);
        *data = temp;
        return 0;
} 

// assumes start in shift_DR, will return to RTI unless more is set
int
loadData (unm_adapter *adapter,u64 data, int len, int more)
{
        int i, bit;

        DPRINTK(1, INFO, "loading data %llx data %d more %d\n",
                data, len, more);
        for (i=0; i< (len-1); i++) {
                bit = (data >> i) & 1;
                CLOCK_IN_BIT(bit,0);
        }
        if (more) {
                bit = (data >> (len-1)) & 1;
                CLOCK_IN_BIT(bit,0);
        } else {
                bit = (data>>(len - 1)) & 1;
                // last data, go to Exit1-DR
                CLOCK_IN_BIT(bit,1);
                // go to Update DR
                changeState(adapter,1,1);
                // go to RTI
                changeState(adapter,0,1);
        }

        return 0;
}

#define CRB_REG_EX_PC                   0x3c

static void nx_msleep(unsigned long msecs)
{
        set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_timeout((HZ * msecs + 999) / 1000);
}

int pinit_from_rom(unm_adapter *adapter, int verbose)
{
        int addr, val,status;
        int n, i;
        int init_delay=0;
        struct crb_addr_pair *buf;
        unsigned long off;
        unsigned long flags;

        struct crb_addr_pair {
                long addr;
                long data;
        };

        // resetall
        status = unm_nic_get_board_info(adapter);
        if (status)
                printk ("%s: pinit_from_rom: Error getting board info\n",
				unm_nic_driver_name);

        UNM_CRB_WRITELIT_ADAPTER(UNM_ROMUSB_GLB_SW_RESET, 0xffffffff, adapter);

        if (verbose) {
                int val;
                if (rom_fast_read(adapter, 0x4008, &val) == 0)
                        printk("P2 ROM board type: 0x%08x\n", val);
                else
                        printk("Could not read board type\n");
                if (rom_fast_read(adapter, 0x400c, &val) == 0)
                        printk("P2 ROM board  num: 0x%08x\n", val);
                else
                        printk("Could not read board number\n");
                if (rom_fast_read(adapter, 0x4010, &val) == 0)
                        printk("P2 ROM chip   num: 0x%08x\n", val);
                else
                        printk("Could not read chip number\n");
        }

        if (rom_fast_read(adapter, 0, &n) == 0 &&
             (n & 0x800000000ULL)) {
                n &= ~0x80000000ULL;
                if (n  < 1024) {
                        if (verbose) printk("%s: %d CRB init values found"
                                                " in ROM.\n",unm_nic_driver_name, n);
                } else {
			printk("%s:n=0x%x Error! UNM card flash not"
                                        " initialized.\n", __FUNCTION__, n);
                        return -1;
                }
                buf = kmalloc(n*sizeof(struct crb_addr_pair), GFP_KERNEL);
                if (buf==NULL) {
                        printk("%s: pinit_from_rom: Unable to malloc memory.\n",
				unm_nic_driver_name);
                        return -1;
                }
                for  (i=0; i< n; i++) {
                        if (rom_fast_read(adapter, 8*i + 4, &val) != 0 ||
                            rom_fast_read(adapter, 8*i + 8, &addr) != 0)
                                goto error_ret;

                        buf[i].addr=addr;
                        buf[i].data=val;

                        if (verbose)
                                printk("%s: PCI:     0x%08x == 0x%08x\n",
                                       unm_nic_driver_name,
                                       (unsigned int)decode_crb_addr(
						(unsigned long)addr), val);
                }
                for  (i=0; i< n; i++) {
                        off = decode_crb_addr((unsigned long)buf[i].addr) + UNM_PCI_CRBSPACE;
                        /* skipping cold reboot MAGIC */
                        if (off == UNM_CAM_RAM(0x1fc)) {
                                continue;
                        }
                        if (off == ADDR_ERROR) {
                                printk("%s: Err: Unknown addr: 0x%08lx\n",
                                       unm_nic_driver_name, buf[i].addr);
                                continue;
                        }                        
                        /* After writing this register, HW needs time for CRB */
                        /* to quiet down (else crb_window returns 0xffffffff) */
                        if (off == UNM_ROMUSB_GLB_SW_RESET) {
                                init_delay=1;
                                /* hold xdma in reset also */
                                buf[i].data = 0x8000ff;
                        }


                        if (ADDR_IN_WINDOW1(off))
                        {
                                read_lock(&adapter->adapter_lock);
                                UNM_NIC_PCI_WRITE_32(buf[i].data, CRB_NORMALIZE(adapter, off));
                                read_unlock(&adapter->adapter_lock);
                        }
                        else
                        {
                                write_lock_irqsave(&adapter->adapter_lock, flags);
                                //_unm_nic_write_crb(adapter, off, buf[i].data);
                                unm_nic_pci_change_crbwindow(adapter, 0);
                                off = pci_base_offset(adapter, off);
                                UNM_NIC_PCI_WRITE_32(buf[i].data, (void *)off);

                                unm_nic_pci_change_crbwindow(adapter, 1);
                                write_unlock_irqrestore(&adapter->adapter_lock, flags);
                        }
                        if (init_delay==1) {
                            nx_msleep(1000);
                            init_delay=0;
                        }

                        nx_msleep(1);
                }
                kfree(buf);

                // disable_peg_cache_all
                // unreset_net_cache
		val = UNM_CRB_READ_VAL_ADAPTER(UNM_ROMUSB_GLB_SW_RESET, adapter);
                UNM_CRB_WRITELIT_ADAPTER(UNM_ROMUSB_GLB_SW_RESET, (val & 0xffffff0f), adapter);
                // p2dn replyCount
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_D+0xec, 0x1e, adapter);
                // disable_peg_cache 0
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_D+0x4c,8, adapter);
                // disable_peg_cache 1
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_I+0x4c,8, adapter);

                // peg_clr_all
                // peg_clr 0
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_0+0x8,0, adapter);
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_0+0xc,0, adapter);
                // peg_clr 1
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_1+0x8,0, adapter);
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_1+0xc,0, adapter);
                // peg_clr 2
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_2+0x8,0, adapter);
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_2+0xc,0, adapter);
                // peg_clr 3
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_3+0x8,0, adapter);
                UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PEG_NET_3+0xc,0, adapter);

        } else {
            printk("NX: ERROR Reading crb_init area\n");
            return -1;

        }

        return 0;
error_ret:
        return -1;
}

#define CACHE_DISABLE 1
#define CACHE_ENABLE  0

int
phantom_init(struct unm_adapter_s *adapter, int pegtune_val)
{
    int val=0;


    /* For initial boot after power up (indicated by PEGTUNE_DONE reg being
    *  cleared), we should wait for HW to wake up before allowing any other
    *  card access (ethtool, etc). We don't want to interfere with ROM shadowing
    *  or the bootloader.
    */
    if (!pegtune_val) {
#if 1
            do {
                    long timeout = 1 * HZ;
                    while(timeout) {
                            if (signal_pending(current)) {
                                    printk( "%s: Got a signal, exiting\n", __FUNCTION__ );
                                    printk( "%s: val=0x%x, pegtune_val=0x%x\n", __FUNCTION__,
                                            val, pegtune_val );
                                    return -1;
                            }
                            set_current_state(TASK_INTERRUPTIBLE);
                            timeout = schedule_timeout(timeout);
                    }
                    val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter, CRB_CMDPEG_STATE));
                    pegtune_val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                    UNM_ROMUSB_GLB_PEGTUNE_DONE));
            } while ((val != PHAN_INITIALIZE_COMPLETE) && (val != PHAN_INITIALIZE_ACK));
#else
            int loops=0;
            val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter, CRB_CMDPEG_STATE));
            while (val != PHAN_INITIALIZE_COMPLETE && loops < 2000000) {
                    udelay(100);
                    schedule();
                    /* Window 1 call */
                    val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                                        CRB_CMDPEG_STATE));
                    loops++;
            }
            if (val != PHAN_INITIALIZE_COMPLETE) {
                    printk("WARNING: Initial boot wait loop failed...\n");
            return -1;
            }
#endif
    }

    return 0;
}

int
load_firmware(struct unm_adapter_s *adapter)
{
        int  i;
        long data, size = 0;
        long flashaddr = BOOTLD_START, memaddr = BOOTLD_START;

        size = (16*1024)/4;

        read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(1, CRB_NORMALIZE(adapter,
                                                UNM_ROMUSB_GLB_CAS_RST));
        read_unlock(&adapter->adapter_lock);


        for (i = 0; i < size; i++) {
                if (rom_fast_read(adapter, flashaddr, (int *)&data) != 0) {
                        DPRINTK(1,ERR, "Error in rom_fast_read(). Will skip"
                                "loading flash image\n");
                        return -1;
                }
                unm_nic_pci_mem_write(adapter, memaddr, &data, 4);
                flashaddr += 4;
                memaddr   += 4;
        }
       udelay(100);
       read_lock(&adapter->adapter_lock);
       UNM_NIC_PCI_WRITE_32(0x3fff, CRB_NORMALIZE(adapter,
                                               UNM_ROMUSB_GLB_CHIP_CLK_CTRL));
       UNM_NIC_PCI_WRITE_32(0, CRB_NORMALIZE(adapter,
                                               UNM_ROMUSB_GLB_CAS_RST));
       read_unlock(&adapter->adapter_lock);
       return 0;

}
