/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 *  Main source file for NetXen NIC Driver on Linux
 *
 *  $Id: unm_nic_main.c,v 1.343.2.9 2007/03/17 08:28:59 sanjeev Exp $
 */


#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/types.h>
//#include <linux/kgdb-defs.h>
#include <linux/compiler.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,27) || defined(__VMKERNEL_MODULE__)
#include <linux/ethtool.h>
#endif
#include <linux/mii.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>

#include <linux/mm.h>
#ifndef __VMKERNEL_MODULE__
#include <linux/mman.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#include <linux/wrapper.h>
#endif

#include <asm/system.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include "kernel_compatibility.h"
#include "unm_nic_hw.h"
#include "unm_nic_config.h"
#ifdef __VMKERNEL_MODULE__
#include "vmklinux_dist.h"
#include "smp_drv.h"
#include "asm/page.h"
#endif

#undef UNM_LOOPBACK
#undef SINGLE_DMA_BUF
#define UNM_NIC_HW_CSUM
#define UNM_NIC_NAPI
//#undef UNM_NIC_NAPI

#if defined(PEGNET_NIC) && (UNM_HOST_MSG_TRANSPORT==UNM_HOST_MSG_TRANSPORT_NIC)

/*
 * Define NAPI when the NIC path is taken for messages from the card to host.
 * Currently non NAPI does not work because code in pegnet part uses bottom
 * half version of the spin_lock so cannot be used in irq routines. It hangs
 * the system.
 */
#ifndef UNM_NIC_NAPI
#define UNM_NIC_NAPI
#endif  /* UNM_NIC_NAPI */

#endif  /* PEGNET_NIC */


// #define UNM_SKB_TASKLET

#if 1
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define UNM_NETIF_F_TSO
#endif
#endif

#ifndef DMA_64BIT_MASK
#define DMA_64BIT_MASK			 0xffffffffffffffffULL
#endif

#define DMA_35BIT_MASK                   0x00000007ffffffffULL

#ifndef DMA_32BIT_MASK
#define DMA_32BIT_MASK                   0x00000000ffffffffULL
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,19)
#if defined(CONFIG_VLAN_8021Q) || defined(CONFIG_VLAN_8021Q_MODULE)
//#define UNM_NIC_HW_VLAN 1
#else
#define UNM_NIC_HW_VLAN 0
#endif
#else
#define UNM_NIC_HW_VLAN 0
#endif                 //LINUX_KERNEL_VERSION >= 2.4.19

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
#define	CHECKSUM_HW	CHECKSUM_PARTIAL
#endif /* KERNEL_VERSION(2,4,19) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
#define	TASK_PARAM		struct work_struct *
#define	NX_INIT_WORK(a, b, c)	INIT_WORK(a, b)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define	TASK_PARAM		unsigned long
#define	NX_INIT_WORK(a, b, c)	INIT_WORK(a, b, c)
#else
#define	TASK_PARAM		unsigned long
#define	NX_INIT_WORK(a, b, c)	INIT_TQUEUE(a, b, c)
#endif

#include "unm_nic.h"

#define DEFINE_GLOBAL_RECV_CRB
#include "nic_phan_reg.h"

#include "unm_nic_ioctl.h"
#include "nic_cmn.h"
#include "unm_nic_config.h"

#include "unm_nic_hw.h"
#include "unm_version.h"
#include "unm_brdcfg.h"
#include "nx_nic_linux_tnic_api.h"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_LICENSE("GPL");

/*static int unm_nic_debug = NETIF_MSG_DRV | NETIF_MSG_PROBE; TODO */
int unm_nic_debug;

char unm_nic_driver_name[]   = DRIVER_NAME;
char unm_nic_driver_string[] = DRIVER_VERSION_STRING
                               UNM_NIC_LINUX_VERSIONID
                               "-"
                               UNM_NIC_BUILD_NO
                               " generated "
                               UNM_NIC_TIMESTAMP;

#ifndef ARCH_KMALLOC_MINALIGN
#define ARCH_KMALLOC_MINALIGN 0
#endif
#ifndef ARCH_KMALLOC_FLAGS
#define ARCH_KMALLOC_FLAGS SLAB_HWCACHE_ALIGN
#endif

#define UNM_NETDEV_WEIGHT 120
#define ADAPTER_UP_MAGIC 777

kmem_cache_t           *cmd_desc_cache = NULL;
kmem_cache_t           *cmd_buff_cache = NULL;
cmdDescType0_t         *cmd_desc_array[NR_CPUS];
struct unm_cmd_buffer  *cmd_buff_array[NR_CPUS];

static int unm_nic_probe_err = 0;

#define RCV_DESC_RINGSIZE         (sizeof(rcvDesc_t) * adapter->MaxRxDescCount)
#define STATUS_DESC_RINGSIZE    (sizeof(statusDesc_t)* adapter->MaxRxDescCount)
#define JUMBO_DESC_RINGSIZE  (sizeof(rcvDesc_t) * adapter->MaxJumboRxDescCount)
#define LRO_DESC_RINGSIZE      (sizeof(rcvDesc_t) * adapter->MaxLroRxDescCount)
#define TX_RINGSIZE   (sizeof(struct unm_cmd_buffer) * adapter->MaxTxDescCount)
#define RCV_BUFFSIZE  (sizeof(struct unm_rx_buffer) * rcv_desc->MaxRxDescCount)

#define UNM_NIC_INT_RESET       0x2004
#define UNM_DB_MAPSIZE_BYTES    0x1000
#define UNM_CMD_PRODUCER_OFFSET                 0
#define UNM_RCV_STATUS_CONSUMER_OFFSET          0
#define UNM_RCV_PRODUCER_OFFSET                 0
#ifdef PEGNET_NIC
#define UNM_RCV_PEG_DB_ID  2
#define UNM_XMIT_PEG_DB_ID 3
#else
#define UNM_RCV_PEG_DB_ID  1
#define UNM_XMIT_PEG_DB_ID 2
#endif

uint8_t  nx_p2_id = NX_P2_C0;
unsigned long long unm_dma_mask = DMA_35BIT_MASK;
#ifdef BOUNCE_BUFFERS
/* Number of bounce buffers. Has to be a power of two */
#define NUM_BOUNCE 256
char *tx_bounce_ptr[NUM_BOUNCE];
char *rx_bounce_ptr[NUM_BOUNCE];

struct bounce_index tx_bounce_head, tx_bounce_tail,
               rx_bounce_head, rx_bounce_tail;

spinlock_t rx_bounce_lock, tx_bounce_lock;

#define BOUNCE_BUFFER_ORDER 2
#define BOUNCE_BUFFER_SIZE (PAGE_SIZE << BOUNCE_BUFFER_ORDER)
#endif

/* Local functions to UNM NIC driver */
static int  __devinit unm_nic_probe(struct pci_dev *pdev,
                                    struct pci_device_id *ent);
static void __devexit unm_nic_remove(struct pci_dev *pdev);
static int    unm_nic_open(struct net_device *netdev);
static int    unm_nic_close(struct net_device *netdev);
static int    unm_nic_xmit_frame(struct sk_buff *, struct net_device *);
static int    unm_nic_set_mac(struct net_device *netdev, void *p);
static int    unm_nic_change_mtu(struct net_device *netdev, int new_mtu);
static int    unm_nic_hw_resources(struct unm_adapter_s *adapter);
static void   unm_nic_set_multi(struct net_device *netdev);
static void   initialize_adapter_sw(struct unm_adapter_s *adapter);
static int    init_firmware (struct unm_adapter_s *adapter);
static void   unm_post_rx_buffers(struct unm_adapter_s *adapter, uint32_t ctx,
                                    uint32_t type);
static void   unm_post_rx_buffers_nodb(struct unm_adapter_s *adapter, uint32_t ctx,
                                    uint32_t type);
static void   unm_process_rcv(struct unm_adapter_s *adapter,int ctx,
                                statusDesc_t *desc);
static void   unm_tx_timeout(struct net_device *netdev);
static void   unm_tx_timeout_task(TASK_PARAM adapid);
static int    unm_process_cmd_ring(unsigned long data);
static void   unm_nic_down(struct unm_adapter_s *adapter);
static int    unm_nic_do_ioctl(struct unm_adapter_s *adapter, void*u_data);
static int    initialize_adapter_hw(struct unm_adapter_s *adapter);
static void   unm_watchdog_task(TASK_PARAM adapid);
void   unm_nic_enable_int(unm_adapter *adapter);
static void   unm_nic_disable_int(unm_adapter *adapter);
static uint32_t unm_process_rcv_ring(struct unm_adapter_s *, int, int);
static void   unm_nic_clear_stats(struct unm_adapter_s *);
static struct net_device_stats *unm_nic_get_stats(struct net_device *netdev);
static void   unm_nic_irq_enable(struct unm_adapter_s *adapter);
static int    unm_handle_int(unm_adapter *, struct net_device *);
static int    unm_nic_ioctl(struct net_device *netdev,
                            struct ifreq *ifr, int cmd);
static int    unm_nic_port_read_proc(char *buf, char **start, off_t offset,
                                        int count, int *eof, void *data);
#ifdef UNM_NIC_NAPI
static int    unm_nic_rx_has_work(struct unm_adapter_s *adapter);
static int    unm_nic_tx_has_work(struct unm_adapter_s *adapter);
static int    unm_nic_poll(struct net_device *dev, int *budget);
#endif
#ifdef CONFIG_NET_POLL_CONTROLLER
static void unm_nic_poll_controller(struct net_device *netdev);
#endif
#ifdef UNM_NETIF_F_TSO
static inline void unm_tso_check(struct unm_adapter_s *adapter,
                                 cmdDescType0_t *desc,
                                 struct sk_buff *skb);
#endif
static inline void  unm_tx_csum(struct unm_adapter_s *adapter,
                                cmdDescType0_t *desc,
                                struct sk_buff *skb);
#ifdef OLD_KERNEL
static void   unm_intr(int irq, void *data, struct pt_regs *regs);
#else
static irqreturn_t unm_intr(int irq, void *data, struct pt_regs *regs);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,27)
extern void set_ethtool_ops(struct net_device *netdev);
#endif

#ifdef PEGNET_NIC
static void unm_init_pending_cmd_desc(unm_pending_cmd_descs_t *pending_cmds);
static void unm_destroy_pending_cmd_desc(unm_pending_cmd_descs_t *pending_cmds);
static void unm_proc_pending_cmd_desc(unm_adapter *adapter);
static inline int unm_get_pending_cmd_desc_cnt(unm_pending_cmd_descs_t *pending_cmds);

#endif


#if defined(XGB_DEBUG)
static void dump_skb(struct sk_buff *skb);
static int skb_is_sane(struct sk_buff *skb);
#endif

spinlock_t global_lock;

unsigned int            loopback_start = 0;

static struct           proc_dir_entry *unm_proc_dir_entry = NULL;

static unsigned char *tx_user_packet_data = NULL, *rx_user_packet_data;
static unsigned int tx_user_packet_length = 0, rx_datalen, rx_user_pos = 0;
static int capture_input = 0, captured = 0;

int physical_port[] = {0, 1, 2, 3};

#define DIDX_DIF(p, c)        \
   ((p >= c) ? (p - c) : (MAX_CMD_DESCRIPTORS_HOST - p + c))

#define CIS_WATERMARK         0x60

#define PEG_LOOP         1000  /* loop to pegstuck? */

#define QMAJ(hdr)        ((hdr >> 19) & 0xF)
#define QMIN(hdr)        (hdr & 0x3ffff)
#define QSUB(hdr)        ((hdr >> 18) & 0x1)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/*
 * In unm_nic_down(), we must wait for any pending callback requests into
 * unm_watchdog_task() to complete; eg otherwise the watchdog_timer could be
 * reenabled right after it is deleted in unm_nic_down(). FLUSH_SCHEDULED_WORK()
 * does this synchronization.
 *
 * Normally, schedule_work()/flush_scheduled_work() could have worked, but
 * unm_nic_close() is invoked with kernel rtnl lock held. netif_carrier_off()
 * call in unm_nic_close() triggers a schedule_work(&linkwatch_work), and a
 * subsequent call to flush_scheduled_work() in unm_nic_down() would cause
 * linkwatch_event() to be executed which also attempts to acquire the rtnl
 * lock thus causing a deadlock.
 */
#define	SCHEDULE_WORK(tp)	queue_work(unm_workq, tp)
#define	FLUSH_SCHEDULED_WORK()	flush_workqueue(unm_workq)
static struct workqueue_struct *unm_workq;
static void   unm_watchdog(unsigned long);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */

#if 0
/* VLAD - may be useful for future debugging */
void
unm_hexdump(unsigned char *buf, int len)
{
        int                         ii, jj;
        unsigned char         *cur_p = buf;

	printk("%d bytes @ 0x%X\n", len, buf);
	for (ii = 0; len > 0; ++ii) {
		if ((ii % 32) == 0)
			printk("        0  1  2  3  4  5  6  7  8  9"
				"  A  B  C  D  E  F\n");
		printk("0x%4.4x ", ii);
		for (jj = 0; (jj < 16) && (jj < len); ++jj)
			printk("%2.2x ", *(cur_p + jj));
		if (jj < 16)
			for (; jj < 16; ++ jj)
				printk("   ");
		printk("\n");
		cur_p += jj;
		len -= jj;
	}
}
/* VLAD */
#endif

static inline void unm_nic_update_cmd_producer(struct unm_adapter_s *adapter,
						uint32_t crb_producer) 
{
	switch (adapter->portnum) {
		case 0:
			UNM_NIC_PCI_WRITE_32(crb_producer, CRB_NORMALIZE
					(adapter, CRB_CMD_PRODUCER_OFFSET));
			return;
		case 1:
			UNM_NIC_PCI_WRITE_32(crb_producer, CRB_NORMALIZE
					(adapter, CRB_CMD_PRODUCER_OFFSET_1));
			return;
		case 2:
			UNM_NIC_PCI_WRITE_32(crb_producer, CRB_NORMALIZE
					(adapter, CRB_CMD_PRODUCER_OFFSET_2));
			return;
		case 3:
			UNM_NIC_PCI_WRITE_32(crb_producer, CRB_NORMALIZE
					(adapter, CRB_CMD_PRODUCER_OFFSET_3));
			return;
		default:
			printk("We tried to update CRB_CMD_PRODUCER_OFFSET for"
					"invalid PCI function id %d\n",
					adapter->portnum);
			return;
	}
}
	 
static inline void unm_nic_update_cmd_consumer(struct unm_adapter_s *adapter,
						uint32_t crb_consumer) 
{
	switch (adapter->portnum) {
		case 0:
			UNM_NIC_PCI_WRITE_32(crb_consumer, CRB_NORMALIZE
					(adapter, CRB_CMD_CONSUMER_OFFSET));
			return;
		case 1:
			UNM_NIC_PCI_WRITE_32(crb_consumer, CRB_NORMALIZE
					(adapter, CRB_CMD_CONSUMER_OFFSET_1));
			return;
		case 2:
			UNM_NIC_PCI_WRITE_32(crb_consumer, CRB_NORMALIZE
					(adapter, CRB_CMD_CONSUMER_OFFSET_2));
			return;
		case 3:
			UNM_NIC_PCI_WRITE_32(crb_consumer, CRB_NORMALIZE
					(adapter, CRB_CMD_CONSUMER_OFFSET_3));
			return;
		default:
			printk("We tried to update CRB_CMD_PRODUCER_OFFSET for"
					"invalid PCI function id %d\n",
					adapter->portnum);
			return;
	}
}

void __devinit
unm_check_options(unm_adapter *adapter)
{
        switch(adapter->ahw.boardcfg.board_type){
        case UNM_BRDTYPE_P2_SB31_10G:
        case UNM_BRDTYPE_P2_SB31_10G_CX4:
        case UNM_BRDTYPE_P2_SB31_10G_IMEZ:
        case UNM_BRDTYPE_P2_SB31_10G_HMEZ:
		adapter->MaxRxDescCount      = MAX_RCV_DESCRIPTORS;
                break;

        case UNM_BRDTYPE_P2_SB35_4G:
	case UNM_BRDTYPE_P2_SB31_2G:
		adapter->MaxRxDescCount      = MAX_RCV_DESCRIPTORS_1G;
                break;

        default:
		adapter->MaxRxDescCount      = MAX_RCV_DESCRIPTORS_1G;

                printk("%s: Unknown(%x)\n", unm_nic_driver_name,
			adapter->ahw.boardcfg.board_type);
                break;
        }

        adapter->MaxTxDescCount      = MAX_CMD_DESCRIPTORS_HOST;
	adapter->MaxJumboRxDescCount = MAX_JUMBO_RCV_DESCRIPTORS;
	adapter->MaxLroRxDescCount   = MAX_LRO_RCV_DESCRIPTORS;
        return;
}

static inline void unm_tx_csum(struct unm_adapter_s *adapter,
                               cmdDescType0_t *desc,
                               struct sk_buff *skb)
{
#ifdef UNM_NIC_HW_CSUM
        if (skb->ip_summed == CHECKSUM_HW) {
                if (skb->nh.iph->protocol == IPPROTO_TCP) {
                        desc->opcode = TX_TCP_PKT;
                } else if (skb->nh.iph->protocol == IPPROTO_UDP) {
                        desc->opcode = TX_UDP_PKT;
                } else {
                        return;
                }
                //adapter->stats.xmitcsummed++;
        }
        desc->tcpHdrOffset = skb->h.raw - skb->data;
        desc->ipHdrOffset  = skb->nh.raw - skb->data;
#endif /* UNM_NIC_HW_CSUM */
}
#ifdef UNM_NETIF_F_TSO
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
#define TSO_SIZE(x)   ((x)->gso_size)
#else
#define TSO_SIZE(x)   ((x)->tso_size)
#endif

static inline void unm_tso_check(struct unm_adapter_s *adapter,
                                 cmdDescType0_t *desc,
                                 struct sk_buff *skb)
{
#ifndef UNM_NIC_HW_CSUM
#error "Hardware checksum must be enabled while running with LSO."
#endif
        if (desc->mss) {
            desc->totalHdrLength = sizeof(struct ethhdr) +
                               ((skb->nh.iph)->ihl *sizeof(u32)) +
                               ((skb->h.th)->doff * sizeof(u32));
            desc->opcode = TX_TCP_LSO;
                //adapter->stats.xmitcsummed++;
        } else if (skb->ip_summed == CHECKSUM_HW) {
                if (skb->nh.iph->protocol == IPPROTO_TCP) {
                        desc->opcode = TX_TCP_PKT;
                } else if (skb->nh.iph->protocol == IPPROTO_UDP) {
                        desc->opcode = TX_UDP_PKT;
                } else {
                        return;
                }
                //adapter->stats.xmitcsummed++;
        }
        desc->tcpHdrOffset = skb->h.raw - skb->data;
        desc->ipHdrOffset  = skb->nh.raw - skb->data;
}
#endif /* UNM_NETIF_F_TSO */

/*  PCI Device ID Table  */
#define NETXEN_PCI_ID(device_id)   PCI_DEVICE(PCI_VENDOR_ID_NX, device_id)

static struct pci_device_id unm_pci_tbl[] __devinitdata = {
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_QG) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_XG) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_CX4) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_IMEZ) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_HMEZ) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_IMEZ_DUP) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_HMEZ_DUP) },
        {0,}
};


MODULE_DEVICE_TABLE(pci, unm_pci_tbl);

#define SUCCESS 0
static int
is_flash_supported(struct unm_adapter_s *adapter)
{
    int     locs[] = {0, 0x4, 0x100, 0x4000, 0x4128};
    int     addr, val01, val02, i, j;

    /* if the flash size is not 4Mb, make huge war cry and die */
    for (j = 1; j < 4; j++) {
        addr  = j * 0x100000;
        for (i = 0; i < (sizeof(locs)/sizeof(locs[0])); i++) {
            if(rom_fast_read(adapter, locs[i], &val01) == 0 &&
                           rom_fast_read(adapter, (addr+locs[i]),
                                         &val02) == 0) {
                                if (val01 == val02) {
                                        return -1;
                                }
                        }
                        else
                                return -1;
            }
    }

    return 0;
}

static int unm_get_flash_block(struct unm_adapter_s *adapter, int base, int size, uint32_t *buf)
{
        int        i, addr;
        uint32_t *ptr32;

        addr  = base;
        ptr32 = buf;
        for (i = 0; i < size / sizeof(uint32_t); i++) {
                if (rom_fast_read(adapter, addr, ptr32) == -1) {
                        return -1;
                }
                ptr32++;
                addr += sizeof(uint32_t);
        }
        if ((char *)buf + size > (char *)ptr32) {
                uint32_t local;

                if (rom_fast_read(adapter, addr, &local) == -1) {
                        return -1;
                }
                memcpy(ptr32, &local, (char *)buf+size-(char *)ptr32);
        }

        return 0;
}

static int get_flash_mac_addr(struct unm_adapter_s *adapter, uint64_t mac[])
{
        uint32_t *pmac = (uint32_t *)&mac[0];

        if (unm_get_flash_block(adapter,
                    USER_START + offsetof(unm_user_info_t, mac_addr),
                    FLASH_NUM_PORTS * sizeof(U64), pmac) == -1) {
                return -1;
        }
        if (*mac == ~0ULL) {
                if (unm_get_flash_block(adapter,
                     USER_START_OLD + offsetof(unm_old_user_info_t, mac_addr),
                     FLASH_NUM_PORTS * sizeof(U64), pmac) == -1) {
                        return -1;
                }
                if (*mac == ~0ULL) {
                        return -1;
                }
        }

        return 0;
}


#ifdef  PEGNET_NIC

/*
 * Initialize buffers required for the adapter in pegnet_nic case.
 */
static int initialize_adapter_offload(unm_adapter *adapter)
{
        uint64_t        addr;
        uint32_t        hi;
        uint32_t        lo;

        adapter->dummy_dma.addr =
                pci_alloc_consistent(adapter->ahw.pdev,
                                     UNM_HOST_DUMMY_DMA_SIZE,
                                     &adapter->dummy_dma.phys_addr);
        if(adapter->dummy_dma.addr == NULL) {
                printk("%s: ERROR: Could not allocate dummy DMA memory\n",
                       __FUNCTION__);
                return (-ENOMEM);
        }

        addr = (uint64_t)adapter->dummy_dma.phys_addr;
        hi = (addr >> 32) & 0xffffffff;
        lo = addr & 0xffffffff;

	read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(hi, CRB_NORMALIZE(adapter,
                                               CRB_HOST_DUMMY_BUF_ADDR_HI));
        UNM_NIC_PCI_WRITE_32(lo, CRB_NORMALIZE(adapter,
                                               CRB_HOST_DUMMY_BUF_ADDR_LO));
	read_unlock(&adapter->adapter_lock);

        return (0);
}

/*
 * Free buffers for the offload part from the adapter.
 */
static void free_adapter_offload(unm_adapter *adapter)
{
        if (adapter->dummy_dma.addr) {
                pci_free_consistent(adapter->ahw.pdev, UNM_HOST_DUMMY_DMA_SIZE,
                                    adapter->dummy_dma.addr,
                                    adapter->dummy_dma.phys_addr);
                adapter->dummy_dma.addr = NULL;
        }
}
#endif  /* PEGNET_NIC */

#ifdef BOUNCE_BUFFERS
/* Whenever we cross the 16K boundary of bounce buffer, we use the next
* 16K buffer and wrap up if its the last buffer.
*/
static int
unm_get_next_bounce_buffer(struct bounce_index *head, struct bounce_index *tail,
struct bounce_index *biret, unsigned len)
{
         struct bounce_index tmpbi;
 
         tmpbi.buffer_index = head->buffer_index;
         tmpbi.buffer_offset = head->buffer_offset;
 
         if((tmpbi.buffer_offset + len) > BOUNCE_BUFFER_SIZE){
                 if((tmpbi.buffer_index == tail->buffer_index) &&
                         (tmpbi.buffer_offset < tail->buffer_offset)) {
                         return -1;
                 }
                 tmpbi.buffer_index = (tmpbi.buffer_index + 1) & (NUM_BOUNCE - 1);
                 tmpbi.buffer_offset = 0;
         }
 
         if(tmpbi.buffer_index == tail->buffer_index &&
                 tmpbi.buffer_offset < tail->buffer_offset &&
                 (tmpbi.buffer_offset + len) >= tail->buffer_offset)
         {
                 return -1;
         }
         head->buffer_index = tmpbi.buffer_index;
         head->buffer_offset = tmpbi.buffer_offset + len;
         *biret = tmpbi;
         return 0;
 }
 
static void
unm_free_bounce_buffers(void)
{
         int i;
 
         for(i = 0; i < NUM_BOUNCE && tx_bounce_ptr[i]; i++) {
                 free_pages((unsigned long)tx_bounce_ptr[i], BOUNCE_BUFFER_ORDER);
                 tx_bounce_ptr[i] = NULL;
         }
 
         for(i = 0; i < NUM_BOUNCE && rx_bounce_ptr[i]; i++) {
                 free_pages((unsigned long)rx_bounce_ptr[i], BOUNCE_BUFFER_ORDER);
                 rx_bounce_ptr[i] = NULL;
         }
}
 
/* We have 4MB space reserved for bounce buffers.
 * The 4MB space is divided in 256 chunks of 16K buffers.
 */
static int
unm_alloc_bounce_buffers(void)
{
         int i;
 
         memset (tx_bounce_ptr, 0, sizeof(tx_bounce_ptr));
         memset (rx_bounce_ptr, 0, sizeof(rx_bounce_ptr));
 
         for(i = 0; i < NUM_BOUNCE; i++) {
                 tx_bounce_ptr[i] = (char *)__get_free_pages(GFP_KERNEL,
                                                         BOUNCE_BUFFER_ORDER);
                 if(!tx_bounce_ptr[i])
                         goto err_out;
                 if(virt_to_phys(tx_bounce_ptr[i])
                                + BOUNCE_BUFFER_SIZE > unm_dma_mask){
 
                         free_pages((unsigned long)tx_bounce_ptr[i], BOUNCE_BUFFER_ORDER);
                         tx_bounce_ptr[i] = (char *)__get_free_pages(GFP_DMA,
                                                         BOUNCE_BUFFER_ORDER);
                 }
                 if(!tx_bounce_ptr[i])
                         goto err_out;
 
         }
         tx_bounce_head.buffer_index = tx_bounce_tail.buffer_index = 0;
         tx_bounce_head.buffer_offset = tx_bounce_tail.buffer_offset = 0;
 
         for(i = 0; i<NUM_BOUNCE; i++) {
                 rx_bounce_ptr[i] = (char *)__get_free_pages(GFP_KERNEL, BOUNCE_BUFFER_ORDER);
                 if(!rx_bounce_ptr[i])
                         goto err_out;
                 if(virt_to_phys(rx_bounce_ptr[i])
                                + BOUNCE_BUFFER_SIZE > unm_dma_mask){
 
                         free_pages((unsigned long)rx_bounce_ptr[i], BOUNCE_BUFFER_ORDER);
                         rx_bounce_ptr[i] = (char *)__get_free_pages(GFP_DMA,
                                                         BOUNCE_BUFFER_ORDER);
                 }
                 if(!rx_bounce_ptr[i])
                         goto err_out;
 
         }
         rx_bounce_head.buffer_index = rx_bounce_tail.buffer_index = 0;
         rx_bounce_head.buffer_offset = rx_bounce_tail.buffer_offset = 0;
         return 0;
 
err_out:
         unm_free_bounce_buffers();
         return -ENOMEM;
 
}
 
#endif /* bounce buffers */
 
#define	ADAPTER_LIST_SIZE	12
int unm_cards_found;

static void reset_hw(struct unm_adapter_s *adapter)
{
        writel(0, CRB_NORMALIZE(adapter, CRB_CMDPEG_STATE));
        pinit_from_rom(adapter, 0);
        udelay(500);
        load_firmware(adapter);
}

static void cleanup_adapter(struct unm_adapter_s *adapter)
{
	struct pci_dev		*pdev;
	struct net_device	*netdev;
	int			i;

	if (!adapter) {
		return;
	}
	pdev = adapter->ahw.pdev;
	netdev = adapter->netdev;
	if (pdev == NULL || netdev == NULL) {
		return;
	}

	remove_proc_entry(netdev->name, unm_proc_dir_entry);
	unregister_netdev(netdev);

        for (i = 0; i < MAX_RCV_CTX; ++i) {
		int ring;
		unm_recv_context_t *recv_ctx;
		unm_rcv_desc_ctx_t *rcv_desc;

                recv_ctx = &adapter->recv_ctx[i];
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        rcv_desc = &recv_ctx->rcv_desc[ring];
                        if (rcv_desc->rx_buf_arr != NULL) {
                                pci_free_consistent(pdev, RCV_BUFFSIZE,
                                                       rcv_desc->rx_buf_arr,
                                                       rcv_desc->rx_buf_phys);
                                rcv_desc->rx_buf_arr = NULL;
                        }
                }
        }

#if defined(CONFIG_PCI_MSI)
	if ((adapter->flags & UNM_NIC_MSI_ENABLED)) {
		pci_disable_msi(pdev);
	}
#endif

        if (adapter->cmd_buf_arr != NULL) {
        	vfree (adapter->cmd_buf_arr);
		adapter->cmd_buf_arr = NULL;
	}

	if (adapter->ahw.pci_base0) {
		iounmap ((uint8_t *)adapter->ahw.pci_base0);
		adapter->ahw.pci_base0 = 0UL;
	}
	if (adapter->ahw.pci_base1) {
		iounmap ((uint8_t *)adapter->ahw.pci_base1);
		adapter->ahw.pci_base1 = 0UL;
	}
	if (adapter->ahw.pci_base2) {
		iounmap ((uint8_t *)adapter->ahw.pci_base2);
		adapter->ahw.pci_base2 = 0UL;
	}
	if (adapter->ahw.db_base) {
		iounmap ((uint8_t *)adapter->ahw.db_base);
		adapter->ahw.db_base = 0UL;
	}

        pci_release_regions(pdev);
        pci_disable_device(pdev);
        pci_set_drvdata(pdev, NULL);
#ifdef OLD_KERNEL
                kfree(netdev);
#else
                free_netdev(netdev);
#endif
}
int
check_hw_init(struct unm_adapter_s *adapter)
{
       u32 val=0;
       int ret=0;
       val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                                       UNM_CAM_RAM(0x1fc)));
       printk("NetXen: read 0x%08x for init reg.\n",val);
       if (val == 0x55555555) {
           /* This is the first boot after power up */
           val = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                                  UNM_ROMUSB_GLB_SW_RESET));
           printk("NetXen: read 0x%08x for reset reg.\n",val);
           if (val != 0x80000f) {
               ret=-1;
           }

           /* clear the register for future unloads/loads */
           UNM_NIC_PCI_WRITE_32(0, CRB_NORMALIZE(adapter,
                                               UNM_CAM_RAM(0x1fc)));

       }
       return ret;
}


/*
unm_nic_probe()

Linux system will invoke this after identifying the vendor ID and device Id
in the pci_tbl where this module will search for UNM vendor and device ID for
quad port adapter.

*/
static int __devinit
unm_nic_probe(struct pci_dev *pdev, struct pci_device_id *ent)
{
        struct net_device     *netdev  = NULL;
        struct unm_adapter_s  *adapter = NULL;
        uint8_t               *mem_ptr0 = NULL;
        uint8_t               *mem_ptr1 = NULL;
        uint8_t               *mem_ptr2 = NULL;
        uint8_t               *db_ptr  = NULL;
        unsigned long          mem_base, mem_len, db_base, db_len;
        int                    pci_using_dac, i = 0, err;
        int                    ring;
        unm_recv_context_t    *recv_ctx    = NULL;
        unm_rcv_desc_ctx_t    *rcv_desc    = NULL;
        struct unm_cmd_buffer *cmd_buf_arr = NULL;
        struct proc_dir_entry *procfile;
        uint64_t               mac_addr[FLASH_NUM_PORTS+1];
        static int	       valid_mac=0;
	int		       pci_func_id = PCI_FUNC(pdev->devfn);

        if (pdev->class != 0x020000) {
                DPRINTK(1, ERR, "NetXen function %d, class %x will not be enabled.\n",
                                   pci_func_id, pdev->class);
                return -ENODEV;
        }

        printk(KERN_INFO "%s \n", unm_nic_driver_string);

        if((err = pci_enable_device(pdev))) {
                DPRINTK(1, ERR, "Cannot enable PCI device, aborting. %d\n",
                                    err);
                unm_nic_probe_err = err;
                return err;
        }

        if (!(pci_resource_flags(pdev, 0) & IORESOURCE_MEM)) {
                DPRINTK(1, ERR, "Cannot find proper PCI device "
                                    "base address, aborting. %p\n", pdev);
                err = -ENODEV;
                goto err_out_disable_pdev;
        }

        if((err = pci_request_regions(pdev, unm_nic_driver_name))) {
                DPRINTK(1, ERR, "Cannot find proper PCI resources"
                                    " aborting:%d\n", err);
err_out_disable_pdev:
		pci_disable_device(pdev);
		pci_set_drvdata(pdev, NULL);
		unm_nic_probe_err = err;
		return err;
        }
        
        pci_set_master(pdev);

        pci_read_config_byte(pdev, PCI_REVISION_ID, &nx_p2_id);

#ifndef CONFIG_IA64
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        if((pci_set_dma_mask(pdev, PCI_DMA_64BIT) == 0) &&
            (pci_set_consistent_dma_mask(pdev, DMA_64BIT_MASK) == 0))
#else
        if(pci_set_dma_mask(pdev, PCI_DMA_64BIT) == 0)
#endif
	{
                pci_using_dac = 1;
        } else {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
                if((err = pci_set_dma_mask(pdev, PCI_DMA_32BIT)) ||
                   (err = pci_set_consistent_dma_mask(pdev, DMA_32BIT_MASK)))
#else
                if((err = pci_set_dma_mask(pdev, PCI_DMA_32BIT)))
#endif
		{
                        DPRINTK(1, ERR, "No usable DMA configuration,"
                                        " aborting:%d\n", err);
                        goto err_ret;
                }
                pci_using_dac = 0;
        }
	if (nx_p2_id == NX_P2_C1) {
		unm_dma_mask = DMA_35BIT_MASK;
	} else {
		unm_dma_mask = DMA_32BIT_MASK;
	}
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        if((err = pci_set_dma_mask(pdev, PCI_DMA_32BIT)) ||
           (err = pci_set_consistent_dma_mask(pdev, PCI_DMA_32BIT)))
#else
        if((err = pci_set_dma_mask(pdev, PCI_DMA_32BIT)))
#endif
	{
                DPRINTK(1, ERR, "No usable DMA configuration,"
                                        " aborting:%d\n", err);
                goto err_ret;
        }

        pci_using_dac = 0;
	unm_dma_mask = DMA_32BIT_MASK;
#endif          /*CONFIG_IA64 */

	DPRINTK(1, INFO, "pci_using_dac: %d\n", pci_using_dac);

	netdev = alloc_etherdev(sizeof(struct unm_adapter_s));

	if (!netdev) {
		printk(KERN_ERR "%s: Failed to allocate memory for the "
				"device block.Check system memory resource"
				" usage.\n", unm_nic_driver_name);
		err = -ENOMEM;
		goto err_ret;
	}

	SET_MODULE_OWNER(netdev);
	SET_NETDEV_DEV(netdev, &pdev->dev);

	adapter       = netdev->priv;
        memset(adapter, 0 , sizeof(struct unm_adapter_s));

	adapter->ahw.pdev      = pdev;
	adapter->ahw.pci_func  = pci_func_id;
	rwlock_init(&adapter->adapter_lock);
	spin_lock_init(&adapter->tx_lock);
	spin_lock_init(&adapter->lock);
        spin_lock_init(&adapter->tnic.lock);

	/* remap phys address */
	mem_base = pci_resource_start(pdev, 0);    /* 0 is for BAR 0 */
	mem_len  = pci_resource_len(pdev, 0);

	/* FSL 128 Meg of memory */
	DPRINTK(1, INFO, "ioremap from %lx a size of %lx\n", mem_base,
			mem_len);
	mem_ptr0 = ioremap(mem_base, FIRST_PAGE_GROUP_SIZE);
	mem_ptr1 = ioremap(mem_base + SECOND_PAGE_GROUP_START, SECOND_PAGE_GROUP_SIZE);
	mem_ptr2 = ioremap(mem_base + THIRD_PAGE_GROUP_START, THIRD_PAGE_GROUP_SIZE);

	DPRINTK(1, INFO, "ioremaped at 0 -> %p, 1 -> %p, 2 -> %p\n", mem_ptr0, mem_ptr1, mem_ptr2);

	db_base = pci_resource_start(pdev, 4);  /* doorbell is on bar 4 */
	db_len  = pci_resource_len(pdev, 4);


	DPRINTK(1, INFO, "doorbell ioremap from %lx a size of %lx\n", db_base,
			db_len);

	db_ptr = ioremap(db_base, UNM_DB_MAPSIZE_BYTES);
	adapter->ahw.pci_base0  = (unsigned long)mem_ptr0;
	adapter->ahw.pci_len0   = FIRST_PAGE_GROUP_SIZE;
	adapter->ahw.pci_base1  = (unsigned long)mem_ptr1;
	adapter->ahw.pci_len1   = SECOND_PAGE_GROUP_SIZE;
	adapter->ahw.pci_base2  = (unsigned long)mem_ptr2;
	adapter->ahw.pci_len2   = THIRD_PAGE_GROUP_SIZE;
	adapter->ahw.crb_base  = PCI_OFFSET_SECOND_RANGE(adapter, UNM_PCI_CRBSPACE);
	adapter->ahw.db_base   = (unsigned long)db_ptr;
	adapter->ahw.db_len    = db_len;
	if ((mem_ptr0 == 0UL) || (mem_ptr1 == 0UL) || (mem_ptr2 == 0UL)) {
		DPRINTK(1, ERR, "Cannot remap adapter memory aborting.:0 -> %p, 1 -> %p, 2 -> %p\n",
				mem_ptr0, mem_ptr1, mem_ptr2);
		err = -EIO;
		goto err_ret;
	}
	if (db_len == 0) {
		printk(KERN_ERR"%s: doorbell is disabled\n",
				unm_nic_driver_name);
		err = -EIO;
		goto err_ret;
	}
	if (db_ptr == 0UL) {
		printk(KERN_ERR"%s: Failed to allocate doorbell map.",
				unm_nic_driver_name);
		err = -EIO;
		goto err_ret;
	}
	DPRINTK(1, INFO, "doorbell ioremaped at %p\n", db_ptr);

	adapter->netdev  = netdev;
	adapter->pdev    = pdev;

        /* This will be reset for mezz cards  */
	adapter->portnum = pci_func_id;
	adapter->status   &= ~NETDEV_STATUS;

#if defined(UNM_NIC_HW_CSUM)
        adapter->rx_csum = 1;
#endif

	netdev->open               = unm_nic_open;
	netdev->stop               = unm_nic_close;
	netdev->hard_start_xmit    = unm_nic_xmit_frame;
	netdev->get_stats          = unm_nic_get_stats;
	netdev->set_multicast_list = unm_nic_set_multi;
	netdev->set_mac_address    = unm_nic_set_mac;
	netdev->change_mtu         = unm_nic_change_mtu;
	netdev->do_ioctl           = unm_nic_ioctl;
	netdev->tx_timeout         = unm_tx_timeout;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,27) || defined(__VMKERNEL_MODULE__)
	set_ethtool_ops(netdev);
	/* FIXME: maybe SET_ETHTOOL_OPS(netdev,&unm_nic_ethtool_ops);*/
#endif
#ifdef UNM_NIC_NAPI
	netdev->poll               = unm_nic_poll;
	netdev->weight             = UNM_NETDEV_WEIGHT;
#endif
#ifdef CONFIG_NET_POLL_CONTROLLER
	netdev->poll_controller    = unm_nic_poll_controller;
#endif
#ifdef UNM_NIC_HW_CSUM
	/* ScatterGather support */
	netdev->features           = NETIF_F_SG ;
	netdev->features          |= NETIF_F_IP_CSUM;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,19)
#ifdef UNM_NIC_HW_VLAN
	netdev->features          |= NETIF_F_HW_VLAN_TX;
	netdev->features          |= NETIF_F_HW_VLAN_RX;
#endif
#endif     //LINUX_KERNEL_VERSION >= 2.4.19
#ifdef UNM_NETIF_F_TSO
	netdev->features          |= NETIF_F_TSO;
#endif

	if(pci_using_dac)
		netdev->features  |= NETIF_F_HIGHDMA;

#if defined(CONFIG_PCI_MSI)
	adapter->flags |= UNM_NIC_MSI_ENABLED;
	if (pci_enable_msi(pdev)) {
		adapter->flags &= ~UNM_NIC_MSI_ENABLED;
		printk(KERN_WARNING "%s: unable to allocate MSI interrupt"
				" error\n", unm_nic_driver_name);
	}
#endif
	netdev->irq                = pdev->irq;
	NX_INIT_WORK(adapter->tx_timeout_task,
	    (void (*)(void *))unm_tx_timeout_task, netdev);

        /*
         * Set the CRB window to invalid. If any register in window 0 is
         * accessed it should set the window to 0 and then reset it to 1.
         */
        adapter->curr_window = 255;

        /*
         * Initialize the HW so that we can get the board type first. Based on
	 * the board type the ring size is chosen.
         */
        if (initialize_adapter_hw(adapter) != 0) {
            err = -EIO;
            goto err_ret;
        }

        unm_check_options(adapter);

        cmd_buf_arr = (struct unm_cmd_buffer*) vmalloc(TX_RINGSIZE);
        adapter->cmd_buf_arr = cmd_buf_arr;

        if (cmd_buf_arr == NULL) {
                printk(KERN_ERR "%s: Failed to allocate memory for the TX cmd "
		       "buffer. Check system memory resource usage.\n",
                        unm_nic_driver_name);
                err = -ENOMEM;
                goto err_ret;
        }
        memset (cmd_buf_arr, 0, TX_RINGSIZE);

        for (i = 0; i < MAX_RCV_CTX; ++i) {
                recv_ctx = &adapter->recv_ctx[i];
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        rcv_desc = &recv_ctx->rcv_desc[ring];
                        switch (RCV_DESC_TYPE(ring)) {
                        case RCV_RING_STD:
                                rcv_desc->MaxRxDescCount =
                                                adapter->MaxRxDescCount;
                                rcv_desc->flags = RCV_DESC_NORMAL;
                                rcv_desc->dma_size = RX_DMA_MAP_LEN;
                                rcv_desc->skb_size = MAX_RX_BUFFER_LENGTH;
                                break;

                        case RCV_RING_JUMBO:
                                rcv_desc->MaxRxDescCount =
                                                adapter->MaxJumboRxDescCount;
                                rcv_desc->flags = RCV_DESC_JUMBO;
                                rcv_desc->dma_size = RX_JUMBO_DMA_MAP_LEN;
                                rcv_desc->skb_size =
                                                MAX_RX_JUMBO_BUFFER_LENGTH;
                                break;

                        case RCV_RING_LRO:
                                rcv_desc->MaxRxDescCount =
                                                adapter->MaxLroRxDescCount;
                                rcv_desc->flags = RCV_DESC_LRO;
                                rcv_desc->dma_size = RX_LRO_DMA_MAP_LEN;
                                rcv_desc->skb_size =
                                                MAX_RX_LRO_BUFFER_LENGTH;
                                break;

                        default:
                                printk(KERN_ERR"%s: unm_nic_probe: bad receive "
                                       "descriptor type %d\n",
                                        unm_nic_driver_name, RCV_DESC_TYPE(ring));
                                break;
                        }
                        rcv_desc->rx_buf_arr = (struct unm_rx_buffer*)
                                                    pci_alloc_consistent(pdev,
                                                    RCV_BUFFSIZE,
                                                    &rcv_desc->rx_buf_phys);

                        if(rcv_desc->rx_buf_arr == NULL) {
                              printk(KERN_ERR"%s: Failed to allocate memory for"
                                       " the device block.Check system memory "
                                       "resource usage.\n",unm_nic_driver_name);
                                err = -ENOMEM;
                                goto err_ret;
                        }
                        memset (rcv_desc->rx_buf_arr, 0, RCV_BUFFSIZE);
                }

        }

        initialize_adapter_sw(adapter); /* initialize the buffers in adapter */

	/* Mezz cards have PCI function 0,2,3 enabled */
	if (adapter->ahw.boardcfg.board_type == UNM_BRDTYPE_P2_SB31_10G_IMEZ) {
		if (pci_func_id >= 2) {
                        /* also swap the port numbers */
			adapter->portnum = pci_func_id - 2;
		}
	}

	if(adapter->portnum == 0) {
#ifdef CONFIG_IA64
        	pinit_from_rom(adapter, 0);
	        udelay(500);
        	load_firmware(adapter);
	/*
	 * Done later in the function now.
         reset_hw(adapter);
	 */
#endif
	}

        init_timer(&adapter->watchdog_timer);
        adapter->ahw.xg_linkup = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        adapter->watchdog_timer.function = &unm_watchdog;
#else
        adapter->watchdog_timer.function = &unm_watchdog_task;
#endif
        adapter->watchdog_timer.data = (unsigned long) adapter;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        NX_INIT_WORK(&adapter->watchdog_task,
                 (void (*)(void *))unm_watchdog_task, adapter);
#endif
        adapter->ahw.vendor_id = pdev->vendor;
        adapter->ahw.device_id = pdev->device;
        adapter->procCmdBufCounter = 0;
        pci_read_config_byte(pdev, PCI_REVISION_ID, &adapter->ahw.revision_id);
        pci_read_config_word(pdev, PCI_COMMAND, &adapter->ahw.pci_cmd_word);

	/* make sure Window == 1 */
        unm_nic_pci_change_crbwindow(adapter,1);

        /*
         * Initialize all the CRB registers here.
         */
        /* Window = 1 */
	unm_nic_update_cmd_producer(adapter, 0);
	unm_nic_update_cmd_consumer(adapter, 0);
        UNM_NIC_PCI_WRITE_32(0, CRB_NORMALIZE(adapter, CRB_HOST_CMD_ADDR_LO));

	if (is_flash_supported(adapter) == 0 &&
			get_flash_mac_addr(adapter, mac_addr) == 0) {
		valid_mac = 1;
	} else {
		valid_mac = 0;
	}

	if (valid_mac) {
		unsigned char *p = (unsigned char *)&mac_addr[adapter->portnum];
		adapter->mac_addr[0] = *(p+5);
		adapter->mac_addr[1] = *(p+4);
		adapter->mac_addr[2] = *(p+3);
		adapter->mac_addr[3] = *(p+2);
		adapter->mac_addr[4] = *(p+1);
		adapter->mac_addr[5] = *(p+0);

		if(!is_valid_ether_addr(adapter->mac_addr)) {
			printk(KERN_ERR "%s: Bad MAC address %02x:%02x:%02x:%02x:%02x:%02x.\n",
					unm_nic_driver_name,
					adapter->mac_addr[0],
					adapter->mac_addr[1],
					adapter->mac_addr[2],
					adapter->mac_addr[3],
					adapter->mac_addr[4],
					adapter->mac_addr[5]);
		} else {
			memcpy(netdev->dev_addr, adapter->mac_addr, netdev->addr_len);
			if (unm_nic_macaddr_set (adapter, adapter->mac_addr)!= 0) {
                            err = -EIO;
                            goto err_ret;
                        }
		}
	}
#ifdef PEGNET_NIC
	/*
	 * Initialize the pegnet_cmd_desc overflow data structure.
	 */
#ifndef __VMKERNEL_MODULE__
	unm_init_pending_cmd_desc(&adapter->pending_cmds);
#endif
	/* fill the adapter id field with the board serial num */
	unm_nic_get_serial_num(adapter);
#endif  /* PEGNET_NIC */


	if(adapter->portnum == 0) {
#ifdef  PEGNET_NIC
        	/* do this before waking up pegs so that we have valid dummy dma addr*/
	        err = initialize_adapter_offload(adapter);
        	if (err) {
                	goto err_ret;
	        }
#endif  /* PEGNET_NIC */

		/* Check here to make sure that pegtune finished correctly */
		if (check_hw_init(adapter)!= 0) {
			printk("ERROR in NetXen HW init sequence.\n");
			err = -ENODEV;
			goto err_ret;
		}
		printk("State: 0x%0x\n",
		       UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
							 CRB_CMDPEG_STATE)));

		/*
		 * Tell the hardware our version number.
		 */
		i = (_UNM_NIC_LINUX_MAJOR << 16) | ((_UNM_NIC_LINUX_MINOR << 8)) |
					(_UNM_NIC_LINUX_SUBVERSION);
		UNM_NIC_PCI_WRITE_32(i, CRB_NORMALIZE(adapter, CRB_DRIVER_VERSION));
		/* Unlock the HW, prompting the boot sequence */
		UNM_NIC_PCI_WRITE_32(1, CRB_NORMALIZE(adapter,
						      UNM_ROMUSB_GLB_PEGTUNE_DONE));

        	/* Handshake with the card before we register the devices. */
	        phantom_init(adapter, 0);

#if 1
		/* We seem to have a bug in the initial card bootup sequence.
		 * Reloading the firmware fixes it. */

		reset_hw(adapter);

		phantom_init(adapter, 0);
#endif

	}

	/*
	 * See if the firmware gave us a virtual-physical port mapping.
	 */
	i = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter, CRB_V2P(adapter->portnum)));
	if (i != 0x55555555) {
		physical_port[adapter->portnum] = i;
	}
        unm_nic_erase_pxe(adapter);

	netif_carrier_off(netdev);
	netif_stop_queue(netdev);

	if((err = register_netdev(netdev))) {
 	        DPRINTK(1, ERR, "register_netdev failed port %d\n",
                                 pci_func_id);
	}

#ifdef __VMKERNEL_MODULE__
	if (err = vmk_net_register_dev(netdev, pdev, TRUE) != 0){
		printk("Registering %s with vmkernel failed\n",netdev->name);
	}
#endif
	procfile = create_proc_entry(netdev->name, S_IRUGO,
			unm_proc_dir_entry);
	if (procfile) {
		procfile->data = netdev;
		procfile->owner = THIS_MODULE;
		procfile->read_proc = unm_nic_port_read_proc;
	}

        pci_set_drvdata(pdev, netdev);

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
                printk("%s: QUAD GbE board initialized\n",unm_nic_driver_name);
                break;

        case UNM_NIC_XGBE:
                printk("%s: XGbE board initialized\n",unm_nic_driver_name);
                break;
        }
	adapter->driver_mismatch = 0;
        unm_nic_probe_err = 0;
        return 0;

err_ret:
	cleanup_adapter(adapter);
        unm_nic_probe_err = err;
        return err;
}

static void
initialize_adapter_sw (struct unm_adapter_s    *adapter)
{
        int                ctx, ring;
        uint32_t           i;
        uint32_t           num_rx_bufs = 0;
        unm_rcv_desc_ctx_t *rcv_desc;

        adapter->freeCmdCount = adapter->MaxTxDescCount;
        DPRINTK(1, INFO, "initializing some queues: %p\n", adapter);
        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        struct unm_rx_buffer *rxBuf;
                        rcv_desc    = &adapter->recv_ctx[ctx].rcv_desc[ring];
                        rcv_desc->rcv_free = rcv_desc->MaxRxDescCount;
                        /* Initialize the queues for both receive and
                           command buffers */
                        rxBuf       = rcv_desc->rx_buf_arr;
                        num_rx_bufs = rcv_desc->MaxRxDescCount;
                        rcv_desc->begin_alloc = 0;
                        /* Initialize the free queues */
                        TAILQ_INIT (&rcv_desc->free_rxbuf_list);
                        /*
                         * Now go through all of them, set reference handles
                         * and put them in the queues.
                         */
                        for (i=0; i < num_rx_bufs; i ++) {
                                rxBuf->refHandle = i;
                                rxBuf->state = UNM_BUFFER_FREE;
                                TAILQ_INSERT_TAIL(&rcv_desc->free_rxbuf_list,
                                                  rxBuf, link);

                                DPRINTK(1, INFO, "Rx buf:ctx%d i(%d) rxBuf:"
                                                    "%p\n", ctx, i,rxBuf);
                                rxBuf ++;
                        }
                }
        }
        DPRINTK(1, INFO, "initialized buffers for %s and %s\n",
                            "adapter->free_cmd_buf_list",
                            "adapter->free_rxbuf");
        return;
}

static int
initialize_adapter_hw(struct unm_adapter_s *adapter)
{
        uint32_t value = 0;
        unm_board_info_t *board_info = &(adapter->ahw.boardcfg);
        int      ports = 0;

        if (unm_nic_get_board_info(adapter) != 0) {
            printk("%s: Error getting board config info.\n",
                    unm_nic_driver_name);
            return -1;
        }

        GET_BRD_PORTS_BY_TYPE(board_info->board_type, ports);
        if (ports == 0) {
                printk(KERN_ERR "%s: Unknown board type\n",
		       unm_nic_driver_name);
	}
        adapter->ahw.max_ports = ports;

        return value;
}

static int
init_firmware (struct unm_adapter_s *adapter)
{
        uint32_t   state = 0, loops = 0, err = 0;

        /* Window 1 call */
        read_lock(&adapter->adapter_lock);
        state = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter, CRB_CMDPEG_STATE));
        read_unlock(&adapter->adapter_lock);

        if (state == PHAN_INITIALIZE_ACK)
                return 0;

        while (state != PHAN_INITIALIZE_COMPLETE && loops < 200000) {
                udelay(100);
                schedule();
                /* Window 1 call */
                read_lock(&adapter->adapter_lock);
                state = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                                        CRB_CMDPEG_STATE));
                read_unlock(&adapter->adapter_lock);

                loops++;
        }
        if (loops >= 200000) {
                printk(KERN_ERR "Cmd Peg initialization not complete:%x.\n",
                                state);
                err = -EIO;
                return err;
        }
        /* Window 1 call */
        read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(MPORT_MULTI_FUNCTION_MODE,
                             CRB_NORMALIZE(adapter, CRB_MPORT_MODE));
        UNM_NIC_PCI_WRITE_32(PHAN_INITIALIZE_ACK,
                             CRB_NORMALIZE(adapter, CRB_CMDPEG_STATE));
        read_unlock(&adapter->adapter_lock);

        return err;
}

static void
unm_free_hw_resources(struct unm_adapter_s *adapter)
{
        unm_recv_context_t *recv_ctx;
        unm_rcv_desc_ctx_t *rcv_desc;
        int                 ctx, ring;

        if (adapter->ctxDesc != NULL) {
                pci_free_consistent(adapter->ctxDesc_pdev,
                                    sizeof(RingContext) +
                                    sizeof(uint32_t),
                                    adapter->ctxDesc,
                                    adapter->ctxDesc_physAddr);
                adapter->ctxDesc = NULL;
        }

        if (adapter->ahw.cmdDescHead != NULL) {
                pci_free_consistent(adapter->ahw.cmdDesc_pdev,
                                    sizeof(cmdDescType0_t) *
                                    adapter->MaxTxDescCount,
                                    adapter->ahw.cmdDescHead,
                                    adapter->ahw.cmdDesc_physAddr);
                adapter->ahw.cmdDescHead = NULL;
        }
        /* Special handling: there are 2 ports on this board */
        if (adapter->ahw.boardcfg.board_type == UNM_BRDTYPE_P2_SB31_10G_IMEZ) {
                adapter->ahw.max_ports = 2;
        }

#ifdef EPG_WORKAROUND
        if (adapter->ahw.pauseAddr != NULL) {
                pci_free_consistent(adapter->ahw.pause_pdev,512,
                                    adapter->ahw.pauseAddr,
                                    adapter->ahw.pause_physAddr);
                adapter->ahw.pauseAddr = NULL;
        }
#endif

        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                recv_ctx = &adapter->recv_ctx[ctx];
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        rcv_desc = &recv_ctx->rcv_desc[ring];

                        if (rcv_desc->desc_head != NULL) {
                                pci_free_consistent(rcv_desc->phys_pdev,
                                                    RCV_DESC_RINGSIZE,
                                                    rcv_desc->desc_head,
                                                    rcv_desc->phys_addr);
                                rcv_desc->desc_head = NULL;
                        }
                }

                if (recv_ctx->rcvStatusDescHead != NULL) {
                        pci_free_consistent(recv_ctx->rcvStatusDesc_pdev,
                                            STATUS_DESC_RINGSIZE,
                                            recv_ctx->rcvStatusDescHead,
                                            recv_ctx->rcvStatusDesc_physAddr);
                        recv_ctx->rcvStatusDescHead = NULL;
                }
        }

        return;
}

static void __devexit
unm_nic_remove(struct pci_dev *pdev)
{
        struct unm_adapter_s *adapter;
        struct net_device    *netdev;
        struct unm_rx_buffer *buffer;
        unm_recv_context_t   *recv_ctx;
        unm_rcv_desc_ctx_t   *rcv_desc;
        int                   i;
        int                   ctx, ring;

        netdev = pci_get_drvdata(pdev);
        adapter = netdev_priv(netdev);
        if (adapter == NULL) {
                return;
        }

	unm_nic_stop_port(adapter);
#ifndef SINGLE_DMA_BUF
        if ((adapter->flags & UNM_NIC_MSI_ENABLED)) {
		read_lock(&adapter->adapter_lock);
                unm_nic_disable_int(adapter);
		read_unlock(&adapter->adapter_lock);
        }
#endif

#ifdef  PEGNET_NIC
#ifndef __VMKERNEL_MODULE__
	unm_destroy_pending_cmd_desc(&adapter->pending_cmds);
#endif  /* ! __VMKERNEL_MODULE__ */

	if (adapter->portnum == 0) {
		free_adapter_offload(adapter);
	}
#endif  /* PEGNET_NIC */

#ifndef SINGLE_DMA_BUF
	if(adapter->is_up == ADAPTER_UP_MAGIC)
		free_irq(netdev->irq, adapter);
#endif
        unm_free_hw_resources(adapter);

        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                recv_ctx = &adapter->recv_ctx[ctx];
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        rcv_desc = &recv_ctx->rcv_desc[ring];
                        for (i = 0; i < rcv_desc->MaxRxDescCount; ++i) {
                                buffer = &(rcv_desc->rx_buf_arr[i]);
                                if (buffer->state == UNM_BUFFER_FREE) {
                                        continue;
                                }
                                pci_unmap_single(pdev, buffer->dma,
                                                    rcv_desc->dma_size,
                                                    PCI_DMA_FROMDEVICE);
                                if (buffer->skb != NULL) {
                                        dev_kfree_skb_any(buffer->skb);
                                }
                        }
                }
        }

	cleanup_adapter(adapter);
}

/*
 * Called when a network interface is made active
 * Returns 0 on success, negative value on failure
 */
static int
unm_nic_open(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;
        int                   err = 0;
        int                   ctx, ring;
        unsigned char         mac_addr[MAX_ADDR_LEN];
        int                   i;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
        if (!try_module_get(THIS_MODULE))
                return -ENODEV;
#endif

        mod_timer(&adapter->watchdog_timer, jiffies);

	if (adapter->is_up != ADAPTER_UP_MAGIC) {
		err = init_firmware (adapter);
		if (err != 0) {
			printk(KERN_ERR "Failed to init firmware\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			module_put(THIS_MODULE);
#endif
			return -EIO;
		}
		unm_nic_flash_print(adapter);

		/* setup all the resources for the Phantom... */
		/* this include the descriptors for rcv, tx, and status */
		unm_nic_clear_stats(adapter);
		err = unm_nic_hw_resources (adapter);
		if (err) {
			DPRINTK(1, ERR, "Error in setting hw resources:"
					"%d\n", err);
			return err;
		}
		for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
			for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
				unm_post_rx_buffers(adapter, ctx, ring);
			}
		}

#ifndef SINGLE_DMA_BUF
		err = request_irq(netdev->irq, &unm_intr,
				SA_SHIRQ | SA_SAMPLE_RANDOM, netdev->name,
				adapter);
		if (err) {
			printk(KERN_ERR"%s: %s Unable to regirster the interrupt"
					" service routine\n", unm_nic_driver_name,
					netdev->name);
			unm_free_hw_resources(adapter);
			return (err);
		}
		//              tasklet_enable(&adapter->tx_tasklet);
#endif
		adapter->is_up = ADAPTER_UP_MAGIC;
		//                update_core_clock(port);
	}
#ifndef SINGLE_DMA_BUF
	read_lock(&adapter->adapter_lock);
	unm_nic_enable_int(adapter);
	read_unlock(&adapter->adapter_lock);
#endif
        spin_lock_init(&adapter->stats_lock);

        /* Done here again so that even if phantom sw overwrote it, we set it 
         * Flaky Mac address registers on quadgig may require several writes.
        */
        if (adapter->ahw.board_type == UNM_NIC_GBE) {
                unsigned char *p = mac_addr;
                for (i = 0; i < 10; i++) {
                        unm_nic_macaddr_set (adapter, adapter->mac_addr);
                        unm_niu_macaddr_get(adapter, physical_port[adapter->portnum], p);
                        if (memcmp(mac_addr, adapter->mac_addr, 6) == 0)
                                break;
                }

                if (i == 10) {
                        printk(KERN_ERR "%s: cannot set Mac addr for %s\n",
                               unm_nic_driver_name, netdev->name);

                        printk(KERN_ERR "MAC address set: "
                               "%02x:%02x:%02x:%02x:%02x:%02x.\n",
                               adapter->mac_addr[0],
                               adapter->mac_addr[1],
                               adapter->mac_addr[2],
                               adapter->mac_addr[3],
                               adapter->mac_addr[4],
                               adapter->mac_addr[5]);

                        printk(KERN_ERR "MAC address get: "
                               "%02x:%02x:%02x:%02x:%02x:%02x.\n",
                               mac_addr[0],
                               mac_addr[1],
                               mac_addr[2],
                               mac_addr[3],
                               mac_addr[4],
                               mac_addr[5]);
            }
        } else {
                if (unm_nic_macaddr_set (adapter, adapter->mac_addr)!=0) {
                    return -EIO;
                }
        }
        if (unm_nic_init_port (adapter) != 0) {
                printk(KERN_ERR "%s: Failed to initialize the port %d\n",
                                unm_nic_driver_name, adapter->portnum);
                unm_nic_down(adapter);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
                module_put(THIS_MODULE);
#endif
                return -EIO;
        }
        unm_nic_set_link_parameters(adapter);

        unm_nic_set_multi(netdev);
        if(!adapter->driver_mismatch) {
               netif_start_queue(netdev);
        }

        adapter->state = PORT_UP;
        return 0;
}

static void *nx_alloc(struct pci_dev *pdev,
                      size_t sz, dma_addr_t *ptr,
                      struct pci_dev **used_dev)
{
        void *addr;

        addr = pci_alloc_consistent(pdev, sz, ptr);
        if ((unsigned long long)((*ptr) + sz) < unm_dma_mask) {
                *used_dev = pdev;
                return addr;
        }
        pci_free_consistent(pdev, sz, addr, *ptr);
        addr = pci_alloc_consistent(NULL, sz, ptr);
        *used_dev = NULL;
        return addr;
}

/*
 * check if the firmware has been downloaded and ready to run  and
 * setup the address for the descriptors in the adapter
 */
static int
unm_nic_hw_resources (struct unm_adapter_s *adapter)
{
        struct _hardware_context *hw=&adapter->ahw;
        uint32_t                  state = 0;
        void                     *addr;
#ifdef EPG_WORKAROUND
        void                     *pause_addr;
#endif
        int                       loops = 0, err = 0;
        int                       ctx, ring;
	int			  func_id = adapter->portnum;
        unm_recv_context_t       *recv_ctx;
        unm_rcv_desc_ctx_t       *rcv_desc;

        DPRINTK(1, INFO, "crb_base: %lx %lx", UNM_PCI_CRBSPACE,
		PCI_OFFSET_SECOND_RANGE(adapter, UNM_PCI_CRBSPACE));
        DPRINTK(1, INFO, "cam base: %lx %lx", UNM_CRB_CAM,
		pci_base_offset(adapter, UNM_CRB_CAM));
        DPRINTK(1, INFO, "cam RAM: %lx %lx", UNM_CAM_RAM_BASE,
		pci_base_offset(adapter, UNM_CAM_RAM_BASE));
        DPRINTK(1, INFO, "NIC base:%lx %lx\n",NIC_CRB_BASE,
		pci_base_offset(adapter, NIC_CRB_BASE));

        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                DPRINTK(1, INFO, "Command Peg ready..waiting for rcv peg\n");
                loops = 0;
                state = 0;
                /* Window 1 call */
                read_lock(&adapter->adapter_lock);
                state = UNM_NIC_PCI_READ_32(
                                CRB_NORMALIZE(adapter,
                                        recv_crb_registers[ctx].CRB_RCVPEG_STATE));
                read_unlock(&adapter->adapter_lock);


                while (state != PHAN_PEG_RCV_INITIALIZED && loops < 20) {
                        udelay(100);
                        schedule();
                        /* Window 1 call */

                        read_lock(&adapter->adapter_lock);
                        state = UNM_NIC_PCI_READ_32(
                                        CRB_NORMALIZE(adapter,
                                        recv_crb_registers[ctx].CRB_RCVPEG_STATE));
                        read_unlock(&adapter->adapter_lock);

                        loops++;
                }
                if (loops >= 20) {
                        DPRINTK(1, ERR, "Rcv Peg initialization not complete:"
                                        "%x.\n", state);
                        err = -EIO;
                        return err;
                }
        }
        DPRINTK(1, INFO, "Recieve Peg ready too. starting stuff\n");


        addr = nx_alloc(adapter->ahw.pdev,
                                sizeof(RingContext) + sizeof(uint32_t),
                                (dma_addr_t *)&adapter->ctxDesc_physAddr,
                                &adapter->ctxDesc_pdev);

        printk("ctxDesc_physAddr: 0x%llx\n", (U64)adapter->ctxDesc_physAddr);

        if (addr == NULL) {
                DPRINTK(1,ERR, "bad return from "
                                "pci_alloc_consistent\n");
                err = -ENOMEM;
                return err;
        }
        memset(addr, 0, sizeof(RingContext));
        adapter->ctxDesc = (RingContext *) addr;
        adapter->ctxDesc->CtxId = adapter->portnum;
        adapter->ctxDesc->CMD_CONSUMER_OFFSET = adapter->ctxDesc_physAddr
                                                + sizeof(RingContext);
        adapter->cmdConsumer = (uint32_t *)(((char *) addr) + sizeof(RingContext));

        addr = nx_alloc(adapter->ahw.pdev,
                        sizeof(cmdDescType0_t)* adapter->MaxTxDescCount,
                        (dma_addr_t *)&hw->cmdDesc_physAddr,
                        &adapter->ahw.cmdDesc_pdev);

        printk("cmdDesc_physAddr: 0x%llx\n", (U64)hw->cmdDesc_physAddr);

        if (addr == NULL) {
                DPRINTK(1,ERR, "bad return from pci_alloc_consistent\n");
                unm_free_hw_resources(adapter);
                err = -ENOMEM;
                return err;
        }

        adapter->ctxDesc->CmdRingAddrLo = hw->cmdDesc_physAddr & 0xffffffffUL;
        adapter->ctxDesc->CmdRingAddrHi = ((U64)hw->cmdDesc_physAddr >> 32);
        adapter->ctxDesc->CmdRingSize = adapter->MaxTxDescCount;

#ifdef EPG_WORKAROUND
        pause_addr = nx_alloc(adapter->ahw.pdev,512,
                                    (dma_addr_t *)&hw->pause_physAddr, &hw->pause_pdev);

        if (pause_addr == NULL) {
                DPRINTK(1,ERR, "bad return from pci_alloc_consistent\n");
                err = -ENOMEM;
                return err;
        }
        hw->pauseAddr = (char *)pause_addr;
        {
                uint64_t *ptr = (uint64_t *)pause_addr;
                *ptr++ = 0ULL;
                *ptr++ = 0ULL;
                *ptr++ = 0x200ULL;
                *ptr++ = 0x0ULL;
                *ptr++ = 0x2200010000c28001ULL;
                *ptr++ = 0x0100088866554433ULL;
        }
#endif

        hw->cmdDescHead = (cmdDescType0_t *)addr;

        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                recv_ctx = &adapter->recv_ctx[ctx];

                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        rcv_desc = &recv_ctx->rcv_desc[ring];
                        addr = nx_alloc(adapter->ahw.pdev,
                                                        RCV_DESC_RINGSIZE,
                                                        (dma_addr_t *)
                                                        &rcv_desc->phys_addr,
							&rcv_desc->phys_pdev);
                        printk("rcv %d physAddr: 0x%llx\n", ring,
                                (U64)rcv_desc->phys_addr);

                        if (addr == NULL) {
                                DPRINTK(1,ERR, "bad return from "
                                                "pci_alloc_consistent\n");
                                unm_free_hw_resources(adapter);
                                err = -ENOMEM;
                                return err;
                        }
                        rcv_desc->desc_head = (rcvDesc_t *) addr;

                        adapter->ctxDesc->RcvContext[ring].RcvRingAddrLo =
                                   rcv_desc->phys_addr & 0xffffffffUL;
                        adapter->ctxDesc->RcvContext[ring].RcvRingAddrHi =
                                   ((U64)rcv_desc->phys_addr>>32);
                        adapter->ctxDesc->RcvContext[ring].RcvRingSize =
                                rcv_desc->MaxRxDescCount;
                }

                addr = nx_alloc(adapter->ahw.pdev,
                                           STATUS_DESC_RINGSIZE,
                                           (dma_addr_t *)
                                           &recv_ctx->rcvStatusDesc_physAddr,
					   &recv_ctx->rcvStatusDesc_pdev);
                printk("rcvStatusDesc_physAddr: 0x%llx\n",
                        (U64)recv_ctx->rcvStatusDesc_physAddr);

                if (addr == NULL) {
                        DPRINTK(1,ERR, "bad return from"
                                       " pci_alloc_consistent\n");
                        unm_free_hw_resources(adapter);
                        err = -ENOMEM;
                        return err;
                }
                recv_ctx->rcvStatusDescHead = (statusDesc_t *) addr;

                adapter->ctxDesc->StsRingAddrLo =
                         recv_ctx->rcvStatusDesc_physAddr & 0xffffffffUL;
                adapter->ctxDesc->StsRingAddrHi =
                         ((U64)recv_ctx->rcvStatusDesc_physAddr>>32);
                adapter->ctxDesc->StsRingSize = adapter->MaxRxDescCount;


        }
        /* Window = 1 */

	read_lock(&adapter->adapter_lock);
#ifdef EPG_WORKAROUND
        /* write this before the host cmd ring addr */
        UNM_NIC_PCI_WRITE_32((hw->pause_physAddr & 0xffffffffUL),
                                CRB_NORMALIZE(adapter, CRB_PAUSE_ADDR_LO));
        UNM_NIC_PCI_WRITE_32(((U64)hw->pause_physAddr>>32),
                                CRB_NORMALIZE(adapter, CRB_PAUSE_ADDR_HI));
#endif
        UNM_NIC_PCI_WRITE_32(adapter->ctxDesc_physAddr,
                                CRB_NORMALIZE(adapter, CRB_CTX_ADDR_REG_LO(func_id)));
        UNM_NIC_PCI_WRITE_32(((U64)adapter->ctxDesc_physAddr>>32),
                                CRB_NORMALIZE(adapter, CRB_CTX_ADDR_REG_HI(func_id)));
        UNM_NIC_PCI_WRITE_32(UNM_CTX_SIGNATURE | func_id,
                                CRB_NORMALIZE(adapter, CRB_CTX_SIGNATURE_REG(func_id)));
	read_unlock(&adapter->adapter_lock);
        return err;
}

/*
 * This will be called when all the ports of the adapter are removed.
 * This will cleanup and disable interrupts and irq.
 */
static void
unm_nic_down(struct unm_adapter_s *adapter)
{
        int i,j;
        struct unm_cmd_buffer *cmd_buff;
        struct unm_skb_frag    *buffrag;

        adapter->state = PORT_DOWN;
        /* should we disable to phy for the port    */

#if 0
		spin_lock_bh(&adapter->tnic.lock);
		if (!adapter->tnic.registered) {
#ifndef SINGLE_DMA_BUF
			read_lock(&adapter->adapter_lock);
			unm_nic_disable_int(adapter);
			read_unlock(&adapter->adapter_lock);
#endif
		}
		spin_unlock_bh(&adapter->tnic.lock);
#endif

                cmd_buff = adapter->cmd_buf_arr;
                for (i = 0; i < adapter->MaxTxDescCount; i++) {
                        buffrag = cmd_buff->fragArray;
                        if (buffrag->dma) {
                                pci_unmap_single(adapter->pdev, buffrag->dma,
                                        buffrag->length, PCI_DMA_TODEVICE);
                                buffrag->dma = (uint64_t)NULL;
                        }
                        for (j =0; j < cmd_buff->fragCount; j++) {
                                buffrag++;
                                if (buffrag->dma) {
                                        pci_unmap_page(adapter->pdev,
                                                        buffrag->dma,
                                                        buffrag->length,
                                                        PCI_DMA_TODEVICE);
                                        buffrag->dma = (uint64_t)NULL;
                                }
                        }
                        /* Free the skb we received in unm_nic_xmit_frame */
                        if(cmd_buff->skb) {
                                dev_kfree_skb_any(cmd_buff->skb);
                                cmd_buff->skb = NULL;
                        }
                        cmd_buff++;
                }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	FLUSH_SCHEDULED_WORK();
#endif
        del_timer_sync(&adapter->watchdog_timer);
}

/**
 * unm_nic_close - Disables a network interface entry point
 **/
static int
unm_nic_close(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);

        netif_carrier_off(netdev);
        netif_stop_queue(netdev);

        unm_nic_down(adapter);
        /*del_timer(&unm_nic_timer);*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
        module_put(THIS_MODULE);
#endif

        return 0;
}

static int
unm_nic_set_mac(struct net_device *netdev, void *p)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        struct sockaddr      *addr    = p;

        if (netif_running(netdev))
                return -EBUSY;

        if (!is_valid_ether_addr(addr->sa_data))
                return -EADDRNOTAVAIL;

        DPRINTK(1, INFO, "valid ether addr\n");
        memcpy(netdev->dev_addr,  addr->sa_data, netdev->addr_len);
        memcpy(adapter->mac_addr, addr->sa_data, netdev->addr_len);

        return (unm_nic_macaddr_set (adapter, addr->sa_data));
}

/**
 * unm_nic_set_multi - Multicast
 **/

/* TODO */
static void
unm_nic_set_multi(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        struct dev_mc_list *mc_ptr;

        mc_ptr = netdev->mc_list;
        if (netdev->flags & IFF_PROMISC) {
                unm_nic_set_promisc_mode(adapter);
        } else {
                unm_nic_unset_promisc_mode(adapter);
        }

        /* TBD FSL            */

}

static int
unm_nic_xmit_frame(struct sk_buff *skb, struct net_device *netdev)
{
        unsigned int              nr_frags = 0;
        struct unm_adapter_s     *adapter = netdev_priv(netdev);
        struct _hardware_context *hw = &adapter->ahw;
        //struct pci_dev           *pdev;
        unsigned int              firstSegLen  = skb->len - skb->data_len;
        struct unm_skb_frag      *buffrag;
        unsigned int              i;

        uint32_t                  producer = 0;
        uint32_t                  saved_producer = 0;
        cmdDescType0_t               *hwdesc;
        int                       k;
        struct unm_cmd_buffer    *pbuf = NULL;
        static int                dropped_packet = 0;
        //struct unm_skb_frag      *frag;
        int                       fragCount;
        uint32_t                  local_producer = 0;
        uint32_t                  MaxTxDescCount = 0;
        uint32_t                  lastCmdConsumer = 0;
        int                       no_of_desc;
#ifdef BOUNCE_BUFFERS
        struct bounce_index       tmpbi;
        char                     *bounce_data;
        unsigned long             bounce_flags;
#endif

        adapter->stats.xmitcalled++;
        nr_frags = skb_shinfo(skb)->nr_frags;
        fragCount = 1 + nr_frags;

        if(unlikely(skb->len <= 0)) {
                dev_kfree_skb_any(skb);
                adapter->stats.badskblen++;
                return 0;
        }

        if (fragCount > MAX_BUFFERS_PER_CMD) {
                printk("%s: %s unm_nic_xmit_frame: fragCount (%d) too large, "
                            "can handle only %d frags\n", unm_nic_driver_name,
                                netdev->name, fragCount, MAX_BUFFERS_PER_CMD);
                goto drop_packet;
         }

        /*
         * Everything is set up. Now, we just need to transmit it out.
         * Note that we have to copy the contents of buffer over to
         * right place. Later on, this can be optimized out by de-coupling the
         * producer index from the buffer index.
         */
retry_getting_window:
        spin_lock_bh(&adapter->tx_lock);
        if(adapter->total_threads >= MAX_XMIT_PRODUCERS){
                spin_unlock_bh(&adapter->tx_lock);
                /*
                 * Yield CPU
                 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                if(!in_atomic())
                        schedule();
                else {
#endif
                        for(i = 0; i < 20; i++)
                                cpu_relax();    /*This a nop instr on i386*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                }
#endif
                goto retry_getting_window;
        }
        local_producer = adapter->cmdProducer;
        /* There 4 fragments per descriptor */
        no_of_desc = (fragCount + 3) >> 2;
#ifdef UNM_NETIF_F_TSO
        if(netdev->features & NETIF_F_TSO){
                if (TSO_SIZE(skb_shinfo(skb)) >  0) {
                        no_of_desc++;
                        if(((skb->nh.iph)->ihl *sizeof(u32)) +
                                    ((skb->h.th)->doff * sizeof(u32)) +
                                    sizeof(struct ethhdr) >
                                    (sizeof(cmdDescType0_t) - IP_ALIGNMENT_BYTES)) {
                                no_of_desc++;
                        }
                }
        }
#endif
        k = adapter->cmdProducer;
        MaxTxDescCount = adapter->MaxTxDescCount;
        lastCmdConsumer = adapter->lastCmdConsumer;
        if((k + no_of_desc) >=
                ((lastCmdConsumer <= k) ? lastCmdConsumer + MaxTxDescCount :
                    lastCmdConsumer)) {
                spin_unlock_bh(&adapter->tx_lock);
                /*port->stats.nocmddescriptor ++;
                DPRINTK(1, ERR, "No command descriptors available,"
                                " producer = %d, consumer = %d count=%llu,"
                                " dropping packet\n", producer,
                                adapter->lastCmdConsumer,
                                port->stats.nocmddescriptor);
				*/
                goto requeue_packet;
        }
        k = get_index_range(k, MaxTxDescCount,no_of_desc);
        adapter->cmdProducer = k;
        adapter->total_threads++;
        adapter->num_threads++;

        spin_unlock_bh(&adapter->tx_lock);
        /* Copy the descriptors into the hardware    */
        producer = local_producer;
        saved_producer = producer;
        hwdesc = &hw->cmdDescHead[producer];
        memset(hwdesc, 0, sizeof(cmdDescType0_t));
        /* Take skb->data itself */
        pbuf = &adapter->cmd_buf_arr[producer];

#ifdef UNM_NETIF_F_TSO
        if(netdev->features & NETIF_F_TSO){
                if (TSO_SIZE(skb_shinfo(skb)) >  0) {
                        pbuf->mss=TSO_SIZE(skb_shinfo(skb));
                        hwdesc->mss=TSO_SIZE(skb_shinfo(skb));
		//	printk("mss = %d\n",hwdesc->mss);
                } else {
                        pbuf->mss=0;
                        hwdesc->mss=0;
                }
        }
        else
#endif
        {
                pbuf->mss=0;
                hwdesc->mss=0;
        }

        //pbuf->no_of_descriptors = no_of_desc;
        pbuf->totalLength = skb->len;
        pbuf->skb = skb;
        pbuf->cmd = TX_ETHER_PKT;
        pbuf->fragCount = fragCount;
        pbuf->port = adapter->portnum;
        buffrag = &pbuf->fragArray[0];

        buffrag->dma = pci_map_single(adapter->pdev, skb->data, firstSegLen,
                                        PCI_DMA_TODEVICE);
#ifdef BOUNCE_BUFFERS
        pbuf->tx_bounce_buff = 0;
#endif

        buffrag->length = firstSegLen;
        hwdesc->totalLength = skb->len;
        hwdesc->numOfBuffers = fragCount;
        hwdesc->opcode = TX_ETHER_PKT;

        hwdesc->port = adapter->portnum;
	hwdesc->ctx_id = adapter->portnum;
        hwdesc->buffer1Length = firstSegLen;
        hwdesc->AddrBuffer1   = buffrag->dma;

#ifdef BOUNCE_BUFFERS
        if(buffrag->dma + firstSegLen > unm_dma_mask) {
               pci_unmap_single(adapter->pdev, buffrag->dma, firstSegLen,
                                             PCI_DMA_TODEVICE);
               spin_lock_irqsave(&tx_bounce_lock, bounce_flags);
               if (unm_get_next_bounce_buffer(&tx_bounce_head, &tx_bounce_tail,
                       &tmpbi, firstSegLen)) {
                       spin_unlock_irqrestore(&tx_bounce_lock, bounce_flags);
                       goto requeue_packet;
               }

               spin_unlock_irqrestore(&tx_bounce_lock, bounce_flags);

               pbuf->tx_bounce_buff = IS_BOUNCE;
               bounce_data = tx_bounce_ptr[tmpbi.buffer_index] +
                       tmpbi.buffer_offset;
               buffrag->dma = pci_map_single(adapter->pdev, bounce_data,
                                             firstSegLen, PCI_DMA_TODEVICE);
               hwdesc->AddrBuffer1 = buffrag->dma;
               memcpy(bounce_data, skb->data, firstSegLen);
               pbuf->bnext = tx_bounce_head;
        }
#endif

        for (i=1, k=1; i < fragCount; i ++, k++) {
                struct skb_frag_struct *frag;
                int    len, temp_len ;
                unsigned long offset;
                dma_addr_t temp_dma;
#ifdef BOUNCE_BUFFERS
                struct page *bounce_frag_page;
                __uint32_t bounce_page_offset;
#endif

                /* move to next desc. if there is a need */
                if ((i & 0x3) == 0){
                        k = 0;

                        producer = get_next_index(producer,
                                                    adapter->MaxTxDescCount);
                        hwdesc = &hw->cmdDescHead[producer];
                        memset(hwdesc, 0, sizeof(cmdDescType0_t));
                }
                frag   = &skb_shinfo(skb)->frags[i - 1];
                len    = frag->size;
                offset = frag->page_offset;

                temp_len = len;
                temp_dma    = pci_map_page(adapter->pdev, frag->page, offset,
                                            len, PCI_DMA_TODEVICE);
                buffrag++;
                buffrag->dma = temp_dma;
                buffrag->length = temp_len;
#ifdef BOUNCE_BUFFERS
                if(temp_dma + temp_len > unm_dma_mask){
                       pci_unmap_single(adapter->pdev, temp_dma, len,
                               PCI_DMA_TODEVICE);
                       spin_lock_irqsave(&tx_bounce_lock, bounce_flags);

                       if (unm_get_next_bounce_buffer(&tx_bounce_head,
                               &tx_bounce_tail, &tmpbi, len)) {
                               spin_unlock_irqrestore(&tx_bounce_lock, bounce_flags);
                               goto requeue_packet;
                       }
                       spin_unlock_irqrestore(&tx_bounce_lock, bounce_flags);

                       pbuf->tx_bounce_buff = IS_BOUNCE;
                       bounce_data = tx_bounce_ptr[tmpbi.buffer_index] +
                              tmpbi.buffer_offset;

                       bounce_frag_page = virt_to_page(bounce_data);
                       bounce_page_offset = (unsigned long)bounce_data -
                                              (unsigned long)((unsigned long)
                                                bounce_data & PAGE_MASK);
                       temp_dma = pci_map_page(adapter->pdev, bounce_frag_page,
                                                  bounce_page_offset, len,
                                                  PCI_DMA_TODEVICE);
                       buffrag->dma = temp_dma;
			memcpy(bounce_data, kmap_atomic(frag->page, 
                                                       KM_SKB_DATA_SOFTIRQ) 
                                               + offset, len);
                       kunmap(frag->page);
                       pbuf->bnext = tx_bounce_head;
                }
#endif

                DPRINTK(1, INFO, "for loop. i=%d k=%d\n", i, k);
                switch (k) {
                    case 0:
                            hwdesc->buffer1Length = temp_len;
                            hwdesc->AddrBuffer1   = temp_dma ;
                            break;
                    case 1:
                            hwdesc->buffer2Length = temp_len;
                            hwdesc->AddrBuffer2   = temp_dma ;
                            break;
                    case 2:
                            hwdesc->buffer3Length = temp_len;
                            hwdesc->AddrBuffer3   = temp_dma ;
                            break;
                    case 3:
                            hwdesc->buffer4Length = temp_len;
                            hwdesc->AddrBuffer4   = temp_dma ;
                            break;
                }
                frag ++;
        }
        producer = get_next_index(producer, adapter->MaxTxDescCount);

#ifdef UNM_NETIF_F_TSO
        /* might change opcode to TX_TCP_LSO */
        if(netdev->features & NETIF_F_TSO)
                unm_tso_check(adapter, &hw->cmdDescHead[saved_producer], skb);
        else
#endif
        unm_tx_csum(adapter, &hw->cmdDescHead[saved_producer], skb);

        if (vlan_tx_tag_get(skb)) {
                hw->cmdDescHead[saved_producer].flags |= FLAGS_VLAN_TAGGED;
        }

        /* For LSO, we need to copy the MAC/IP/TCP headers into
         * the descriptor ring
         */
        if (hw->cmdDescHead[saved_producer].opcode == TX_TCP_LSO) {
                int hdrLen, firstHdrLen, moreHdr;
                hdrLen = hw->cmdDescHead[saved_producer].totalHdrLength;
                if (hdrLen > (sizeof(cmdDescType0_t)-IP_ALIGNMENT_BYTES)) {
                        firstHdrLen = sizeof(cmdDescType0_t) -
                                        IP_ALIGNMENT_BYTES;
                        moreHdr=1;
                } else {
                        firstHdrLen = hdrLen;
                        moreHdr=0;
                }
                /* copy the MAC/IP/TCP headers to the cmd descriptor list */
                hwdesc = &hw->cmdDescHead[producer];

                /* copy the first 64 bytes */
                memcpy(((void *)hwdesc)+IP_ALIGNMENT_BYTES,
                        (void *)(skb->data), firstHdrLen);
#if 0
		{
			__uint64_t *temp = (__uint64_t *)hwdesc;
			int i;
			for (i=0; i< 8; i++) {
				printk("port%d firstheader%d: %llx\n",adapter->portnum,i, *temp); 
				temp++;
			}
		}
#endif
                producer = get_next_index(producer, MaxTxDescCount);

                if (moreHdr) {
                        hwdesc = &hw->cmdDescHead[producer];
                        /* copy the next 64 bytes - should be enough except
                         * for pathological case
                         */
                        memcpy((void *)hwdesc, (void *)(skb->data) +
                                    firstHdrLen, hdrLen-firstHdrLen);
#if 0
			{
				__uint64_t *temp = (__uint64_t *)hwdesc; 
				int i;
				for (i=0; i< 8; i++) {
					printk("port%d secondheader%d: %llx\n",adapter->portnum,i, *temp); 
					temp++;
				}
			}
#endif

                        producer = get_next_index(producer, MaxTxDescCount);
                }
        }
        spin_lock_bh(&adapter->tx_lock);
	adapter->stats.txbytes += hw->cmdDescHead[saved_producer].totalLength;
        /* Code to update the adapter considering how many producer threads
           are currently working */
        if ((-- adapter->num_threads) == 0){
                /* This is the last thread */
                uint32_t crb_producer = adapter->cmdProducer;
#if 0
                ctx_msg  msg = {0};
#endif
                hw->cmdProducer = adapter->cmdProducer;

		 read_lock(&adapter->adapter_lock);
		 unm_nic_update_cmd_producer(adapter, crb_producer);
/*                 UNM_NIC_PCI_WRITE_32(crb_producer,
                                          CRB_NORMALIZE(adapter,
                                                 CRB_CMD_PRODUCER_OFFSET));
*/		 read_unlock(&adapter->adapter_lock);
#if 0
                msg.PegId = UNM_XMIT_PEG_DB_ID;
                msg.privId = 1;
                msg.Count = crb_producer;
                msg.CtxId = adapter->portnum;
                msg.Opcode = UNM_CMD_PRODUCER;
                read_lock(&adapter->adapter_lock);
		switch(adapter->portnum){
			UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
					DB_NORMALIZE(adapter, UNM_CMD_PRODUCER_OFFSET));
		}
                read_unlock(&adapter->adapter_lock);
#endif

                adapter->total_threads = 0;
        }
#if 0
        else {
                uint32_t crb_producer = 0;
                ctx_msg  msg = {0};
                /*
		 *read_lock(&adapter->adapter_lock);
                 *crb_producer =
                 *        UNM_NIC_PCI_READ_32(
                 *                CRB_NORMALIZE(adapter,
                 *                        CRB_CMD_PRODUCER_OFFSET));
		 *read_unlock(&adapter->adapter_lock);
                 */
                /*
                 * This piece of code works only if cmd producer has
                 * its own doorbell page
                 */

	        read_lock(&adapter->adapter_lock);
                *((uint32_t *) &msg) =
                        UNM_NIC_PCI_READ_32(DB_NORMALIZE(adapter,
                                    UNM_CMD_PRODUCER_OFFSET));
	        read_unlock(&adapter->adapter_lock);
                crb_producer = msg.Count;
                if (crb_producer == local_producer) {
                        crb_producer = get_index_range(crb_producer,
                                                        MaxTxDescCount,
                                                        no_of_desc);
                        /*
			 *read_lock(&adapter->adapter_lock);
                         *UNM_NIC_PCI_WRITE_32(crb_producer,
                         *                 CRB_NORMALIZE(adapter,
                         *                         CRB_CMD_PRODUCER_OFFSET));
			 *read_unlock(&adapter->adapter_lock);
                         */
                        msg.PegId = UNM_XMIT_PEG_DB_ID;
                        msg.privId = 1;
                        msg.Count = crb_producer;
                        msg.CtxId = adapter->portnum;
                        msg.Opcode = UNM_CMD_PRODUCER;
                        read_lock(&adapter->adapter_lock);
                        UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                                DB_NORMALIZE(adapter,
                                        UNM_CMD_PRODUCER_OFFSET));
                        read_unlock(&adapter->adapter_lock);

                        hw->cmdProducer = crb_producer;
                }
        }
#endif
        //port->stats.xmitfinished++;
	adapter->stats.xmitfinished++;
        spin_unlock_bh(&adapter->tx_lock);

        netdev->trans_start = jiffies;

        DPRINTK(1, INFO, "wrote CMD producer %x to phantom\n", producer);

        DPRINTK(1, INFO, "Done. Send\n");
        goto unm_nic_xmit_success;

requeue_packet:
        spin_lock_bh(&adapter->tx_lock);
        netif_stop_queue(netdev);
        //port->flags |= NETDEV_STATUS;
        adapter->status |= NETDEV_STATUS;
        spin_unlock_bh(&adapter->tx_lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
        return NETDEV_TX_BUSY;
#else
        return 1;
#endif

drop_packet:
#if 0
        pbuf = &adapter->cmd_buf_arr[producer];
        frag = &pbuf->fragArray[0];
        pdev = adapter->pdev;
        pci_unmap_single(pdev, frag->dma, frag->length, PCI_DMA_TODEVICE);
        for (i = 1; i < fragCount; i ++) {
                frag ++;    /* Get the next frag */
                pci_unmap_page(pdev, frag->dma, frag->length,
                                PCI_DMA_TODEVICE);
        }

        //port->stats.txdropped++;
	adapter->stats.txdropped++;
        dev_kfree_skb_any(pbuf->skb);
        pbuf->skb = NULL;
#endif
	adapter->stats.txdropped++;
        dev_kfree_skb_any(skb);
        if ((++dropped_packet & 0xff) == 0xff) {
                printk("%s: %s droppped packets = %d\n", unm_nic_driver_name,
                                netdev->name, dropped_packet);
        }
        //netif_stop_queue(netdev);
unm_nic_xmit_success:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
        return NETDEV_TX_OK;
#else
        return 0;
#endif
}


static int
unm_loopback_xmit_frame(struct sk_buff *skb, struct net_device *netdev)
{
        unsigned long flags;
        int rv;

        local_irq_save(flags);
        rv = unm_nic_xmit_frame(skb, netdev);
        local_irq_restore(flags);
        return rv;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void unm_watchdog(unsigned long v)
{
	struct unm_adapter_s	*adapter = (struct unm_adapter_s *)v;
	SCHEDULE_WORK(&adapter->watchdog_task);
}
#endif

static int unm_nic_check_temp(struct unm_adapter_s *adapter)
{
    uint32_t temp, temp_state, temp_val;
	int rv = 0;

    temp = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                                    CRB_TEMP_STATE));

	temp_state = nx_get_temp_state(temp);
	temp_val = nx_get_temp_val(temp);

	if (temp_state == NX_TEMP_PANIC) {
                netdev_list_t *this;
                printk(KERN_ALERT
               "%s: Device temperature %d degrees C exceeds maximum allowed. "
               "Hardware has been shut down.\n", unm_nic_driver_name, temp_val);
		for (this = adapter->netlist; this != NULL; this = this->next) {
			netif_carrier_off(this->netdev);
			netif_stop_queue(this->netdev);
		}
		rv = 1;
	}
    adapter->temp = temp_state;
	return rv;
}


static void
unm_watchdog_task(TASK_PARAM adapid)
{
	struct net_device *netdev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	struct unm_adapter_s *adapter = container_of(adapid,	
	    struct unm_adapter_s, watchdog_task);
#else
        struct unm_adapter_s *adapter = (struct unm_adapter_s *)adapid;
#endif
	unsigned long        flags;

	if ((adapter->portnum  == 0) && unm_nic_check_temp(adapter)) {
		/*We return without turning on the netdev queue as there
		 *was an overheated device
		 */
		return;
	}
	if (adapter->driver_mismatch) {
		/*We return without turning on the netdev queue as there
		 *was a mismatch in driver and firmware version
		 */
		return;
	}

	UNM_WRITE_LOCK_IRQS(&adapter->adapter_lock, flags);
	unm_nic_pci_change_crbwindow(adapter, 1);
	UNM_WRITE_UNLOCK_IRQR(&adapter->adapter_lock, flags);

	unm_nic_handle_phy_intr(adapter);

	netdev = adapter->netdev;
	if ((netdev->flags & IFF_UP) && !netif_carrier_ok(netdev) &&
			unm_link_ok(adapter) ) {
		printk(KERN_INFO"%s %s (port %d), Link is up\n",
				unm_nic_driver_name, netdev->name, adapter->portnum);
		netif_carrier_on(netdev);
		netif_wake_queue(netdev);
	} else if(!(netdev->flags & IFF_UP) && netif_carrier_ok(netdev)) {
		printk(KERN_ERR "%s %s Link is Down\n",
				unm_nic_driver_name, netdev->name);
		netif_carrier_off(netdev);
		netif_stop_queue(netdev);
	}
	mod_timer(&adapter->watchdog_timer, jiffies + 2 * HZ);
}

static void
unm_tx_timeout(struct net_device *netdev)
{
        struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        //SCHEDULE_WORK(adapter->tx_timeout_task + adapter->portnum);
        SCHEDULE_WORK(adapter->tx_timeout_task);
#else
        //schedule_task(adapter->tx_timeout_task + adapter->portnum);
        schedule_task(adapter->tx_timeout_task);
#endif
}

static void
unm_tx_timeout_task(TASK_PARAM adapid)
{
	struct net_device	*netdev = (struct net_device *)adapid;
#if 0
        struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;
        unsigned long flags;

        printk(KERN_ERR "%s %s: transmit timeout, resetting.\n",
                        unm_nic_driver_name, netdev->name);
        spin_lock_irqsave(&adapter->lock, flags);
        unm_nic_close(netdev);
        unm_nic_open(netdev);
        spin_unlock_irqrestore(&adapter->lock, flags);
        netdev->trans_start = jiffies;
        netif_wake_queue(netdev);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
        printk(KERN_ERR "%s : transmit timeout, NEED to reset.\n",
                        unm_nic_driver_name);
#else
        printk(KERN_ERR "%s %s: transmit timeout, NEED to reset.\n",
                        unm_nic_driver_name, netdev->name);
#endif

}

static void
unm_nic_clear_stats(struct unm_adapter_s *adapter)
{
        memset(&adapter->stats, 0, sizeof(adapter->stats));
        return;
}

/*
 * unm_nic_get_stats - Get System Network Statistics
 * @netdev: network interface device structure
 */
static struct net_device_stats *
unm_nic_get_stats(struct net_device *netdev)
{
        struct unm_adapter_s         *adapter  = netdev_priv(netdev);
        struct net_device_stats *stats = &adapter->net_stats;

        /* unm_update_stats(port);   TBD FSL */
        memset(stats, 0, sizeof(*stats));

        /* total packets received   */
        stats->rx_packets         = adapter->stats.no_rcv;
        /* total packets transmitted    */
        stats->tx_packets         = adapter->stats.xmitedframes +
                                    adapter->stats.xmitfinished;
        /* total bytes received     */
        stats->rx_bytes           = adapter->stats.rxbytes;
        /* total bytes transmitted  */
        stats->tx_bytes           = adapter->stats.txbytes;
        /* bad packets received     */
        stats->rx_errors          = adapter->stats.rcvdbadskb;
        /* packet transmit problems */
        stats->tx_errors          = adapter->stats.nocmddescriptor;
        /* no space in linux buffers    */
        stats->rx_dropped         = adapter->stats.updropped;
        /* no space available in linux  */
        stats->tx_dropped         = adapter->stats.txdropped;

        return stats;
}

/*
 * unm_nic_change_mtu - Change the Maximum Transfer Unit
 * Returns 0 on success, negative on failure
 */
static int
unm_nic_change_mtu(struct net_device *netdev, int new_mtu)
{
        struct unm_adapter_s       *adapter    = netdev_priv(netdev);

        if (new_mtu & 0xffff0000)
                return -EINVAL;

        if (new_mtu > 8000) {
                printk(KERN_ERR"%s: %s MTU > 8000 is not supported\n",
                        unm_nic_driver_name, netdev->name);
                return -EINVAL;
        }

        unm_nic_set_mtu(adapter, new_mtu);
        netdev->mtu  = new_mtu;
        adapter->mtu = new_mtu;

        return 0;
}

/**
 * unm_update_stats - Update the board statistics counters
 * @adapter: board private structure
 **/

/* TODO */
void
unm_update_stats(struct unm_adapter_s *adapter)
{
        unsigned long flags;

        spin_lock_irqsave(&adapter->stats_lock, flags);

        /* TBD    FSL stats for the port        */

        spin_unlock_irqrestore(&adapter->stats_lock, flags);
}

/**
 * unm_nic_irq_enable - Enable default interrupt generation settings
 * @adapter: board private structure
 **/

/* TODO */
static inline void
unm_nic_irq_enable(struct unm_adapter_s *adapter)
{
        DPRINTK(1, INFO, "enabling irq %p\n", adapter);
        /* enable the interrupts  TBD FSL */
}

static void
unm_nic_disable_int(unm_adapter *adapter)
{
        uint32_t     mask;

        DPRINTK(1,INFO,"Entered ISR Disable \n");
        mask = 0x7ff;

        /*
         * ISR_INT_MASK: Can be read from window 0 or 1.
         */
#if 0		
	TO BE enabled when hardware/firmware starts using bits properly
	switch(adapter->portnum) {
		case 0:
			mask |= TARGET_INTERRUPT_MASK_0;
			break;
		case 1:
			mask |= TARGET_INTERRUPT_MASK_1;
			break;
		case 2:
			mask |= TARGET_INTERRUPT_MASK_2;
			break;
		case 3:
			mask |= TARGET_INTERRUPT_MASK_3;
			break;
	}
#endif
        UNM_NIC_PCI_WRITE_32(mask, (void *)(PCI_OFFSET_SECOND_RANGE(adapter, ISR_INT_MASK)));

        DPRINTK(1,INFO,"Done with Disable Int\n");

        return;
}

void
unm_nic_enable_int(unm_adapter *adapter)
{
        u32                mask;

        DPRINTK(1, INFO, "Entered ISR Enable \n");

        //UNM_LOCK(&adapter->adapter_lock, flags);

        switch (adapter->ahw.board_type) {
        case UNM_NIC_GBE:
            mask  =  0x77b;
            break;
        case UNM_NIC_XGBE:
            mask  =  0x77f;
            break;
        default:
            mask  =  0x7ff;
            break;
        }
#if 0
	TO BE enabled when hardware/firmware starts using bits properly
	switch(adapter->portnum) {
		case 0:
			mask |= !TARGET_INTERRUPT_MASK_0;
			break;
		case 1:
			mask |= !TARGET_INTERRUPT_MASK_1;
			break;
		case 2:
			mask |= !TARGET_INTERRUPT_MASK_2;
			break;
		case 3:
			mask |= !TARGET_INTERRUPT_MASK_3;
			break;
	}
#endif
	
        UNM_NIC_PCI_WRITE_32(mask, (void *)(PCI_OFFSET_SECOND_RANGE(adapter, ISR_INT_MASK)));

        if (!(adapter->flags & UNM_NIC_MSI_ENABLED)) {
               mask = 0xbff;
	       UNM_NIC_PCI_WRITE_32(0X0, CRB_NORMALIZE(adapter, CRB_INT_VECTOR));
               UNM_NIC_PCI_WRITE_32(mask, (void *)(PCI_OFFSET_SECOND_RANGE(adapter, ISR_INT_TARGET_MASK)));
        }

        //UNM_UNLOCK(&adapter->adapter_lock, flags);

        DPRINTK(1,INFO,"Done with enable Int\n");

        return;
}

#define UNM_NIC_INT_RESET 0x2004

#define find_diff_among(a,b,range) ((a)<=(b)?((b)-(a)):((b)+(range)-(a)))

static int
unm_handle_int(unm_adapter *adapter, struct net_device *netdev)
{
        u32                  ret = 0;
#if !defined(UNM_NIC_NAPI)
        static volatile u32  rcvd = 0;
        int                  ctx;
#endif

        DPRINTK(1, INFO, "Entered handle ISR\n");
        adapter->stats.ints++;

        if (!(adapter->flags & UNM_NIC_MSI_ENABLED)) {
                int count = 0;
                u32 mask;
		u32 our_int = 0;
		read_lock(&adapter->adapter_lock);
		our_int = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter, CRB_INT_VECTOR));
                if ((our_int & (0x80 << adapter->portnum)) == 0) {
                        /* not our interrupt */
                        read_unlock(&adapter->adapter_lock);
                        return ret;
                }
                unm_nic_disable_int(adapter);
                /* Window = 0 or 1 */
                do {
                        UNM_NIC_PCI_WRITE_32(0xffffffff, (void *)
                              (PCI_OFFSET_SECOND_RANGE(adapter, ISR_INT_TARGET_STATUS)));
                        mask = UNM_NIC_PCI_READ_32((void *)
                                (pci_base_offset(adapter, ISR_INT_VECTOR)));
                } while (((mask & 0x80) != 0) && (++count < 32));
                if ((mask & 0x80) != 0) {
                        printk("Could not disable interrupt completely\n");
                }
                read_unlock(&adapter->adapter_lock);
        }

        //adapter->stats.hostints++;

#if !defined(UNM_NIC_NAPI)
        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                rcvd += unm_process_rcv_ring(adapter, ctx, MAX_STATUS_HANDLE);
        }
#else
        if (unm_nic_rx_has_work(adapter) || unm_nic_tx_has_work(adapter)) {
		if (netif_running(netdev)) {
			ret = netif_rx_schedule_prep(netdev);
		} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
			ret = __netif_rx_schedule_prep(netdev);
#else
			/*
			 * There is no API in these versions of the kernel.
			 */
			ret = !test_and_set_bit(__LINK_STATE_RX_SCHED,
						&netdev->state);
#endif
		}
		if (ret) {
			/*
			 * Interrupts are already disabled.
			 */
			__netif_rx_schedule(netdev);
		} else {
			static unsigned int intcount = 0;
			if ((++intcount & 0xfff) == 0xfff) {
			//	printk(KERN_ERR "%s: %s interrupt %d while in poll\n",
			//	       unm_nic_driver_name, netdev->name, intcount);
			}
		}
		ret = 1;
	}
#endif
#if 0
			/*
			 * Try reclaiming transmit buffers.
			 */
        if(find_diff_among(adapter->lastCmdConsumer,
                adapter->cmdProducer, adapter->MaxTxDescCount) >
                (adapter->MaxTxDescCount>>5)) {
                if (loopback_start) {
                        unm_process_cmd_ring((unsigned long)adapter);
                } else {
                        unm_nic_read_crb (adapter, CRB_CMD_CONSUMER_OFFSET,
                                          &consumer);
                        adapter->cmdPegConsumer = consumer;
                        if((find_diff_among(adapter->lastCmdConsumer,
                                            consumer, adapter->MaxTxDescCount))
                                > (adapter->MaxTxDescCount>>5))
                                tasklet_schedule(&adapter->tx_tasklet);
                }
        }
#endif
        if (adapter->ahw.board_type == UNM_NIC_GBE)
                unm_nic_handle_phy_intr(adapter);
        if (ret == 0) {
                /* TODO: enable non-rx type interrupts */
		read_lock(&adapter->adapter_lock);
                unm_nic_enable_int(adapter);
		read_unlock(&adapter->adapter_lock);
        }

        return (ret);
}

/**
 * unm_intr - Interrupt Handler
 * @irq: interrupt number
 * data points to adapter stucture (which may be handling more than 1 port
 **/
#ifdef OLD_KERNEL
void
#else
irqreturn_t
#endif
unm_intr(int irq, void *data, struct pt_regs *regs)
{
        struct unm_adapter_s *adapter;
        struct net_device    *netdev;

        if(unlikely(!irq)) {
#ifdef OLD_KERNEL
                return;
#else
                return IRQ_NONE;  /* Not our interrupt */
#endif
        }

        adapter = (struct unm_adapter_s *)data;
        //unm_recv_context_t  *recv_ctx = &(adapter->recv_ctx[0]);
	netdev  = adapter->netdev;

	/* process our status queue */
	if(netif_running(netdev) || adapter->tnic.registered) {
		unm_handle_int(adapter, netdev);
	}

	/*
	 * None of the devices are up for 
	 */

#ifdef OLD_KERNEL
        return;
#else
        return IRQ_HANDLED;
#endif
}

#ifdef UNM_NIC_NAPI
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,21)
static inline void __netif_rx_complete(struct net_device *dev)
{
        if (!test_bit(__LINK_STATE_RX_SCHED, &dev->state))
                BUG();
        list_del(&dev->poll_list);
        smp_mb__before_clear_bit();
        clear_bit(__LINK_STATE_RX_SCHED, &dev->state);
}
#endif /* LINUX_VERSION_CODE < 2.4.23 */

#ifdef PEGNET_NIC
#ifndef __VMKERNEL_MODULE__
/*
 * Check if there are any command descriptors pending.
 */
static inline int unm_get_pending_cmd_desc_cnt(unm_pending_cmd_descs_t *pending_cmds)
{
        return (pending_cmds->cnt);
}
#endif
#endif

static int
unm_nic_poll(struct net_device *netdev, int *budget)
{
        struct unm_adapter_s *adapter    = netdev_priv(netdev);
        int                   work_to_do = min(*budget, netdev->quota);
        int                   done = 1;
        int                   ctx;
        int                   this_work_done;
        int                   work_done = 0;

        adapter->stats.polled++;
        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
        /*
         * TODO : Fairness issue. This will give undue weight to the
         * receive context 0.
         */

        /*
         * TODO : To avoid starvation, we give each of our receivers,
         * a fraction of the quota. Sometimes, it might happen that we
         * have enough quota to process every packet, but since all the
         * packets are on one context, it gets only half of the quota,
         * and ends up not processing it.
         */
                this_work_done = unm_process_rcv_ring(adapter, ctx,
                                                      work_to_do/MAX_RCV_CTX);
                work_done += this_work_done;
        }

        netdev->quota -= work_done;
        *budget -= work_done;

        if (work_done >= work_to_do && unm_nic_rx_has_work(adapter) != 0) {
                done  = 0;
        }

	if (unm_process_cmd_ring((unsigned long)adapter) == 0) {
                done = 0;
        }

        DPRINTK(1, INFO, "new work_done: %d work_to_do: %d\n",
                            work_done, work_to_do);
        if (done) {
                netif_rx_complete(netdev);
                /*unm_nic_hw_write(adapter,(uint64_t)ISR_INT_VECTOR,
                                    &reset_val,4);*/
		read_lock(&adapter->adapter_lock);
                unm_nic_enable_int(adapter);
		read_unlock(&adapter->adapter_lock);
        }

#ifdef  PEGNET_NIC
#ifndef __VMKERNEL_MODULE__
        /*
         * This comes after enabling the interrupt. The reason is that if done
         * before the interrupt is turned and the following routine could fill
         * up the cmd ring and the card could finish the processing and try
         * interrupting the host even before the interrupt is turned on again.
         * This could lead to missed interrupt and has an outside chance of
         * deadlock.
         */
        spin_lock(&adapter->tx_lock);
        if (unm_get_pending_cmd_desc_cnt(&adapter->pending_cmds)) {
                unm_proc_pending_cmd_desc(adapter);
        }
        spin_unlock(&adapter->tx_lock);
#endif
#endif  /* PEGNET_NIC */

        return (done ? 0 : 1);
}

/*
 * returns: 1 if there is work for the receive side
 *          0 if there is none
 */
static int
unm_nic_rx_has_work(struct unm_adapter_s *adapter)
{
        int ctx;

        for (ctx = 0; ctx < MAX_RCV_CTX; ++ctx) {
                unm_recv_context_t *recv_ctx = &(adapter->recv_ctx[ctx]);
                uint32_t             consumer;
                statusDesc_t        *descHead;
                statusDesc_t        *desc;    /* used to read status desc here */

                consumer = recv_ctx->statusRxConsumer;
                descHead = recv_ctx->rcvStatusDescHead;
                desc = &descHead[consumer];

                if ((desc->owner & STATUS_OWNER_HOST)) {
                    return 1;
                }
        }

        return 0;
}

/*
 * returns: 1 if there is work for the transmit side
 *          0 if there is none
 */
static int
unm_nic_tx_has_work(struct unm_adapter_s *adapter)
{
        if (find_diff_among(adapter->lastCmdConsumer,
                            adapter->cmdProducer,
                            adapter->MaxTxDescCount) > 0) {
                /* (adapter->MaxTxDescCount >> 5)) { */

                return 1;

        }

        return 0;
}
#endif

#ifdef CONFIG_NET_POLL_CONTROLLER
static void
unm_nic_poll_controller(struct net_device *netdev)
{
        struct unm_adapter_s        *adapter= netdev_priv(netdev);
        disable_irq(netdev->irq);
        unm_intr(netdev->irq, adapter, NULL);
        enable_irq(netdev->irq);
}
#endif

/*
 * unm_process_rcv() send the received packet to the protocol stack.
 * and if the number of receives exceeds RX_BUFFERS_REFILL, then we
 * invoke the routine to send more rx buffers to the Phantom...
 */
static inline void
unm_process_rcv(struct unm_adapter_s *adapter, int ctx, statusDesc_t *desc)
{
        struct pci_dev       *pdev     = adapter->pdev;
        struct net_device    *netdev   = adapter->netdev;
        int                   index    = desc->referenceHandle;
        unm_recv_context_t   *recv_ctx = &(adapter->recv_ctx[ctx]);
        struct unm_rx_buffer *buffer;
        struct sk_buff       *skb;
        uint32_t              length   = desc->totalLength;
        uint32_t              desc_ctx;
        unm_rcv_desc_ctx_t   *rcv_desc;
        int                   ret;
#ifdef UNM_XTRA_DEBUG
        int                   i;
#endif

#ifdef __VMKERNEL_MODULE__
	if(length < ETH_ZLEN)
		length = ETH_ZLEN;
#endif
        desc_ctx = desc->type;
        if (unlikely(desc_ctx >= NUM_RCV_DESC_RINGS)) {
                printk("%s: %s Bad Rcv descriptor ring\n", unm_nic_driver_name,
                        netdev->name);
                return;
        }

        rcv_desc = &recv_ctx->rcv_desc[desc_ctx];
        if (unlikely(index > rcv_desc->MaxRxDescCount)) {
                DPRINTK(1, ERR, "Got a buffer index:%x for %s desc type."
                       "Max is %x\n", index, RCV_DESC_TYPE_NAME(desc_ctx),
                       rcv_desc->MaxRxDescCount);
                return;
        }
        buffer   = &rcv_desc->rx_buf_arr[index];

        if(desc_ctx == RCV_DESC_LRO_CTXID) {
                buffer->lro_current_frags++;
                if (desc->last_lro_frag) {
                    buffer->lro_expected_frags=desc->lro_count;
                    buffer->lro_length=length;
                }
                if(buffer->lro_current_frags != buffer->lro_expected_frags) {
                    if (buffer->lro_expected_frags != 0) {
                        printk("LRO: (refhandle:%x) recv frag. wait for last."
                               " flags: %x expected:%d have:%d\n", index,
                               desc->last_lro_frag, buffer->lro_expected_frags,
                               buffer->lro_current_frags);
                    }

                    return;
                }
        }

        pci_unmap_single(pdev, buffer->dma, rcv_desc->dma_size,
                         PCI_DMA_FROMDEVICE);

        skb = (struct sk_buff *)buffer->skb;
#ifdef BOUNCE_BUFFERS
        if(buffer->rx_bounce_buff == IS_BOUNCE)
        {
                buffer->rx_bounce_buff = 0;
                memcpy(skb->data, buffer->bounce_ptr, rcv_desc->dma_size);
                rx_bounce_tail = buffer->bnext;
        }
#endif

        if (unlikely(skb == NULL)) {
                /*
                 * This should not happen and if it does, it is serious,
                 * catch it
                 */
                uint32_t temp1,temp2;
                /* Window = 0 or 1 */
		read_lock(&adapter->adapter_lock);
                temp1 = UNM_NIC_PCI_READ_32((void *)(PCI_OFFSET_SECOND_RANGE(adapter,
                                               + (UNM_CRB_PCIE+0x20000))));
                temp2 = UNM_NIC_PCI_READ_32((void *)(PCI_OFFSET_SECOND_RANGE(adapter,
                                               + (UNM_CRB_PCIE+0x20008))));
		read_unlock(&adapter->adapter_lock);

                /* Window = 0 */
                unm_nic_write_w0(adapter, UNM_CRB_PEG_NET_0+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_1+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_2+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_3+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_NIU+0x70000, 0);

                printk("got NULL skb for index %d  desc %p", index, desc);
                printk("of %s type\n", RCV_DESC_TYPE_NAME(desc_ctx));
                printk("%s: %s Halted the pegs and stopped the NIU consumer\n",
                        unm_nic_driver_name, netdev->name);
                printk("Curr DMA Addr: 0x%x, status:0x%x\n",
                        temp1,temp2);
                printk("prev desc %p: refh:%x\n",desc-1,
                       (desc-1)->referenceHandle);
                printk("next desc %p: refh:%x\n",desc+1,
                       (desc+1)->referenceHandle);
                printk("next desc %p: refh:%x\n",desc+2,
                       (desc+2)->referenceHandle);
                printk("and mine again %p: refh:%x\n",desc,
                       (desc)->referenceHandle);

                adapter->stats.rcvdbadskb++;
                return;
        }
#if defined(XGB_DEBUG)
        if (!skb_is_sane(skb)) {
                dump_skb(skb);
                printk(KERN_ERR"%s: %s index:%d skb(%p) dma(%p) is not sane\n",
                        unm_nic_driver_name,netdev->name,index,skb,buffer->dma);
                /* Window = 0 */
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_0+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_1+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_2+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_PEG_NET_3+0x3c, 1);
                unm_nic_write_w0 (adapter, UNM_CRB_NIU+0x70000, 0);

                return;
        }
#endif
#if defined(UNM_NIC_HW_CSUM)
        if (likely((adapter->rx_csum) && (desc->status == STATUS_CKSUM_OK))) {
                adapter->stats.csummed++;
                skb->ip_summed = CHECKSUM_UNNECESSARY;
        } else {
                skb->ip_summed = CHECKSUM_NONE;
        }
#endif
        skb->dev = netdev;
        if(desc_ctx == RCV_DESC_LRO_CTXID) {
               /* True length was only available on the last pkt */
               skb_put(skb, buffer->lro_length);
        } else {
               skb_put(skb, length);
        }

#if 1
        skb->protocol = eth_type_trans(skb, netdev);
#else
        /* this is an alternative, but we still need to check eth address */

        if (desc->prot==UNM_PROT_IP) {
            skb->protocol = htons(ETH_P_IP);
            skb->mac.raw=skb->data;
            skb_pull(skb,ETH_HLEN);
            skb->input_dev = netdev;
            eth = eth_hdr(skb);
            if(*eth->h_dest&1)
            {
                if(memcmp(eth->h_dest,dev->broadcast, ETH_ALEN)==0)
                        skb->pkt_type=PACKET_BROADCAST;
                else
                        skb->pkt_type=PACKET_MULTICAST;
            }

        } else {
            skb->protocol = eth_type_trans(skb, netdev);
        }
#endif

        if (capture_input) {
		if( rx_user_packet_data != NULL &&
		    (rx_user_pos + (length - 14)) <= rx_datalen) {
			memcpy(rx_user_packet_data + rx_user_pos, skb->data,
			       length - 14);
			rx_user_pos += (length - 14);
		}
	}

	if (!netif_running(netdev)) {
		kfree_skb(skb);
		goto packet_done_no_stats;
	}

#ifdef PEGNET_NIC
#ifndef __VMKERNEL_MODULE__
	/* At this point we have got a valid skb.
	 * Check whether this is a SYN that pegnet needs to handle
	 */
	if (desc->opcode == UNM_MSGTYPE_NIC_SYN_OFFLOAD) {
		/* check whether pegnet wants this */

		spin_lock(&adapter->tnic.lock);
		if (adapter->tnic.cb) {
			ret = adapter->tnic.cb->nx_syn_handler(netdev, desc->listener_id, skb);
			/* ok, it does */
			if (ret == 0) {
				kfree_skb(skb);
				spin_unlock(&adapter->tnic.lock);
				goto packet_done;
			}
#if 0
			printk("%s: ret:%d\n", __func__, ret);
#endif
		}
		spin_unlock(&adapter->tnic.lock);

		/* fallthrough for regular nic processing */
	}
#endif
#endif

#ifdef UNM_XTRA_DEBUG
        printk("%s: %s reading from %p index %d %d, %d bytes\n",
               unm_nic_driver_name, netdev->name, buffer->dma, index, length);
        for (i=0; &skb->data[i] < skb->tail; i++) {
                printk("%02x%c", skb->data[i], (i+1)%0x10?' ':'\n');
        }
        printk("\n");
#endif
#if defined(UNM_LOOPBACK)
        if (loopback_start) {
                for (i=0; i<6; i++) {
                        tmp = skb->data[i];
                        skb->data[i] = skb->data[i+6];
                        skb->data[i+6] = tmp;
                }
                unm_nic_xmit_frame(skb,netdev);
        }
#else
#ifndef UNM_NIC_NAPI
        ret = netif_rx(skb);
#else
        ret = netif_receive_skb(skb);
#endif
#endif

        /*
         * RH: Do we need these stats on a regular basis. Can we get it from
         * Linux stats.
         */
        switch (ret) {
        case NET_RX_SUCCESS:
                adapter->stats.uphappy++;
                break;

        case NET_RX_CN_LOW:
                adapter->stats.uplcong++;
                break;

        case NET_RX_CN_MOD:
                adapter->stats.upmcong++;
                break;

        case NET_RX_CN_HIGH:
                adapter->stats.uphcong++;
                break;

        case NET_RX_DROP:
                adapter->stats.updropped++;
                break;

        default:
               adapter->stats.updunno++;
                break;
        }

        netdev->last_rx = jiffies;

  packet_done:
        adapter->stats.no_rcv++;
        adapter->stats.rxbytes += length;

  packet_done_no_stats:
        rcv_desc->rcv_free++;
        rcv_desc->rcv_pending--;

        /*
         * We just consumed one buffer so post a buffer.
         * Buffers may come out of order in Rcv when LRO is turned ON.
         */
        buffer->skb                = NULL;
        buffer->state              = UNM_BUFFER_FREE;
        buffer->lro_current_frags  = 0;
        buffer->lro_expected_frags = 0;
        //adapter->stats.post_called++;

        TAILQ_INSERT_TAIL(&rcv_desc->free_rxbuf_list, buffer, link);

        return;
}

/* Process Receive status ring */
static uint32_t
unm_process_rcv_ring(struct unm_adapter_s *adapter, int ctx, int max)
{
        unm_recv_context_t  *recv_ctx = &(adapter->recv_ctx[ctx]);
        statusDesc_t        *descHead = recv_ctx->rcvStatusDescHead;
        statusDesc_t        *desc;    /* used to read status desc here */
        uint32_t             producer = 0; /* will be read from phantom */
        uint32_t             consumer = recv_ctx->statusRxConsumer;
        int                  count    = 0,ring;

        DPRINTK(1, INFO, "procesing receive\n");
        /*
         * we assume in this case that there is only one port and that is
         * port #1...changes need to be done in firmware to indicate port
         * number as part of the descriptor. This way we will be able to get
         * the netdev which is associated with that device.
         */
        while (count < max) {
                desc = &descHead[consumer];
#if defined(DEBUG)
                if (desc->owner != STATUS_OWNER_HOST &&
                    desc->owner != STATUS_OWNER_PHANTOM) {
                    DPRINTK(1, ERR, "desc(%p)owner is %x consumer:%d\n",
                           desc, desc->owner, consumer);
                    break;
                }
#endif
                if (!(desc->owner & STATUS_OWNER_HOST)) {
                    DPRINTK(1, ERR, "desc %p ownedby %s\n", desc,
                            STATUS_OWNER_NAME(desc));
                    //printk("desc %p ownedby %s\n", desc,
                     //       STATUS_OWNER_NAME(desc));
                    break;
                }

#if defined(PEGNET_NIC) && (UNM_HOST_MSG_TRANSPORT==UNM_HOST_MSG_TRANSPORT_NIC)
#ifndef __VMKERNEL_MODULE__
                if (desc->opcode == UNM_MSGTYPE_NIC_RXPKT_DESC
			|| desc->opcode == UNM_MSGTYPE_NIC_SYN_OFFLOAD) {
#endif /* __VMKERNEL_MODULE__ */
#endif  /* PEGNET_NIC */

                        unm_process_rcv(adapter, ctx, desc);

#if defined(PEGNET_NIC) && (UNM_HOST_MSG_TRANSPORT==UNM_HOST_MSG_TRANSPORT_NIC)
#ifndef __VMKERNEL_MODULE__
                } else {
                        unm_msg_t       msg;
                        int             cnt;

                        cnt = desc->descCnt;

                        /*
                         * Check if it is the extended queue message. If it is
                         * then read the rest of the descriptors also.
                         */
                        if (cnt > 1) {
                                uint32_t        tmp_consumer;
                                statusDesc_t    *last_desc;

                                *(uint64_t *)(&msg.hdr) = desc->body[0];
                                msg.body.values[0] = desc->body[1];

                                /*
                                 *
                                 */
                                tmp_consumer = ((consumer + 4) &
                                                (adapter->MaxRxDescCount - 1));
                                last_desc = &descHead[tmp_consumer];
                                if (!(last_desc->owner & STATUS_OWNER_HOST)) {
                                        /*
                                         * The whole queue message is not ready
                                         * yet so break out of here.
                                         */
                                        break;
                                }
                                desc->owner = STATUS_OWNER_PHANTOM;


                                consumer = ((consumer + 1) &
                                            (adapter->MaxRxDescCount - 1));
                                desc = &descHead[consumer];
                                msg.body.values[1] = desc->body[0];
                                msg.body.values[2] = desc->body[1];
                                desc->owner = STATUS_OWNER_PHANTOM;

                                consumer = ((consumer + 1) &
                                            (adapter->MaxRxDescCount - 1));
                                desc = &descHead[consumer];
                                msg.body.values[3] = desc->body[0];
                                msg.body.values[4] = desc->body[1];
                                desc->owner = STATUS_OWNER_PHANTOM;

                                consumer = ((consumer + 1) &
                                            (adapter->MaxRxDescCount - 1));
                                desc = &descHead[consumer];
                                msg.body.values[5] = desc->body[0];
                                msg.body.values[6] = desc->body[1];
                                desc->owner = STATUS_OWNER_PHANTOM;

                                consumer = ((consumer + 1) &
                                            (adapter->MaxRxDescCount - 1));
                                desc = &descHead[consumer];
                        } else {
                                /*
                                 * These messages expect the type field in the
                                 * queue message header to be set and nothing
                                 * else in the header to be set. So we copy
                                 * the desc->body[0] into the header.
                                 * desc->body[0] has the opcode which is same
                                 * as the queue message type.
                                 */
                                msg.hdr.word = desc->body[0];
                                msg.body.values[0] = desc->body[0];
                                msg.body.values[1] = desc->body[1];
                        }

			spin_lock(&adapter->tnic.lock);
			if (adapter->tnic.cb) {
				adapter->tnic.cb->nx_msg_handler(adapter->netdev, &msg);
			}	
			spin_unlock(&adapter->tnic.lock);
                }
#endif /* __VMKERNEL_MODULE__ */
#endif  /* PEGNET_NIC */

                desc->owner = STATUS_OWNER_PHANTOM;
                consumer = (consumer + 1) & (adapter->MaxRxDescCount - 1);
                count++;
        }
        if(count)
                for (ring = 0; ring < NUM_RCV_DESC_RINGS; ring++) {
                        unm_post_rx_buffers_nodb(adapter,ctx, ring);
                }


        /* update the consumer index in phantom */
        if (count) {
#if 0
                ctx_msg  msg = {0};
#endif

                //adapter->stats.process_rcv++;
                recv_ctx->statusRxConsumer = consumer;
                recv_ctx->statusRxProducer = producer;

                /* Window = 1 */

		 read_lock(&adapter->adapter_lock);
                 UNM_NIC_PCI_WRITE_32(consumer,
                         CRB_NORMALIZE(adapter,
                         recv_crb_registers[adapter->portnum].CRB_RCV_STATUS_CONSUMER));
		 read_unlock(&adapter->adapter_lock);
#if 0
                msg.PegId = UNM_RCV_PEG_DB_ID;
                msg.privId = 1;
                msg.Count = consumer;
                msg.CtxId = 0;
                msg.Opcode = UNM_RCV_STATUS_CONSUMER;
                read_lock(&adapter->adapter_lock);
                UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                        DB_NORMALIZE(adapter, UNM_RCV_STATUS_CONSUMER_OFFSET));
                read_unlock(&adapter->adapter_lock);
#endif

        }

        return (count);
}

#if defined(PEGNET_NIC)
#ifndef __VMKERNEL_MODULE__

/*
 * Initialize the allocated resources in pending_cmd_descriptors.
 */
static void unm_init_pending_cmd_desc(unm_pending_cmd_descs_t *pending_cmds)
{
	INIT_LIST_HEAD(&pending_cmds->free_list);
	INIT_LIST_HEAD(&pending_cmds->cmd_list);
        pending_cmds->curr_block = 0;
        pending_cmds->cnt = 0;
}

/*
 * Free up the allocated resources in pending_cmd_descriptors.
 */
static void unm_destroy_pending_cmd_desc(unm_pending_cmd_descs_t *pending_cmds)
{
        unm_pending_cmd_desc_block_t    *block;

	if (pending_cmds->curr_block) {
        	kfree(pending_cmds->curr_block);
	}

	while (!list_empty(&pending_cmds->free_list)) {

                block = list_entry(pending_cmds->free_list.next,
                                   unm_pending_cmd_desc_block_t, link);
		list_del(&block->link);
		kfree(block);
	}

	while (!list_empty(&pending_cmds->cmd_list)) {

                block = list_entry(pending_cmds->cmd_list.next,
                                   unm_pending_cmd_desc_block_t, link);
		list_del(&block->link);
		kfree(block);
	}
}

/*
 *
 */
static unm_pending_cmd_desc_block_t *unm_dequeue_pending_cmd_desc(
        unm_pending_cmd_descs_t *pending_cmds, uint32_t length)
{
	unm_pending_cmd_desc_block_t    *block;


        if (!list_empty(&pending_cmds->cmd_list)) {
                if (length < MAX_PENDING_DESC_BLOCK_SIZE) {
                        return (NULL);
                }
                block = list_entry(pending_cmds->cmd_list.next,
                                   unm_pending_cmd_desc_block_t, link);
                list_del(&block->link);
        } else {
                block = pending_cmds->curr_block;
                if (block == NULL || length < block->cnt) {
                        return (NULL);
                }
                pending_cmds->curr_block = NULL;
        }

        pending_cmds->cnt -= block->cnt;

        return (block);
}

/*
 *
 */
static inline unm_pending_cmd_desc_block_t *nx_alloc_pend_cmd_desc_block(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	return (kmalloc(sizeof (unm_pending_cmd_desc_block_t),
			in_atomic() ? GFP_ATOMIC : GFP_KERNEL));
#else
	return (kmalloc(sizeof (unm_pending_cmd_desc_block_t),
			in_interrupt() ? GFP_ATOMIC : GFP_KERNEL));
#endif
}

/*
 *
 */
static void unm_free_pend_cmd_desc_block(unm_pending_cmd_descs_t *pending_cmds,
					 unm_pending_cmd_desc_block_t *block)
{
/*         printk("%s: Inside\n", __FUNCTION__); */

        block->cnt = 0;
        list_add_tail(&block->link, &pending_cmds->free_list);
}

/*
 * This is to process the pending q msges for chimney.
 */
static void unm_proc_pending_cmd_desc(unm_adapter *adapter)
{
        uint32_t                producer;
        uint32_t                consumer;
        struct unm_cmd_buffer   *pbuf;
	uint32_t                length = 0;
        unm_pending_cmd_desc_block_t    *block;
        int                     i;

        producer = adapter->cmdProducer;
        consumer = adapter->lastCmdConsumer;
	if (producer == consumer) {
		length = adapter->MaxTxDescCount - 1;
	} else if (producer > consumer) {
		length = adapter->MaxTxDescCount - producer + consumer - 1;
	} else {
		/* consumer > Producer */
		length = consumer - producer - 1;
	}

/*         printk("%s: length = %u\n", __FUNCTION__, length); */

        while (length) {
                block = unm_dequeue_pending_cmd_desc(&adapter->pending_cmds,
                                                     length);
                if (block == NULL) {
                        break;
                }
                length -= block->cnt;

                for (i = 0; i < block->cnt; i++) {
                        pbuf = &adapter->cmd_buf_arr[producer];
                        pbuf->mss = 0;
                        pbuf->totalLength = 0;
                        pbuf->skb = NULL;
                        pbuf->cmd = PEGNET_REQUEST;
                        pbuf->fragCount = 0;
                        pbuf->port = 0;

                        adapter->ahw.cmdDescHead[producer] = block->cmd[i];

                        producer = get_next_index(producer,
                                                  adapter->MaxTxDescCount);
                }
                unm_free_pend_cmd_desc_block(&adapter->pending_cmds, block);
        }

        if (adapter->cmdProducer == producer) {
                return;
        }
        adapter->cmdProducer = producer;

        /*
         * Check if this is the only thread trying to update the producer. If
         * it is then write the CRB. If not the last thread will do it.
         */
        if (adapter->num_threads == 0) {
#if 0
                ctx_msg  msg = {0};
#endif
                adapter->ahw.cmdProducer = adapter->cmdProducer;

                /* write producer index to start the xmit */

#if 1
                read_lock(&adapter->adapter_lock);
		unm_nic_update_cmd_producer(adapter, adapter->cmdProducer);
                read_unlock(&adapter->adapter_lock);
#else
                msg.PegId = UNM_XMIT_PEG_DB_ID;
                msg.privId = 1;
                msg.Count = adapter->cmdProducer;
                msg.CtxId = 0;
                msg.Opcode = UNM_CMD_PRODUCER;
                read_lock(&adapter->adapter_lock);
                UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                        DB_NORMALIZE(adapter, UNM_CMD_PRODUCER_OFFSET));
                read_unlock(&adapter->adapter_lock);
#endif

        }
}

/*
 * This is invoked when we are out of descriptor ring space for
 * pegnet_cmd_desc. It queues them up and is to be processed when the
 * descriptor space is freed.
 */
static void nx_queue_pend_cmd_desc(unm_pending_cmd_descs_t *pending_cmds,
				   cmdDescType0_t *cmd_desc, uint32_t length)
{
	unm_pending_cmd_desc_block_t    *block;

	if (pending_cmds->curr_block) {
		block = pending_cmds->curr_block;

	} else if (!list_empty(&pending_cmds->free_list)) {

                block = list_entry(pending_cmds->free_list.next,
                                   unm_pending_cmd_desc_block_t, link);
		list_del(&block->link);
                block->cnt = 0;
	} else {
		printk("%s: *** BUG *** This should never happen\n",
		       __FUNCTION__);
		return;
	}

	memcpy(&block->cmd[block->cnt], cmd_desc, length);
	block->cnt++;

	if(block->cnt == MAX_PENDING_DESC_BLOCK_SIZE) {
		/*
                 * The array is full, put it in the tail of the cmd_list
                 */
		list_add_tail(&block->link, &pending_cmds->cmd_list);
		block = NULL;
	}

	pending_cmds->cnt++;
	pending_cmds->curr_block = block;
}

/*
 *
 */
static int nx_make_available_pend_cmd_desc(unm_pending_cmd_descs_t *pend,
					   int cnt)
{
	unm_pending_cmd_desc_block_t	*block;

	block = pend->curr_block;
	if (block && block->cnt <= (MAX_PENDING_DESC_BLOCK_SIZE - cnt)) {
		return (0);
	}

	if (!list_empty(&pend->free_list)) {
		return (0);
	}

	block = nx_alloc_pend_cmd_desc_block();
	if (block != NULL) {
		/*
		 * Puts it into the free list.
		 */
		unm_free_pend_cmd_desc_block(pend, block);
		return (0);
	}

	return (-ENOMEM);
}

/*
 *
 */
static int nx_queue_pend_cmd_desc_list(unm_adapter *adapter,
				       pegnet_cmd_desc_t *cmd_desc_arr,
				       int no_of_desc)
{
	int	i;
	int	rv;

	rv = nx_make_available_pend_cmd_desc(&adapter->pending_cmds,
					     no_of_desc);
	if (rv) {
		return (rv);
	}

	i = 0;
	do {
		nx_queue_pend_cmd_desc(&adapter->pending_cmds,
				       (cmdDescType0_t *)&cmd_desc_arr[i],
				       sizeof (pegnet_cmd_desc_t));
		i++;
	} while (i != no_of_desc);

	return (0);
}

/*
 * Compare adapter ids for two NetXen devices.
 */
int nx_nic_cmp_adapter_id(struct net_device *dev1, struct net_device *dev2)
{
	unm_adapter *adapter1 = (unm_adapter *)netdev_priv(dev1);
	unm_adapter *adapter2 = (unm_adapter *)netdev_priv(dev2);
	return memcmp(adapter1->id,adapter2->id,sizeof(adapter1->id));
}

/*
 * Checks if the specified device is a NetXen device.
 */
int nx_nic_is_netxen_device(struct net_device *netdev)
{
	return (netdev->do_ioctl == unm_nic_ioctl);
}


/*
 * Register TNIC driver with Nic driver.
 *
 * Return:
 *	0	- If registered successfully.
 *	1	- If already registered.
 *	-1	- If port or adapter is not set.
 */
int nx_nic_register_tnic(struct net_device *netdev, nx_tnic_adapter_t *data,
			 nx_tnic_callback_t *cb)
{
	struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;
	int		rv = -1;

        if (adapter) {

		spin_lock_bh(&adapter->tnic.lock);

		rv = 1;
		if (!adapter->tnic.registered) {
			adapter->tnic.data = data;
			adapter->tnic.cb = cb;
			adapter->tnic.registered = 1;
			rv = 0;
		}

		spin_unlock_bh(&adapter->tnic.lock);
        }

        return (rv);
}

/*
 * Unregisters a known TNIC adapter.
 */
void nx_nic_unregister_tnic(struct net_device *netdev)
{
	struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;

	spin_lock_bh(&adapter->tnic.lock);

        adapter->tnic.data = NULL;
	adapter->tnic.cb = NULL;

	adapter->tnic.registered = 0;

	spin_unlock_bh(&adapter->tnic.lock);	
}

/*
 * Return the registered TNIC adapter structure.
 *
 * Returns:
 *	The Tnic adapter if already registered else returns NULL.
 */
nx_tnic_adapter_t *nx_nic_get_tnic_adapter(struct net_device *netdev)
{
	struct unm_adapter_s *adapter = (unm_adapter *)netdev_priv(netdev);

        if (adapter) {
		return (adapter->tnic.data);
        }

        return (NULL);
}

/*
 * Gets the port number of the device on a netxen adaptor if it is a netxen
 * device.
 *
 * Parameters:
 *	dev	- The device for which the port is requested.
 *
 * Returns:
 *	-1	- If not netxen device.
 *	port number on the adapter if it is netxen device.
 */
int nx_nic_get_device_port(struct net_device *netdev)
{
	struct unm_adapter_s *adapter = (unm_adapter *)netdev_priv(netdev);

	if (netdev->do_ioctl != unm_nic_ioctl) {
		return (-1);
	}

        return ((int)physical_port[adapter->portnum]);
}

/*
 * Gets the ring context used by the device to talk to the card.
 *
 * Parameters:
 *	dev	- The device for which the ring context is requested.
 *
 * Returns:
 *	-EINVAL	- If not netxen device.
 *	Ring context number if it is netxen device.
 */
int nx_nic_get_device_ring_ctx(struct net_device *netdev)
{
        struct unm_adapter_s	*adapter;

	if (netdev->do_ioctl != unm_nic_ioctl) {
		return (-EINVAL);
	}

        adapter = netdev->priv;
	/*
	 * TODO: send the correct context.
	 */
        return (adapter->portnum);
}

/*
 * Gets the pointer to pci device if the device maps to a netxen device.
 */
struct pci_dev *nx_nic_get_pcidev(struct net_device *dev)
{
	if(dev->priv) {
		return ((struct unm_adapter_s *)(dev->priv))->pdev;
	}
	return (NULL);
}

/*
 * Send a TNIC message to the card.
 */
int nx_nic_send_tnic_msg(struct net_device *dev,
			 pegnet_cmd_desc_t *cmd_desc_arr, int nr_elements)
{
        uint32_t                producer;
        struct unm_cmd_buffer   *pbuf;
	cmdDescType0_t		*cmd_desc;
	int			i;
	int			rv;
	unm_adapter             *adapter = NULL;

	if (dev == NULL) {
                printk("%s: Device is NULL\n", __FUNCTION__);
                return (-1);
	}

	adapter = (unm_adapter *)dev->priv;

#if defined(DEBUG)
        if (adapter == NULL) {
                printk("%s: %s Adapter not initialized, cannot send request "
		       "to card\n", unm_nic_driver_name, __FUNCTION__);
                return (-1);
        }
#endif /* DEBUG */

	if (nr_elements > MAX_PENDING_DESC_BLOCK_SIZE || nr_elements == 0) {
		printk("%s: %s: Too many command descriptors in a request\n",
		       unm_nic_driver_name, __FUNCTION__);
		return (-EINVAL);
	}

	i = 0;

        spin_lock_bh(&adapter->tx_lock);

	/* check if space is available */
	if (unm_get_pending_cmd_desc_cnt(&adapter->pending_cmds) ||
	    ((adapter->cmdProducer + nr_elements) >=
	     ((adapter->lastCmdConsumer <= adapter->cmdProducer) ?
	      adapter->lastCmdConsumer + adapter->MaxTxDescCount :
	      adapter->lastCmdConsumer))) {

		rv = nx_queue_pend_cmd_desc_list(adapter, cmd_desc_arr,
						 nr_elements);
		spin_unlock_bh(&adapter->tx_lock);
		return (rv);
	}

/*         adapter->cmdProducer = get_index_range(adapter->cmdProducer, */
/* 					       MaxTxDescCount, nr_elements); */

	producer = adapter->cmdProducer;
	do {
		cmd_desc = (cmdDescType0_t *)&cmd_desc_arr[i];

		pbuf = &adapter->cmd_buf_arr[producer];
		pbuf->mss = 0;
		pbuf->totalLength = 0;
		pbuf->skb = NULL;
		pbuf->cmd = PEGNET_REQUEST;
		pbuf->fragCount = 0;
		pbuf->port = 0;

		/* adapter->ahw.cmdDescHead[producer] = *(cmdDescType0_t *)cmd_desc; */
		adapter->ahw.cmdDescHead[producer].word0 = cmd_desc->word0;
		adapter->ahw.cmdDescHead[producer].word1 = cmd_desc->word1;
		adapter->ahw.cmdDescHead[producer].word2 = cmd_desc->word2;
		adapter->ahw.cmdDescHead[producer].word3 = cmd_desc->word3;
		adapter->ahw.cmdDescHead[producer].word4 = cmd_desc->word4;
		adapter->ahw.cmdDescHead[producer].word5 = cmd_desc->word5;
		adapter->ahw.cmdDescHead[producer].word6 = cmd_desc->word6;
		adapter->ahw.cmdDescHead[producer].unused = cmd_desc->unused;

		producer = get_next_index(producer, adapter->MaxTxDescCount);
		i++;

	} while (i != nr_elements);

	adapter->cmdProducer = producer;

        /*
         * Check if this is the only thread trying to update the producer. If
         * it is then write the CRB. If not the last thread will do it.
         */
        if (adapter->num_threads == 0) {
#if 0
                ctx_msg  msg = {0};
#endif
                adapter->ahw.cmdProducer = adapter->cmdProducer;

                /* write producer index to start the xmit */

		read_lock(&adapter->adapter_lock);
		unm_nic_update_cmd_producer(adapter, adapter->cmdProducer);
		read_unlock(&adapter->adapter_lock);

#if 0
                msg.PegId = UNM_XMIT_PEG_DB_ID;
                msg.privId = 1;
                msg.Count = adapter->cmdProducer;
                msg.CtxId = 0;
                msg.Opcode = UNM_CMD_PRODUCER;
                read_lock(&adapter->adapter_lock);
                UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
				     DB_NORMALIZE(adapter,
						  UNM_CMD_PRODUCER_OFFSET));
                read_unlock(&adapter->adapter_lock);
#endif
        }

        spin_unlock_bh(&adapter->tx_lock);

        return (0);
}



#endif /* __VMKERNEL_MODULE__ */
#endif /* PEGNET_NIC */

#if defined(XGB_DEBUG)
static void
dump_skb(struct sk_buff *skb)
{
        printk("%s: SKB at %p\n", unm_nic_driver_name, skb);
        printk("    ->next: %p\n", skb->next);
        printk("    ->prev: %p\n", skb->prev);
        printk("    ->list: %p\n", skb->list);
        printk("    ->sk: %p\n",   skb->sk);
        printk("    ->stamp: %lx\n", skb->stamp);
        printk("    ->dev: %p\n",  skb->dev);
        //printk("    ->input_dev: %p\n", skb->input_dev);
        printk("    ->real_dev: %p\n", skb->real_dev);
        printk("    ->h.raw: %p\n",  skb->h.raw);

        printk("    ->len: %lx\n", skb->len);
        printk("    ->data_len:%lx\n", skb->data_len);
        printk("    ->mac_len:%lx\n", skb->mac_len);
        printk("    ->csum:%lx\n", skb->csum);

        printk("    ->head:%p\n", skb->head);
        printk("    ->data:%p\n", skb->data);
        printk("    ->tail:%p\n", skb->tail);
        printk("    ->end:%p\n", skb->end);

        return;
}

static int
skb_is_sane(struct sk_buff *skb)
{
        int ret = 1;

        if (skb_is_nonlinear(skb)) {
                printk("%s: got a non-linear SKB @%p data_len:%d back\n",
                       unm_nic_driver_name, skb, skb->data_len);
                return  0;
        }
#if 0
        if (skb->list) {
                return  0;
        }
#endif
        if (skb->dev != NULL || skb->next != NULL || skb->prev != NULL)
                return  0;
        if (skb->data == NULL || skb->data > skb->tail)
                return  0;
        if (*(unsigned long *)skb->head != 0xc0debabe) {
                printk("signature not found\n");
                return  0;
        }

        return ret;
}
#endif

/* Process Command status ring */
static int
unm_process_cmd_ring(unsigned long data)
{
        uint32_t               lastConsumer;
        uint32_t               consumer;
        unm_adapter           *adapter = (unm_adapter*)data;
        int                    count1 = 0;
        int                    count2 = 0;
        struct unm_cmd_buffer *buffer;
        struct pci_dev        *pdev;
        struct unm_skb_frag   *frag;
        uint32_t               i;
        struct sk_buff        *skb = NULL;
        int                    done;

        spin_lock(&adapter->tx_lock);
        lastConsumer = adapter->lastCmdConsumer;
        DPRINTK(1, INFO, "procesing xmit complete\n");
        /* we assume in this case that there is only one port and that is
         * port #1...changes need to be done in firmware to indicate port
         * number as part of the descriptor. This way we will be able to get
         * the netdev which is associated with that device.
         */
        /* Window = 1 */
        /*
	 *read_lock(&adapter->adapter_lock);
         *consumer = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
         *                                CRB_CMD_CONSUMER_OFFSET));
	 *read_unlock(&adapter->adapter_lock);
         */
        consumer = *(adapter->cmdConsumer);
       /* consumer = adapter->cmdPegConsumer; */

        if (lastConsumer == consumer){    /* Ring is empty    */
                DPRINTK(1, INFO, "lastConsumer %d == consumer %d\n",
                                    lastConsumer, consumer);
                spin_unlock(&adapter->tx_lock);
                return (1);
        }

        adapter->procCmdBufCounter++;
        //adapter->stats.process_xmit++;
        /*
         * Not needed - does not seem to be used anywhere.
         * adapter->cmdConsumer = consumer;
         */
        spin_unlock(&adapter->tx_lock);

        while ((lastConsumer != consumer) && (count1 < MAX_STATUS_HANDLE)) {
                buffer = &adapter->cmd_buf_arr[lastConsumer];
                pdev   = adapter->pdev;
                frag   = &buffer->fragArray[0];
                skb    = buffer->skb;
                if(skb && (cmpxchg(&buffer->skb, skb, 0) == skb)) {
                        pci_unmap_single(pdev, frag->dma, frag->length,
                                            PCI_DMA_TODEVICE);
                        for (i = 1; i < buffer->fragCount; i ++) {
                                DPRINTK(1, INFO, "getting fragment no %d\n",
                                                  i);
                                frag ++;    /* Get the next frag */
                                pci_unmap_page(pdev, frag->dma, frag->length,
                                                PCI_DMA_TODEVICE);
                        }
#if defined(UNM_LOOPBACK)
                        if (loopback_start && (unm_post_skb(adapter, skb)
                                    == 0)) {
                                continue;
                        } else {
#endif /* UNM_LOOPBACK */
#ifdef BOUNCE_BUFFERS
                        if(buffer->tx_bounce_buff == IS_BOUNCE)
                        {
                               buffer->tx_bounce_buff = 0;
                               tx_bounce_tail = buffer->bnext;
                        }
#endif

                        //adapter->stats.skbfreed++;
                        dev_kfree_skb_any(skb);
                        skb = NULL;
#if defined(UNM_LOOPBACK)
                        }
#endif
                } else if(adapter->procCmdBufCounter == 1) {
                        adapter->stats.txnullskb++;
                }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                if (unlikely(netif_queue_stopped(adapter->netdev)
                        && netif_carrier_ok(adapter->netdev))
                        && ((jiffies - adapter->netdev->trans_start) > adapter->netdev->watchdog_timeo))
                {
			//SCHEDULE_WORK(adapter->tx_timeout_task + adapter->portnum);
			SCHEDULE_WORK(adapter->tx_timeout_task);
                }
#endif

                lastConsumer = get_next_index (lastConsumer,
                                                adapter->MaxTxDescCount);
                count1++;
        }
        //adapter->stats.noxmitdone += count1;

        count2 = 0;
        while((lastConsumer != consumer) && (count2 < MAX_STATUS_HANDLE)) {
        	buffer = &adapter->cmd_buf_arr[lastConsumer];
                count2++;
                if (buffer->skb) {
                        break;
		} else {
                        lastConsumer = get_next_index(lastConsumer,
                                                      adapter->MaxTxDescCount);
		}
        }
        spin_lock(&adapter->tx_lock);
        if((--adapter->procCmdBufCounter) == 0) {
                adapter->lastCmdConsumer = lastConsumer;
        }
        if (count1 || count2) {
                        if(netif_queue_stopped(adapter->netdev)
                                && (adapter->state == PORT_UP) 
                                && (adapter->status & NETDEV_STATUS)) {
                                       netif_wake_queue(adapter->netdev);
                                       adapter->status &= ~NETDEV_STATUS;
                        }
        }

        /*
         * If everything is freed up to consumer then check if the ring is full
         * If the ring is full then check if more needs to be freed and
         * schedule the call back again.
         *
         * This happens when there are 2 CPUs. One could be freeing and the
         * other filling it. If the ring is full when we get out of here and
         * the card has already interrupted the host then the host can miss the
         * interrupt.
         *
         * There is still a possible race condition and the host could miss an
         * interrupt. The card has to take care of this.
         */
#if 0
        if (adapter->lastCmdConsumer == consumer &&
            (((adapter->cmdProducer + 1) %
              adapter->MaxTxDescCount) == adapter->lastCmdConsumer)) {
                /*
                 *read_lock(&adapter->adapter_lock);
                 *consumer = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
                 *                                             CRB_CMD_CONSUMER_OFFSET));
                 *read_unlock(&adapter->adapter_lock);
                 */
                consumer = *(adapter->cmdConsumer);
        }
#else
                consumer = *(adapter->cmdConsumer);
#endif
        done = (adapter->lastCmdConsumer == consumer);

        spin_unlock(&adapter->tx_lock);
        DPRINTK(1, INFO, "last consumer is %d in %s\n", lastConsumer,
                         __FUNCTION__);
        return (done);
}

#if defined(UNM_LOOPBACK)
/* FIXME: this causes ip checksum errors on smartbits, at times */
static int
unm_post_skb(struct unm_adapter_s *adapter, struct sk_buff *skb)
{
       /*
        * TODO
        * This routine posts only to context 0. If we need a more
        * sophisticated loopback,  this needs to be fixed.
        */
        const uint                ctx      = 0;
        unm_recv_context_t       *recv_ctx = &(adapter->recv_ctx[ctx]);
        uint                      producer = recv_ctx->rcvProducer;
        struct pci_dev           *pdev     = adapter->ahw.pdev;
        struct _hardware_context *hw       = &(adapter->ahw);
        rcvDesc_t                *pdesc;
        struct unm_rx_buffer     *buffer;

	if ((skb->end - skb->data) < MAX_RX_BUFFER_LENGTH ||
                (skb_shinfo(skb)->nr_frags > 0)) {
                return -1;
        }
        skb->data = skb->head;
        skb->tail = skb->head;
        skb->len = 0;
        skb->cloned = 0;
        skb->data_len = 0;

        buffer   = &rcv_desc->rx_buf_arr[rcv_desc->begin_alloc];

        skb_reserve(skb, IP_ALIGNMENT_BYTES);

        buffer->skb = skb;
        buffer->state  = UNM_BUFFER_BUSY;
        buffer->dma = pci_map_single(pdev, skb->data, RX_DMA_MAP_LEN,
                                         PCI_DMA_FROMDEVICE);

        pdesc = &adapter->recv_ctx[ctx].rcvDescHead[producer];
        pdesc->referenceHandle = buffer->refHandle;
        pdesc->bufferLength    = MAX_RX_BUFFER_LENGTH - IP_ALIGNMENT_BYTES;
        pdesc->AddrBuffer      = buffer->dma;

        producer = get_next_index(producer, adapter->MaxRxDescCount);
        recv_ctx->freeRxCount--;
        recv_ctx->pendingRxCount++;
        recv_ctx->rcvProducer = producer;
        /* Window = 1 */
	read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32((producer-1)&(adapter->MaxRxDescCount-1),
                CRB_NORMALIZE(adapter,
                        recv_crb_registers[ctx].CRB_RCV_PRODUCER_OFFSET));
	read_unlock(&adapter->adapter_lock);

        //adapter->stats.goodskbposts++;
        rcv_desc->begin_alloc = get_next_index(rcv_desc->begin_alloc,
                                                rcv_desc->MaxRxDescCount);

        return 0;
}
#endif /* UNM_LOOPBACK */
/**
 * unm_post_rx_buffers puts buffer in the Phantom memory
 **/
static void
unm_post_rx_buffers(struct unm_adapter_s *adapter, uint32_t ctx,
                        uint32_t ringid) {
        struct pci_dev           *pdev = adapter->ahw.pdev;
        struct sk_buff           *skb;
        unm_recv_context_t       *recv_ctx = &(adapter->recv_ctx[ctx]);
        unm_rcv_desc_ctx_t       *rcv_desc = NULL;
        uint                      producer;
        rcvDesc_t                *pdesc;
        struct unm_rx_buffer     *buffer;
        int                       count = 0;
        int                       index = 0;
        ctx_msg                   msg = {0};
        dma_addr_t            dma;
#ifdef BOUNCE_BUFFERS
        unsigned long         bounce_flags;
        struct bounce_index   tmpbi;
#endif

        //adapter->stats.post_called++;
        rcv_desc     = &recv_ctx->rcv_desc[ringid];

        producer = rcv_desc->producer;
        index = rcv_desc->begin_alloc;
        buffer   = &rcv_desc->rx_buf_arr[index];
        /* We can start writing rx descriptors into the phantom memory. */
        while (!TAILQ_EMPTY(&rcv_desc->free_rxbuf_list)) {
                buffer = TAILQ_FIRST(&rcv_desc->free_rxbuf_list);
#ifdef __VMKERNEL_MODULE__
                skb = alloc_skb(rcv_desc->skb_size, GFP_ATOMIC);
#else /* __VMKERNEL_MODULE__ */
                skb = dev_alloc_skb(rcv_desc->skb_size);
#endif /* ! __VMKERNEL_MODULE__ */
                if(unlikely(!skb)) {
                /*
                 * TODO
                 * We need to schedule the posting of buffers to the pegs.
                 */
                        rcv_desc->begin_alloc = index;
                        DPRINTK(1, ERR, "unm_post_rx_buffers: "
                                        " allocated only %d buffers\n", count);
                        break;
                }

                TAILQ_REMOVE(&rcv_desc->free_rxbuf_list, buffer, link);
                count++ ;    /* now there should be no failure */
                pdesc = &rcv_desc->desc_head[producer];

#if defined(XGB_DEBUG)
                *(unsigned long *)(skb->head) = 0xc0debabe;
                if (skb_is_nonlinear(skb)) {
                        printk("Allocated SKB @%p is nonlinear\n");
                }
#endif
                skb_reserve(skb, IP_ALIGNMENT_BYTES);
#ifdef BOUNCE_BUFFERS
                buffer->rx_bounce_buff = 0;
#endif
                /* This will be setup when we receive the
                 * buffer after it has been filled  FSL  TBD TBD
                 * skb->dev = netdev;
                 */
                dma = pci_map_single(pdev, skb->data,
                                                rcv_desc->dma_size,
                                                PCI_DMA_FROMDEVICE);
#ifdef BOUNCE_BUFFERS
                if (dma + rcv_desc->dma_size > unm_dma_mask) {
                        pci_unmap_single(pdev, dma, rcv_desc->dma_size,
                                PCI_DMA_FROMDEVICE);
                        spin_lock_irqsave(&rx_bounce_lock, bounce_flags);
                        if (unm_get_next_bounce_buffer(&rx_bounce_head, &rx_bounce_tail,
                                &tmpbi, rcv_desc->dma_size)) {

                                spin_unlock_irqrestore(&rx_bounce_lock, bounce_flags);
                                dev_kfree_skb_any(skb);
                                skb = NULL;
                                buffer->skb = NULL;
                                buffer->state  = UNM_BUFFER_FREE;
                                count--;
                                break;
                        }
                        spin_unlock_irqrestore(&rx_bounce_lock, bounce_flags);

                        buffer->rx_bounce_buff = IS_BOUNCE;
                        buffer->bnext = rx_bounce_head;
                        buffer->bounce_ptr  = rx_bounce_ptr[tmpbi.buffer_index] +
                                                tmpbi.buffer_offset;
                        dma = pci_map_single(pdev, buffer->bounce_ptr,
                                rcv_desc->dma_size, PCI_DMA_FROMDEVICE);

                }
#endif
                pdesc->AddrBuffer      = dma;
                buffer->skb = skb;
                buffer->state  = UNM_BUFFER_BUSY;
                buffer->dma = dma;


                /* make a rcv descriptor  */
                pdesc->referenceHandle = buffer->refHandle;
                pdesc->bufferLength    = rcv_desc->dma_size;

                DPRINTK(1, INFO, "done writing descripter\n");
                producer = get_next_index(producer, rcv_desc->MaxRxDescCount);
                index = get_next_index(index, rcv_desc->MaxRxDescCount);
                buffer   = &rcv_desc->rx_buf_arr[index];
        }

        /* if we did allocate buffers, then write the count to Phantom */
        if (count) {
                rcv_desc->begin_alloc = index;
                rcv_desc->rcv_pending    += count;
                //adapter->stats.lastposted = count;
                //adapter->stats.posted    += count;
                rcv_desc->producer        = producer;
                if (rcv_desc->rcv_free >= 32) {
                        rcv_desc->rcv_free        = 0;
                        /* Window = 1 */

			  read_lock(&adapter->adapter_lock);
                          UNM_NIC_PCI_WRITE_32((producer-1)&
                                         (rcv_desc->MaxRxDescCount-1),
                                          CRB_NORMALIZE(adapter,
					    recv_crb_registers[adapter->portnum]
						.rcv_desc_crb[ringid].CRB_RCV_PRODUCER_OFFSET));
                          read_unlock(&adapter->adapter_lock);

                        /*
                         * Write a doorbell msg to tell phanmon of change in
                         * receive ring producer
                         */
                        msg.PegId = UNM_RCV_PEG_DB_ID;
                        msg.privId = 1;
                        msg.Count = (producer-1) & (rcv_desc->MaxRxDescCount-1);
			msg.CtxId = adapter->portnum;
                        msg.Opcode = UNM_RCV_PRODUCER(ringid);
			read_lock(&adapter->adapter_lock);
                        UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                                DB_NORMALIZE(adapter, UNM_RCV_PRODUCER_OFFSET));
			read_unlock(&adapter->adapter_lock);
                        printk("dbmsg: 0x%08x\n", *((U32 *)&msg));
                }
        }
}
/**
 * unm_post_rx_buffers puts buffer in the Phantom memory
 **/
static void
unm_post_rx_buffers_nodb(struct unm_adapter_s *adapter, uint32_t ctx,
                        uint32_t ringid) {
        struct pci_dev           *pdev = adapter->ahw.pdev;
        struct sk_buff           *skb;
        unm_recv_context_t       *recv_ctx = &(adapter->recv_ctx[ctx]);
        unm_rcv_desc_ctx_t       *rcv_desc = NULL;
        uint                      producer;
        rcvDesc_t                *pdesc;
        struct unm_rx_buffer     *buffer;
        int                       count = 0;
        int                       index = 0;
	dma_addr_t                dma;
#ifdef BOUNCE_BUFFERS
        unsigned long         bounce_flags;
        struct bounce_index   tmpbi;
#endif

#if 0
        ctx_msg                   msg = {0};
#endif

        //adapter->stats.post_called++;
        rcv_desc     = &recv_ctx->rcv_desc[ringid];

        producer = rcv_desc->producer;
        index = rcv_desc->begin_alloc;
        buffer   = &rcv_desc->rx_buf_arr[index];
        /* We can start writing rx descriptors into the phantom memory. */
        while (!TAILQ_EMPTY(&rcv_desc->free_rxbuf_list)) {
                buffer = TAILQ_FIRST(&rcv_desc->free_rxbuf_list);
#ifdef __VMKERNEL_MODULE__
                skb = alloc_skb(rcv_desc->skb_size, GFP_ATOMIC);
#else /* __VMKERNEL_MODULE__ */
                skb = dev_alloc_skb(rcv_desc->skb_size);
#endif /* ! __VMKERNEL_MODULE__ */
                if(unlikely(!skb)) {
                /*
                 * TODO
                 * We need to schedule the posting of buffers to the pegs.
                 */
                        rcv_desc->begin_alloc = index;
                        DPRINTK(1, ERR, "unm_post_rx_buffers: "
                                        " allocated only %d buffers\n", count);
                        break;
                }

                TAILQ_REMOVE(&rcv_desc->free_rxbuf_list, buffer, link);
                count++ ;    /* now there should be no failure */
                pdesc = &rcv_desc->desc_head[producer];

#if defined(XGB_DEBUG)
                *(unsigned long *)(skb->head) = 0xc0debabe;
                if (skb_is_nonlinear(skb)) {
                        printk("Allocated SKB @%p is nonlinear\n");
                }
#endif
                skb_reserve(skb, IP_ALIGNMENT_BYTES);
#ifdef BOUNCE_BUFFERS
                buffer->rx_bounce_buff = 0;
#endif
                /* This will be setup when we receive the
                * buffer after it has been filled  FSL  TBD TBD
                * skb->dev = netdev;
                */
                dma = pci_map_single(pdev, skb->data,
                                                rcv_desc->dma_size,
                                                PCI_DMA_FROMDEVICE);
#ifdef BOUNCE_BUFFERS
                if (dma + rcv_desc->dma_size > unm_dma_mask) {
                        pci_unmap_single(pdev, dma, rcv_desc->dma_size,
                                PCI_DMA_FROMDEVICE);
                        spin_lock_irqsave(&rx_bounce_lock, bounce_flags);
                        if (unm_get_next_bounce_buffer(&rx_bounce_head, &rx_bounce_tail,
                                &tmpbi, rcv_desc->dma_size)) {

                                spin_unlock_irqrestore(&rx_bounce_lock, bounce_flags);
                                dev_kfree_skb_any(skb);
                                skb = NULL;
                                buffer->skb = NULL;
                                buffer->state  = UNM_BUFFER_FREE;
                                count--;
                                break;
                        }
                       spin_unlock_irqrestore(&rx_bounce_lock, bounce_flags);

                        buffer->rx_bounce_buff = IS_BOUNCE;
                        buffer->bnext = rx_bounce_head;
                        buffer->bounce_ptr  = rx_bounce_ptr[tmpbi.buffer_index] +
                                                tmpbi.buffer_offset;
                        dma = pci_map_single(pdev, buffer->bounce_ptr,
                                rcv_desc->dma_size, PCI_DMA_FROMDEVICE);

                }
#endif
                pdesc->AddrBuffer      = dma;
                buffer->skb = skb;
                buffer->state  = UNM_BUFFER_BUSY;
                buffer->dma = dma;

                /* make a rcv descriptor  */
                pdesc->referenceHandle = buffer->refHandle;
                pdesc->bufferLength    = rcv_desc->dma_size;

                DPRINTK(1, INFO, "done writing descripter\n");
                producer = get_next_index(producer, rcv_desc->MaxRxDescCount);
                index = get_next_index(index, rcv_desc->MaxRxDescCount);
                buffer   = &rcv_desc->rx_buf_arr[index];
        }

        /* if we did allocate buffers, then write the count to Phantom */
        if (count) {
                rcv_desc->begin_alloc = index;
                rcv_desc->rcv_pending    += count;
                //adapter->stats.lastposted = count;
                //adapter->stats.posted    += count;
                rcv_desc->producer        = producer;
                if (rcv_desc->rcv_free >= 32) {
                        rcv_desc->rcv_free        = 0;
                        /* Window = 1 */

			            read_lock(&adapter->adapter_lock);
                        UNM_NIC_PCI_WRITE_32((producer-1)&
                                             (rcv_desc->MaxRxDescCount-1),
                                             CRB_NORMALIZE(adapter,
                                             recv_crb_registers[adapter->portnum].rcv_desc_crb[ringid].CRB_RCV_PRODUCER_OFFSET));
                                             //recv_crb_registers[0].rcv_desc_crb[ringid].CRB_RCV_PRODUCER_OFFSET));
                                          //rcv_desc_crb->CRB_RCV_PRODUCER_OFFSET));
                        read_unlock(&adapter->adapter_lock);

#if 0
                        /*
                         * Write a doorbell msg to tell phanmon of change in
                         * receive ring producer
                         */
                        msg.PegId = UNM_RCV_PEG_DB_ID;
                        msg.privId = 1;
                        msg.Count = (producer-1) & (rcv_desc->MaxRxDescCount-1);
                        msg.CtxId = 0;
                        msg.Opcode = UNM_RCV_PRODUCER(ringid);
			read_lock(&adapter->adapter_lock);
                        UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                                DB_NORMALIZE(adapter, UNM_RCV_PRODUCER_OFFSET));
			read_unlock(&adapter->adapter_lock);
#endif
                }
        }
}

unsigned char XMIT_DATA[] = {0x50, 0xda, 0x2e, 0xfa, 0x77, 0x00, 0x02,
                 0x2d, 0x8a, 0xa1, 0xde, 0x08, 0x00, 0x45, 0x00,
                 0x00, 0x32, 0x00, 0x00, 0x40, 0x00, 0x40, 0x11,
                 0xc2, 0x95, 0xc0, 0xa8, 0x7b, 0x64, 0xc0, 0xa8,
                 0x7b, 0x70, 0x80, 0x09, 0x1c, 0x0d, 0x00, 0x1e,
                 0xb1, 0x5a, 0x1f, 0x00, 0x00, 0x00, 0x58, 0x58,
                 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58,
                 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58
                };

unsigned int tx_stop = 0;
unm_send_test_t tx_args;

static int
tx_test(struct net_device *netdev, unm_send_test_t *arg)
{
        unsigned int    count = arg->count;
        struct sk_buff *skb;
        unsigned int    len;
        unsigned char  *pkt_data;
        unsigned int    sent = 0;

        if (tx_user_packet_data == NULL) {
                pkt_data = XMIT_DATA;
                len = sizeof(XMIT_DATA);
        } else {
                pkt_data = tx_user_packet_data;
                len = tx_user_packet_length;
        }

        while (tx_stop == 0) {
                skb = alloc_skb(len, GFP_KERNEL);
                if (!skb) {
                        return -ENOMEM;
                }

                                *((uint16_t *)(pkt_data + 14)) = (uint16_t)sent;
                memcpy(skb->data, pkt_data, len);
                skb_put(skb, len);
                unm_nic_xmit_frame(skb, netdev);
                ++sent;
                if (arg->count && --count == 0) {
                        break;
                }

                schedule();
                if (arg->ifg) {
                        if (arg->ifg < 1024) {
                                udelay(arg->ifg);
                        } else if (arg->ifg < 5000) {
                                mdelay(arg->ifg >> 10);
                        } else {
                                mdelay(5);
                        }
                }
                if (signal_pending(current)) {
                        break;
                }
        }
        printk("unm packet generator sent %d bytes %d packets %d bytes each\n",
                            sent*len, sent, len);
        return 0;
}

static int
unm_send_test(struct net_device *netdev, void *ptr)
{
        unm_send_test_t args;

        if (copy_from_user(&args, ptr, sizeof(args))) {
                return -EFAULT;
        }
        switch (args.cmd) {
        case UNM_TX_START:
                if (tx_args.cmd == UNM_TX_SET_PARAM) {
                        if (args.ifg == 0) {
                                args.ifg = tx_args.ifg;
                        }
                        if (args.count == 0) {
                                args.count = tx_args.count;
                        }
                }
                tx_stop = 0;
                return tx_test(netdev, &args);
                break;

        case UNM_TX_STOP:
                tx_stop = 1;
                break;

        case UNM_TX_SET_PARAM:
                tx_args = args;
                break;

        case UNM_TX_SET_PACKET:
                if (tx_user_packet_data) {
                        kfree(tx_user_packet_data);
                }
                tx_user_packet_data = kmalloc(args.count, GFP_KERNEL);
                if (tx_user_packet_data == NULL) {
                        return -ENOMEM;
                }
                if (copy_from_user(tx_user_packet_data,
                        (char *)(uptr_t)args.ifg, args.count)) {
                        kfree(tx_user_packet_data);
                        tx_user_packet_data = NULL;
                        return -EFAULT;
                }
                tx_user_packet_length = args.count;
                break;

        case UNM_LOOPBACK_START:
                loopback_start = 1;
                netdev->hard_start_xmit = &unm_loopback_xmit_frame;
                break;

        case UNM_LOOPBACK_STOP:
                loopback_start = 0;
                netdev->hard_start_xmit = &unm_nic_xmit_frame;
                break;
        default:
                return -EINVAL;
        }
        return 0;
}

static int
unm_irq_test(unm_adapter *adapter)
{
        uint64_t         int_cnt, int_cnt1;
        long                 templ, templ1;

        if ((UNM_CRB_READ_VAL_ADAPTER(UNM_PCIX_PS_REG(PCIX_INT_MASK), adapter)&0x7ff) != 0x7ff) {
                /* Generate Interrupt */
                int_cnt = adapter->stats.ints;
                if (adapter->flags & UNM_NIC_MSI_ENABLED) {
                        /* setup MSI */
                        templ1 = UNM_CRB_READ_VAL_ADAPTER((UNM_CRB_PCIX_MD + 0xc4), adapter);
                        UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PCIX_MD+PCIX_PS_OP_ADDR_LO,
                                (templ1 & 0xffff0000), adapter);
                        templ1 = UNM_CRB_READ_VAL_ADAPTER((UNM_CRB_PCIX_MD + 0xc8), adapter);
                        UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PCIX_MD+PCIX_PS_OP_ADDR_HI, templ1, adapter);
                        templ1 = UNM_CRB_READ_VAL_ADAPTER((UNM_CRB_PCIX_MD + 0xcc), adapter);
                        /* send MSI */
                        templ = UNM_CRB_READ_VAL_ADAPTER((UNM_CRB_PCIX_MD + 0xc4), adapter);
                        UNM_CRB_WRITELIT_ADAPTER(UNM_CRB_PCIX_MD+PCIX_PS_MEM_SPACE+(templ & 0xffff),
                                (uint16_t)templ1, adapter);
                } else {
                        if ((UNM_CRB_READ_VAL_ADAPTER(UNM_PCIX_PS_REG(PCIX_INT_VECTOR), adapter)
                                  & 0x80) == 0x80) {
                        /* for some reason, the INT is stuck, so reset it */
                                UNM_CRB_WRITELIT_ADAPTER(UNM_PCIX_PS_REG(PCIX_TARGET_STATUS), -1, adapter);
                        }
                /* Now generate the interrupt */
                        UNM_CRB_WRITELIT_ADAPTER( UNM_PCIX_PS_REG(PCIX_MSI_F0), 0x100, adapter);
                }
                mdelay(50); /* let interrupt be processed */
                int_cnt1 = adapter->stats.ints;
                return ((int_cnt1 == int_cnt) ? (-1) : (0));
        }
        else
                return (-1);
}

static int
unm_loopback_test(struct net_device *netdev, int fint, void *ptr)
{
        int			ii, ret;
        unsigned char		*data;
        unm_send_test_t		args;
        struct unm_adapter_s	*adapter;

	adapter = (struct unm_adapter_s *)netdev->priv;
	if((fint) && (adapter->ahw.boardcfg.board_type == UNM_BRDTYPE_P2_SB31_10G_IMEZ ||
		      adapter->ahw.boardcfg.board_type == UNM_BRDTYPE_P2_SB31_10G_HMEZ))
		return (-LB_NOT_SUPPORTED);

        if ((ret = copy_from_user(&args, ptr, sizeof(args)))) {
                return (-LB_UCOPY_PARAM_ERR);
        }

        if (tx_user_packet_data != NULL) {
                kfree(tx_user_packet_data);
                tx_user_packet_data = NULL;
        }

	if (args.count != 0) { /* user data */
		tx_user_packet_data = kmalloc(args.count + 16, GFP_KERNEL);
		if (tx_user_packet_data == NULL)
			return (-LB_NOMEM_ERR);
		memset(tx_user_packet_data, 0xFF, 16);
		if (copy_from_user(tx_user_packet_data+16,
				(char *)(uptr_t)args.ifg, args.count)) {
			kfree(tx_user_packet_data);
			tx_user_packet_data = NULL;
			return (-LB_UCOPY_DATA_ERR);
		}
		tx_user_packet_length = args.count + 16;
		rx_datalen = (tx_user_packet_length - 14) * 16;
	} else {  /* use local data */
		tx_user_packet_data = kmalloc(sizeof(XMIT_DATA) + 16, GFP_KERNEL);
		if (tx_user_packet_data == NULL)
			return (-LB_NOMEM_ERR);
		memset(tx_user_packet_data, 0xFF, 14);
		memcpy(tx_user_packet_data + 16, XMIT_DATA, sizeof(XMIT_DATA));
		tx_user_packet_length = sizeof(XMIT_DATA) + 16;
		rx_datalen = (sizeof(XMIT_DATA) + 2) * 16;
	}
	if ((rx_user_packet_data = kmalloc(rx_datalen, GFP_KERNEL)) == NULL){
		ret = -LB_NOMEM_ERR;
		goto end;
	}
	rx_user_pos = 0;

        if (fint == 1) {
                loopback_start = 1;
                xge_loopback(adapter, 1);
        }

        tx_stop = 0;
        capture_input = 1;
        captured = 0;
        args.count = 16;
        if ((ret = tx_test(netdev, &args)) != 0) {
                ret = -LB_TX_NOSKB_ERR;
		capture_input = 0;
                goto end;
        }

        mdelay(1000); /* wait for data to come back */
        capture_input = 0;

        if (rx_user_pos != rx_datalen) {
                ret = -LB_SHORT_DATA_ERR;
                goto end;
        }

        data = rx_user_packet_data;
        /* check received bytes against tx_user_packet_data */
        for (ii = 0; ii < 16; ++ii) {
                if (*((uint16_t *)data) != (uint16_t)ii) {
                        ret = -LB_SEQUENCE_ERR;
                        goto end;
                }
                if (memcmp(tx_user_packet_data + 16, data + 2,
                                 tx_user_packet_length - 16)
                         != 0) {
                        ret = -LB_DATA_ERR;
                        goto end;
                }
                data += (tx_user_packet_length - 14);
        }
        ret = LB_TEST_OK;

end:
        tx_stop = 1;
        if (fint == 1) {
                xge_loopback(adapter, 0);
                loopback_start = 0;
        }
        if (tx_user_packet_data != NULL) {
                kfree(tx_user_packet_data);
                tx_user_packet_data = NULL;
        }
        kfree(rx_user_packet_data);
        rx_user_packet_data = NULL;
        return(ret);
}


#ifdef UNM_IDS_ENABLED
typedef struct {
        unsigned long ring_buffer;
        unsigned long ring_buffer_phys;
        unsigned int ring_buffer_size;
} ring_buffer_t;

ring_buffer_t rbuf[MAX_NUM_PORTS] = { {0, 0, 0} }; /* FIXME */
unsigned int ring_buffer_min = RING_BUFFER_MIN;

static void
unm_nic_unregister_iface(int idx)
{
        unsigned long ring_buffer = rbuf[idx].ring_buffer;
        unsigned long ring_buffer_size = rbuf[idx].ring_buffer_size;
        int i;

        if (rbuf[idx].ring_buffer == 0) {
               /*
                * No map here.
                */
                return;
        }
#ifdef P1
        for (i = 0; i < ring_buffer_size/PAGE_SIZE; ++i) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
                mem_map_unreserve(MAP_NR(ring_buffer + i*PAGE_SIZE));
#else
                ClearPageReserved(virt_to_page(ring_buffer + i*PAGE_SIZE));
#endif
        }
#endif

        free_pages(ring_buffer, get_order(ring_buffer_size));
        rbuf[idx].ring_buffer = 0;
        rbuf[idx].ring_buffer_size = 0;
}

static void
unm_nic_unregister_all(void)
{
        int i;

        for (i = 0; i < MAX_NUM_PORTS; ++i) {
                unm_nic_unregister_iface(i);
        }
}
#endif

#if defined(UNM_IP_FILTER)
/*
 * unm_ip_filter() - provides the hooks to enable the user application to
 * set ip filters on the UNM card.
 */

static int
unm_ip_filter(struct unm_adapter_s *adapter, void *ptr)
{
        uint32_t producer;
        uint32_t next;
        controlCmdDesc_t controldesc;
        cmdDescType0_t *hwdesc;
        unm_ip_filter_request_t req;

	if (copy_from_user(&req, ptr, sizeof(req))) {
                return -EFAULT;
        }

        switch (req.cmd) {
        case UNM_IP_ADDR_ADD:
        case UNM_IP_ADDR_DEL:
                        if (req.request.count <= 0 ||
                                req.request.count > UNM_IP_ADDR_MAX_COUNT) {
                                return -EINVAL;
                        }
                        memcpy(controldesc.ip_addr, &req.request,
                                   sizeof(req.request.count) +
                                   req.request.count*
                                   sizeof(req.request.ip_addr[0]));
                        controldesc.count = req.request.count;

                        break;
        case UNM_IP_ADDR_SHOW:
        case UNM_IP_ADDR_CLEAR:
                        controldesc.count = 0;
                        break;

        default:
                        return -EINVAL;
        }

        /*
         * Now that we got the command, send it out.
         */

        controldesc.opcode = UNM_CONTROL_OP;
        controldesc.cmd = req.cmd;
        spin_lock_bh(&adapter->tx_lock);
        producer = adapter->cmdProducer;
        if ((next = get_next_index(producer, adapter->MaxTxDescCount)) ==
                                    adapter->lastCmdConsumer) {
                spin_unlock_bh(&adapter->tx_lock);
                return -EAGAIN;
        }
        hwdesc = &(adapter->ahw.cmdDescHead[producer]);
        memcpy(hwdesc, &controldesc, sizeof(controldesc));
        adapter->cmdProducer = next;
        adapter->ahw.cmdProducer = next;
        /* Window = 1 */

	read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(next, CRB_NORMALIZE(adapter,
                                                CRB_CMD_PRODUCER_OFFSET));
	read_unlock(&adapter->adapter_lock);

#if 0
        msg.PegId = UNM_XMIT_PEG_DB_ID;
        msg.privId = 1;
        msg.Count = next;
        msg.CtxId = 0;
        msg.Opcode = UNM_CMD_PRODUCER;
        read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(*((__uint32_t *) &msg),
                DB_NORMALIZE(adapter, UNM_CMD_PRODUCER_OFFSET));
        read_unlock(&adapter->adapter_lock);
#endif

        spin_unlock_bh(&adapter->tx_lock);

        return 0;
}
#endif /* UNM_IP_FILTER */

static int
unm_link_test(unm_adapter *adapter)
{
    long rv;

        rv = xge_link_status (adapter);
        return((rv == XG_LINK_DOWN) ? (-1) : (0));
}

int unmtest_pegstuck(unm_crbword_t addr, U64 reg, int loop, struct unm_adapter_s *adapter)
{
        int                         i;
        unm_crbword_t         temp;

        for (i = 0; i < loop; ++i) {
                UNM_CRB_READ_CHECK_ADAPTER(reg, &temp, adapter);
                if (temp != addr)
                        return (0);
        }
        return (-1);
}

#define NIC_NUMPEGS                3

static int
unm_hw_test(unm_adapter *adapter)
{
        unm_crbword_t         temp;
        int                                ii, rc = HW_TEST_OK;
        uint64_t                 base, address;

    /* DMA Status*/
        UNM_CRB_READ_CHECK_ADAPTER(UNM_DMA_COMMAND(0), &temp, adapter);
        if ((temp & 0x3) == 0x3) {
                rc = -HW_DMA_BZ_0;
                goto done;
        }
        UNM_CRB_READ_CHECK_ADAPTER(UNM_DMA_COMMAND(1), &temp, adapter);
        if ((temp & 0x3) == 0x3) {
                rc = -HW_DMA_BZ_1;
                goto done;
        }
        UNM_CRB_READ_CHECK_ADAPTER(UNM_DMA_COMMAND(2), &temp, adapter);
        if ((temp & 0x3) == 0x3) {
                rc = -HW_DMA_BZ_2;
                goto done;
        }
            UNM_CRB_READ_CHECK_ADAPTER(UNM_DMA_COMMAND(3), &temp, adapter);
        if ((temp & 0x3) == 0x3) {
                rc = -HW_DMA_BZ_3;
                goto done;
        }

        /* SRE Status */
        UNM_CRB_READ_CHECK_ADAPTER(UNM_SRE_PBI_ACTIVE_STATUS, &temp, adapter);
        if ((temp & 0x1) == 0) {
                rc = -HW_SRE_PBI_HALT;
                goto done;
        }
        UNM_CRB_READ_CHECK_ADAPTER(UNM_SRE_L1RE_CTL, &temp, adapter);
    if ((temp & 0x20000000) != 0) {
                rc = -HW_SRE_L1IPQ;
                goto done;
        }
        UNM_CRB_READ_CHECK_ADAPTER(UNM_SRE_L2RE_CTL, &temp, adapter);
    if ((temp & 0x20000000) != 0) {
                rc = -HW_SRE_L2IFQ;
                goto done;
        }

        UNM_CRB_READ_CHECK_ADAPTER(UNM_SRE_INT_STATUS, &temp, adapter);
        if ((temp & 0xc0ff) != 0) {
                if ((temp & 0x1) != 0) {
                        rc = -HW_PQ_W_PAUSE;
                        goto done;
                }
                if ((temp & 0x2) != 0) {
                        rc = -HW_PQ_W_FULL;
                        goto done;
                }
                if ((temp & 0x4) != 0) {
                        rc = -HW_IFQ_W_PAUSE;
                        goto done;
                }
                if ((temp & 0x8) != 0) {
                        rc = -HW_IFQ_W_FULL;
                        goto done;
                }
                if ((temp & 0x10) != 0) {
                        rc = -HW_MEN_BP_TOUT;
                        goto done;
                }
                if ((temp & 0x20) != 0) {
                        rc = -HW_DOWN_BP_TOUT;
                        goto done;
                }
                if ((temp & 0x40) != 0) {
                        rc = -HW_FBUFF_POOL_WM;
                        goto done;
                }
                if ((temp & 0x80) != 0) {
                        rc = -HW_PBUF_ERR;
                        goto done;
                }
                if ((temp & 0x4000) != 0) {
                        rc = -HW_FM_MSG_HDR;
                        goto done;
                }
                if ((temp & 0x8000) != 0) {
                        rc = -HW_FM_MSG;
                        goto done;
                }
        }

        UNM_CRB_READ_CHECK_ADAPTER(UNM_SRE_INT_STATUS, &temp, adapter);

        if ((temp & 0x3f00) != 0) {
                if ((temp & 0x100) != 0) {
                        rc = -HW_EPG_MSG_BUF;
                        goto done;
                }
                if ((temp & 0x200) != 0) {
                        rc = -HW_EPG_QREAD_TOUT;
                        goto done;
                }
                if ((temp & 0x400) != 0) {
                        rc = -HW_EPG_QWRITE_TOUT;
                        goto done;
                }
                if ((temp & 0x800) != 0) {
                        rc = -HW_EPG_CQ_W_FULL;
                        goto done;
                }
                if ((temp & 0x1000) != 0) {
                        rc = -HW_EPG_MSG_CHKSM;
                        goto done;
                }
                if ((temp & 0x2000) != 0) {
                        rc = -HW_EPG_MTLQ_TOUT;
                        goto done;
                }
        }

        /* Pegs */

        for (ii = 0; ii < NIC_NUMPEGS; ++ii) {
            base = PEG_NETWORK_BASE(ii);
                address = base | CRB_REG_EX_PC;
                UNM_CRB_READ_CHECK_ADAPTER(address, &temp, adapter);
                rc = unmtest_pegstuck(temp, address, PEG_LOOP, adapter);
                if (rc != 0) {
                        rc = -(HW_PEG0 + ii);
                        goto done;
                }
        }

done:
        return (rc);
}

static int
unm_cis_test(unm_adapter *adapter)
{
        //unm_crbword_t         temp, temp1;
        int                         ret = CIS_TEST_OK;
        //ctx_msg         msg = {0};

        /* Read CAM RAM CRB registers: producer and consumer index for
         command descriptors. Make sure that P==C or P-C=n0,
          where n is acceptable number (what range?)*/
        /*
        UNM_CRB_READ_CHECK_ADAPTER(CRB_CMD_PRODUCER_OFFSET, &temp, adapter);
        UNM_CRB_READ_CHECK_ADAPTER(CRB_CMD_CONSUMER_OFFSET, &temp1, adapter);
        */
#if 0
        read_lock(&adapter->adapter_lock);
        *((uint32_t *) &msg) =
                UNM_NIC_PCI_READ_32(DB_NORMALIZE(adapter,
                            UNM_CMD_PRODUCER_OFFSET(0)));
        read_unlock(&adapter->adapter_lock);
        temp = msg.Count;

        //temp1 = adapter->ctxDesc->CMD_CONSUMER_OFFSET;
        temp1 = *(adapter->cmdConsumer);

        if (DIDX_DIF(temp, temp1) > CIS_WATERMARK)
                ret = -CIS_WMARK;
#endif
        return (ret);
}

static int
unm_cr_test(unm_adapter *adapter)
{

        unm_crbword_t         temp;
        int                         ret = CR_TEST_OK;

        UNM_CRB_READ_CHECK_ADAPTER(UNM_CRB_PCIE, &temp, adapter);

        /* at least one bit of bits 0-2 must be set*/
	    if ((temp & 0xFFFF) != PCI_VENDOR_ID_NX) {
		// Vendor ID is itself wrong. Report definite error
		    ret = -CR_ERROR;
	    }
        return (ret);
}

/*
 * unm_nic_ioctl ()    We provide the tcl/phanmon support through these
 * ioctls.
 */
static int
unm_nic_ioctl(struct net_device *netdev, struct ifreq *ifr, int cmd)
{
        int                   err     = 0;
        unsigned long         nr_bytes= 0;
        struct unm_adapter_s *adapter = netdev_priv(netdev);
        char                  dev_name[UNM_NIC_NAME_LEN];

        DPRINTK(1, INFO, "doing ioctl for %s\n", netdev->name);
        switch (cmd) {
        case UNM_NIC_CMD:
                err = unm_nic_do_ioctl(adapter, (void *) ifr->ifr_data);
                break;

        case UNM_NIC_NAME:
                DPRINTK(1, INFO, "ioctl cmd for UNM\n");
                if (ifr->ifr_data) {
                        sprintf(dev_name, "%s-%d", UNM_NIC_NAME_RSP,
                                adapter->portnum);
                        nr_bytes = copy_to_user((char *)ifr->ifr_data, dev_name,
                                UNM_NIC_NAME_LEN);
                        if(nr_bytes)
                            err = -EIO;
                }
                break;

        case UNM_NIC_SEND_TEST:
                err = unm_send_test(netdev, (void *)ifr->ifr_data);
                break;

#if defined(UNM_IP_FILTER)
        case UNM_NIC_IP_FILTER:
                if(adapter->is_up == ADAPTER_UP_MAGIC)
                   err = unm_ip_filter(adapter, (void *)ifr->ifr_data);
                else {
                   printk(KERN_ERR"%s: Adapter resources are not initialized\n",
                                   unm_nic_driver_name);
                   err = -ENOSYS;
                }
                break;
#endif /* UNM_IP_FILTER */

        case UNM_NIC_IRQ_TEST:
                err = unm_irq_test(adapter);
                break;

        case UNM_NIC_ILB_TEST:
                if(adapter->is_up == ADAPTER_UP_MAGIC)
                    err = unm_loopback_test(netdev, 1, (void *)ifr->ifr_data);
                else {
                    printk(KERN_ERR"%s: Adapter resources not initialized\n",
                        unm_nic_driver_name);
                    err = -ENOSYS;
                }
                break;

        case UNM_NIC_ELB_TEST:
                if(adapter->is_up == ADAPTER_UP_MAGIC)
                    err = unm_loopback_test(netdev, 0, (void *)ifr->ifr_data);
                else {
                    printk(KERN_ERR"%s: Adapter resources not initialized\n",
                                  unm_nic_driver_name);
                    err = -ENOSYS;
                }
                break;

        case UNM_NIC_LINK_TEST:
                err = unm_link_test(adapter);
                break;

        case UNM_NIC_HW_TEST:
                err = unm_hw_test(adapter);
                break;

        case UNM_NIC_CIS_TEST:
                err = unm_cis_test(adapter);
                break;

        case UNM_NIC_CR_TEST:
                err = unm_cr_test(adapter);
                break;

        default:
                DPRINTK(1, INFO, "ioctl cmd %x not supported\n", cmd);
                err = -EOPNOTSUPP;
                break;
        }

        return err;
}

#ifndef __VMKERNEL_MODULE__
static int
unm_nic_suspend(struct pci_dev *pdev, PM_MESSAGE_T state)
{
        struct net_device *netdev = pci_get_drvdata(pdev);
        struct unm_adapter_s   *adapter   = netdev_priv(netdev);

        netif_device_detach(netdev);

        if(netif_running(netdev)) {
                netif_carrier_off(netdev);
                netif_stop_queue(netdev);
                adapter->state = PORT_SUSPEND;
                unm_nic_down(adapter);
        }

/*        pci_save_state(pdev, port->pci_state);        */
/*        pci_disable_device(pdev);        */

        return 0;
}

static int
unm_nic_resume(struct pci_dev *pdev)
{
	uint32_t ret;
        struct net_device *netdev = pci_get_drvdata(pdev);
        struct unm_adapter_s   *adapter = netdev_priv(netdev);
        ret = pci_enable_device(pdev);
        adapter->state = PORT_UP;
        netif_device_attach(netdev);
        return ret;
}
#endif /* __VMKERNEL_MODULE__ */

int
unm_nic_fill_statistics(struct unm_adapter_s *adapter, 
                        struct unm_statistics *unm_stats)
{
	void* addr;
	if(adapter->ahw.board_type == UNM_NIC_XGBE) {
		UNM_WRITE_LOCK(&adapter->adapter_lock);
		unm_nic_pci_change_crbwindow(adapter, 0);
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_TX_BYTE_CNT, &(unm_stats->tx_bytes));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_TX_FRAME_CNT, &(unm_stats->tx_packets));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_RX_BYTE_CNT, &(unm_stats->rx_bytes));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_RX_FRAME_CNT, &(unm_stats->rx_packets));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_AGGR_ERROR_CNT, &(unm_stats->rx_errors));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_CRC_ERROR_CNT, &(unm_stats->rx_CRC_errors));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_OVERSIZE_FRAME_ERR, &(unm_stats->rx_long_length_error));
		UNM_NIC_LOCKED_READ_REG(UNM_NIU_XGE_UNDERSIZE_FRAME_ERR, &(unm_stats->rx_short_length_error));

		/* For reading rx_MAC_error bit different procedure */
/*		UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_TEST_MUX_CTL, 0x15);
		UNM_NIC_LOCKED_READ_REG((UNM_CRB_NIU + 0xC0), &temp);
		unm_stats->rx_MAC_errors = temp & 0xff; */

		unm_nic_pci_change_crbwindow(adapter, 1);
		UNM_WRITE_UNLOCK(&adapter->adapter_lock);
	} else {
		spin_lock_bh(&adapter->tx_lock);
               unm_stats->tx_bytes = adapter->stats.txbytes;
               unm_stats->tx_packets = adapter->stats.xmitedframes +
                                       adapter->stats.xmitfinished;
               unm_stats->rx_bytes = adapter->stats.rxbytes;
               unm_stats->rx_packets = adapter->stats.no_rcv;
               unm_stats->rx_errors = adapter->stats.rcvdbadskb;
               unm_stats->tx_errors = adapter->stats.nocmddescriptor;
               unm_stats->rx_short_length_error = adapter->stats.uplcong;
               unm_stats->rx_long_length_error = adapter->stats.uphcong;
		unm_stats->rx_CRC_errors = 0;
		unm_stats->rx_MAC_errors = 0;
		spin_unlock_bh(&adapter->tx_lock);
	}
	return 0;
}

int
unm_nic_clear_statistics(struct unm_adapter_s *adapter)
{
	void* addr;
	int data = 0;

	UNM_WRITE_LOCK(&adapter->adapter_lock);
	unm_nic_pci_change_crbwindow(adapter, 0);

	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_TX_BYTE_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_TX_FRAME_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_RX_BYTE_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_RX_FRAME_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_AGGR_ERROR_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_CRC_ERROR_CNT, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_OVERSIZE_FRAME_ERR, &data);
	UNM_NIC_LOCKED_WRITE_REG(UNM_NIU_XGE_UNDERSIZE_FRAME_ERR, &data);

	unm_nic_pci_change_crbwindow(adapter, 1);
	UNM_WRITE_UNLOCK(&adapter->adapter_lock);
	unm_nic_clear_stats(adapter);
	return 0;
}
static int
unm_nic_do_ioctl(struct unm_adapter_s *adapter, void * u_data)
{
        unm_nic_ioctl_data_t data;
        unm_nic_ioctl_data_t *up_data;
        int retval = 0;
        struct unm_statistics unm_stats;

        up_data = (void*)u_data;

        DPRINTK(1, INFO, "doing ioctl for %p\n", adapter);
        if (copy_from_user(&data, up_data, sizeof(data))) {
                /* evil user tried to crash the kernel */
                DPRINTK(1, ERR, "bad copy from userland: %d\n",
                                 (int)sizeof(data));
                retval = -EFAULT;
                goto error_out;
        }

        /* Shouldn't access beyond legal limits of  "char u[64];" member */
        if (!data.ptr && (data.size >  sizeof(data.u))) {
                /* evil user tried to crash the kernel */
                DPRINTK(1, ERR, "bad size: %d\n", data.size);
                retval = -EFAULT;
                goto error_out;
        }

        switch (data.cmd) {
        case unm_nic_cmd_pci_read:
                if ((retval = unm_nic_hw_read_ioctl(adapter, data.off,
                                                &(data.u), data.size)))
                        goto error_out;
                if (copy_to_user((void*)&(up_data->u), &(data.u),
                                        data.size)) {
                        DPRINTK(1, ERR, "bad copy to userland: %d\n",
                                        (int)sizeof(data));
                        retval = -EFAULT;
                        goto error_out;
                }
                data.rv = 0;
                break;

        case unm_nic_cmd_pci_write:
                data.rv = unm_nic_hw_write_ioctl(adapter, data.off,  &(data.u),
                                            data.size);
                break;

        case unm_nic_cmd_pci_mem_read:
                DPRINTK(1, INFO, "doing %s for %p\n", "unm_nic_cmd_pci_mm_rd",
                                 adapter);
                if ((unm_nic_pci_mem_read_ioctl(adapter, data.off, &(data.u),
		    data.size) != 0) || (copy_to_user((void*)&(up_data->u),
		    &(data.u), data.size) != 0)) {
                        DPRINTK(1, ERR, "bad copy from card to userland: %d\n",
                                        (int)sizeof(data));
                        retval = -EFAULT;
                        goto error_out;
                }
                data.rv = 0;
                DPRINTK(1, INFO, "read %lx\n", (unsigned long)data.u);
                break;

        case unm_nic_cmd_pci_mem_write:
                if ((data.rv = unm_nic_pci_mem_write_ioctl(adapter, data.off, &(data.u),
                    data.size)) != 0) {
                        retval = -EFAULT;
                        goto error_out;
                }
                break;

        case unm_nic_cmd_pci_config_read:
                switch(data.size) {
                case 1:
                        data.rv = pci_read_config_byte(adapter->ahw.pdev,
                                                        data.off,
                                                        (char *)&(data.u));
                        break;
                case 2:
                        data.rv = pci_read_config_word(adapter->ahw.pdev,
                                                        data.off,
                                                        (short *)&(data.u));
                        break;
                case 4:
                        data.rv = pci_read_config_dword(adapter->ahw.pdev,
                                                        data.off,
                                                        (u32 *)&(data.u));
                        break;
                }
                if(copy_to_user((void*)&(up_data->u), &(data.u), data.size)) {
                        DPRINTK(1, ERR, "bad copy to userland: %d\n",
                                         (int)sizeof(data));
                        goto error_out;
                }
                break;

        case unm_nic_cmd_pci_config_write:
                switch(data.size) {
                case 1:
                        data.rv = pci_write_config_byte(adapter->ahw.pdev,
                                                        data.off,
                                                        *(char *)&(data.u));
                        break;
                case 2:
                        data.rv = pci_write_config_word(adapter->ahw.pdev,
                                                        data.off,
                                                        *(short *)&(data.u));
                        break;
                case 4:
                        data.rv = pci_write_config_dword(adapter->ahw.pdev,
                                                        data.off,
                                                        *(u32 *)&(data.u));
                        break;
                }
                break;

        case unm_nic_cmd_get_version:
                if(copy_to_user((void*)&(up_data->u), UNM_NIC_LINUX_VERSIONID,
                                sizeof(UNM_NIC_LINUX_VERSIONID))) {
                        DPRINTK(1, ERR, "bad copy to userland: %d\n",
                                         (int)sizeof(data));
                        goto error_out;
                }
                break;

	case unm_nic_cmd_get_stats:
		data.rv = unm_nic_fill_statistics(adapter, &unm_stats);
		if(copy_to_user((void*) (up_data->ptr), (void*)&unm_stats,
				sizeof(struct unm_statistics))) {
                        DPRINTK(1, ERR, "bad copy to userland: %d\n",
                                         (int)sizeof(data));
                        goto error_out;
                }
		up_data->rv = data.rv;
		break;

	case unm_nic_cmd_clear_stats:
		data.rv = unm_nic_clear_statistics(adapter);
		up_data->rv = data.rv;
		break;
#if 0  // wait for the unmflash changes
        case unm_nic_cmd_flash_read:
                /* do_rom_fast_read() called instead of rom_fast_read() because
                   rom_lock/rom_unlock is done by separate ioctl */
                if ((retval = do_rom_fast_read(adapter, data.off,
                                               (int *)&(data.u))))
                        goto error_out;
                if (copy_to_user((void*)&(up_data->u), &(data.u),
                                         UNM_FLASH_READ_SIZE)) {
                        DPRINTK(1, ERR, "bad copy to userland: %d\n",
                                        (int)sizeof(data));
                        retval = -EFAULT;
                        goto error_out;
                }
                data.rv = 0;
                break;

        case unm_nic_cmd_flash_write:
              /* do_rom_fast_write() called instead of rom_fast_write() because
                    rom_lock/rom_unlock is done by separate ioctl */
                data.rv = do_rom_fast_write(adapter, data.off,
                                            *(u32 *)&(data.u));
                break;

        case unm_nic_cmd_flash_se:
                 data.rv = rom_se(adapter, data.off);
                 break;
#endif

        default:
                DPRINTK(1, INFO, "bad command %d for %p\n", data.cmd, adapter);
                retval = -EOPNOTSUPP;
                goto error_out;
        }
        put_user(data.rv, &(up_data->rv));
        DPRINTK(1, INFO, "done ioctl for %p well.\n", adapter);

error_out:
        return retval;

}

static int
unm_nic_port_read_proc(char *buf, char **start, off_t offset, int count,
                        int *eof, void *data)
{
        struct net_device *netdev = (struct net_device *)data;
        int                   i, j;
        int                len = 0;
	struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;

        if (netdev == NULL) {
                len = sprintf(buf, "No Statistics available now. Device is"
                                   " NULL\n");
                *eof = 1;
                return len;
        }
        len  = sprintf(buf, "NetXen NIC port statistics\n");
        len += sprintf(buf+len, "\n");
        len += sprintf(buf+len, "Interface Name           : %s\n",
                       netdev->name);
        len += sprintf(buf+len, "Port Number              : %d\n",
                       adapter->portnum);
        len += sprintf(buf+len, "Bad SKB                  : %lld\n",
                       adapter->stats.rcvdbadskb);
        len += sprintf(buf+len, "Xmit called              : %lld\n",
                       adapter->stats.xmitcalled);
        len += sprintf(buf+len, "Xmited Frames            : %lld\n",
                       adapter->stats.xmitedframes);
        len += sprintf(buf+len, "Bad SKB length           : %lld\n",
                       adapter->stats.badskblen);
        len += sprintf(buf+len, "Cmd Desc Error           : %lld\n",
                       adapter->stats.nocmddescriptor);
        len += sprintf(buf+len, "Polled for Rcv           : %lld\n",
                       adapter->stats.polled);
        len += sprintf(buf+len, "Received Desc            : %lld\n",
                       adapter->stats.no_rcv);
        len += sprintf(buf+len, "Rcv to stack             : %lld\n",
                       adapter->stats.uphappy);
        len += sprintf(buf+len, "Stack dropped            : %lld\n",
                       adapter->stats.updropped);
        len += sprintf(buf+len, "Low congestion           : %lld\n",
                       adapter->stats.uplcong);
        len += sprintf(buf+len, "High congestion          : %lld\n",
                       adapter->stats.uphcong);
        len += sprintf(buf+len, "Medium congestion        : %lld\n",
                       adapter->stats.upmcong);
        len += sprintf(buf+len, "Rcv bad return           : %lld\n",
                       adapter->stats.updunno);
        len += sprintf(buf+len, "SKBs Freed               : %lld\n",
                       adapter->stats.skbfreed);
        len += sprintf(buf+len, "Xmit finished            : %lld\n",
                       adapter->stats.xmitfinished);
        len += sprintf(buf+len, "Tx dropped SKBs          : %lld\n",
                       adapter->stats.txdropped);
        len += sprintf(buf+len, "Tx got NULL SKBs         : %lld\n",
                       adapter->stats.txnullskb);
        len += sprintf(buf+len, "Rcv of CSUMed SKB        : %lld\n",
                      adapter->stats.csummed);
        len += sprintf(buf+len, "\n");
        len += sprintf(buf+len, "Ring Statistics\n");
        len += sprintf(buf+len, "Command Producer    : %d\n",
                        adapter->cmdProducer);
        len += sprintf(buf+len, "LastCommand Consumer: %d\n",
                       adapter->lastCmdConsumer);
        for (i = 0; i < MAX_RCV_CTX; ++i) {
                for (j = 0; j < NUM_RCV_DESC_RINGS; j++) {
                        len += sprintf(buf+len, "Rcv Ring %d\n", j);
                        len += sprintf(buf+len, "\tReceive Producer [%d]:"
                                   " %d\n", i,
                                   adapter->recv_ctx[i].rcv_desc[j].producer);
                }
                len += sprintf(buf+len, "Status(Rcv) Producer[%d]: %d\n",
                                    i, adapter->recv_ctx[i].statusRxProducer);
                len += sprintf(buf+len, "Status(Rcv) Consumer[%d]: %d\n",
                                    i, adapter->recv_ctx[i].statusRxConsumer);
        }
        *eof = 1;
        return len;
}

typedef int (*pci_func_t)(struct pci_dev *, const struct pci_device_id *);

static struct pci_driver unm_driver = {
        .name     = unm_nic_driver_name,
        .id_table = unm_pci_tbl,
        .probe    = (pci_func_t)unm_nic_probe,
        .remove   = __devexit_p(unm_nic_remove),
#ifndef __VMKERNEL_MODULE__
        .suspend  = unm_nic_suspend,
        .resume   = unm_nic_resume
#endif
};

#ifdef UNM_IDS_ENABLED
static int
unm_nic_fs_open(struct inode *in, struct file *file)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if (!try_module_get(THIS_MODULE))
                return -1;
#else
        MOD_INC_USE_COUNT;
#endif
        return 0;
}

static int
unm_nic_fs_release(struct inode *in, struct file *file)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        module_put(THIS_MODULE);
#else
        MOD_DEC_USE_COUNT;
#endif
        return 0;
}

static int
unm_nic_send_register_cmd(uint32_t idx)
{
        struct net_device *netdev;
	struct unm_adapter_s *adapter = (struct unm_adapter_s*)netdev->priv;

        /* Window = 1 */
	read_lock(&adapter->adapter_lock);
        UNM_NIC_PCI_WRITE_32(rbuf[idx].ring_buffer_size,
                                 CRB_NORMALIZE(adapter, CRB_MMAP_SIZE(idx)));
        UNM_NIC_PCI_WRITE_32(rbuf[idx].ring_buffer_phys,
                                 CRB_NORMALIZE(adapter, CRB_MMAP_ADDR(idx)));

	read_unlock(&adapter->adapter_lock);


        return 0;
}

/*
 * Let the phantom code know that there is not more dma to be done.
 * Set all the CRB registers to 0.
 */
static void
unm_nic_send_unregister_cmd(void)
{
        struct net_device *netdev;
        struct unm_adapter_s *adapter;
        int idx;

        if (unm_netlist == NULL || unm_netlist->netdev == NULL) {
                return;
        }
        netdev = unm_netlist->netdev;
        adapter = netdev->priv;

        for (idx = 0; idx < adapter->ahw.max_ports; ++idx) {
                /* Window = 1 */
		read_lock(&adapter->adapter_lock);
                UNM_NIC_PCI_WRITE_32(0, CRB_NORMALIZE(adapter, CRB_MMAP_SIZE(idx)));
                UNM_NIC_PCI_WRITE_32(0, CRB_NORMALIZE(adapter, CRB_MMAP_ADDR(idx)));
		read_unlock(&adapter->adapter_lock);
        }
}

/*
 * mmap function for the UNM devices. Note that application provides
 * the offset & the size. The offset encodes the desired interface.
 * The relationship is:
 * For kernel 2.4 & above,
 *           desired_interface = offset / PAGE_SIZE
 *
 * For other kernels,
 *           desirect_interface = offset
 *
 * If a buffer has already been allocated, it is re-used.
 */
static int
unm_nic_mmap(struct file *file, struct vm_area_struct *vma)
{
        int i;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
        unsigned long idx = vma->vm_offset >> PAGE_SHIFT;
#else
        unsigned long idx = vma->vm_pgoff;
#endif
        unsigned long size = vma->vm_end - vma->vm_start;

        if (idx < 0 || idx >= MAX_NUM_PORTS) {
                return -EINVAL;
        }

        vma->vm_flags |= VM_LOCKED;
        if (rbuf[idx].ring_buffer == 0) {
                rbuf[idx].ring_buffer = __get_free_pages(GFP_KERNEL,
                                                         get_order(size));
                if (rbuf[idx].ring_buffer == 0) {
                        vma->vm_flags &= ~VM_LOCKED;
                        return -ENOMEM;
                }
                memset((unsigned char *)rbuf[idx].ring_buffer, 0, size);
                rbuf[idx].ring_buffer_phys =
                        virt_to_phys((caddr_t) rbuf[idx].ring_buffer);
                rbuf[idx].ring_buffer_size=size;
        } else {
                if (size > rbuf[idx].ring_buffer_size) {
                        return -EINVAL;
                }
        }
        for (i = 0; i < rbuf[idx].ring_buffer_size / PAGE_SIZE; ++i) {
                SetPageReserved(virt_to_page(rbuf[idx].ring_buffer +
                                                    (i*PAGE_SIZE)));
        }
#if defined(FORCE_NO_PATCH)
#undef  VM_START
#define VM_START(p)        (p)->vm_start
#elif defined(VM_PATCH) || defined(RED_HAT_LINUX_KERNEL) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define VM_START(p)        (p), (p)->vm_start
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#define VM_START(p)        (p), (p)->vm_start
#else
#define VM_START(p)        (p)->vm_start
#endif /* RED_HAT_LINUX_KERNEL || LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) */
        if (io_remap_page_range(vma, vma->vm_start,
                virt_to_phys((void *)rbuf[idx].ring_buffer) >> PAGE_SHIFT,
                rbuf[idx].ring_buffer_size, vma->vm_page_prot)) {
                vma->vm_flags &= ~VM_LOCKED;
                unm_nic_unregister_iface(idx);
                return -ENXIO;
        }

        if (unm_nic_send_register_cmd(idx) != 0) {
                vma->vm_flags &= ~VM_LOCKED;
                unm_nic_unregister_iface(idx);
                return -ENXIO;
        }
        return 0;
}


static struct file_operations unm_fops =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        owner:    THIS_MODULE,
#endif
        mmap:     unm_nic_mmap,
        open:     unm_nic_fs_open,
        release:  unm_nic_fs_release,
};

int unm_nic_fs_major = 252;
#endif

/* Driver Registration on UNM card    */

static int __init
unm_init_module(void)
{
        int i;
	int err;

#ifdef __VMKERNEL_MODULE__
        if (!vmk_set_module_version("%s", unm_nic_driver_string)) {
                return -ENODEV;
        }
#endif /* __VMKERNEL_MODULE__ */

        unm_proc_dir_entry = proc_mkdir(unm_nic_driver_name, proc_net);
        if (!unm_proc_dir_entry) {
                printk(KERN_WARNING"%s: Unable to create /proc/net/%s",
                            unm_nic_driver_name, unm_nic_driver_name);
                return -ENOMEM;
        }
        unm_proc_dir_entry->owner = THIS_MODULE;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if ((unm_workq = create_singlethread_workqueue("unm")) == NULL) {
		remove_proc_entry(unm_proc_dir_entry->name, proc_net);
		return(-ENOMEM);
	}
#endif

#ifdef UNM_IDS_ENABLED
       /*
        * Register a fops interface as well.
        */
        if (register_chrdev(unm_nic_fs_major, UNM_MAP_DEV, &unm_fops) < 0) {
                printk("%s: Failed to register %s\n", unm_nic_driver_name,
                        UNM_MAP_DEV);
                return -EIO;
        }
#endif

#ifdef BOUNCE_BUFFERS
        spin_lock_init(&tx_bounce_lock);
        spin_lock_init(&rx_bounce_lock);
        if (unm_alloc_bounce_buffers()) {
		return -ENOMEM;
	}
#endif
        cmd_desc_cache = kmem_cache_create("cmd_desc_cache",
                sizeof(cmdDescType0_t) * ((MAX_BUFFERS_PER_CMD/4)+3),
                ARCH_KMALLOC_MINALIGN, ARCH_KMALLOC_FLAGS, NULL, NULL);
        cmd_buff_cache = kmem_cache_create("cmd_buff_cache",
                sizeof(struct unm_cmd_buffer), ARCH_KMALLOC_MINALIGN,
                ARCH_KMALLOC_FLAGS, NULL, NULL);
        for (i=0; i<NR_CPUS; i++) {
                cmd_desc_array[i] =
                    (cmdDescType0_t *) kmem_cache_alloc (cmd_desc_cache,
                                                            GFP_ATOMIC);
                cmd_buff_array[i] =
                    (struct unm_cmd_buffer *) kmem_cache_alloc (cmd_buff_cache,
                                                                GFP_ATOMIC);
        }

        err = pci_module_init(&unm_driver);

        if (err != 0 || unm_nic_probe_err != 0) {
                for (i=0; i<NR_CPUS; i++) {
                        if (cmd_desc_array[i])
                                kmem_cache_free(cmd_desc_cache,
                                                    cmd_desc_array[i]);
                        if (cmd_buff_array[i])
                                kmem_cache_free(cmd_buff_cache,
                                                    cmd_buff_array[i]);
                }
                kmem_cache_destroy(cmd_desc_cache);
                kmem_cache_destroy(cmd_buff_cache);
		cmd_desc_cache = NULL;
		cmd_buff_cache = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		destroy_workqueue(unm_workq);
#endif
		if (unm_proc_dir_entry != NULL) {
			remove_proc_entry(unm_proc_dir_entry->name, proc_net);
			unm_proc_dir_entry = NULL;
		}
#ifdef BOUNCE_BUFFERS
		unm_free_bounce_buffers();
#endif
		return err;
        }

        return 0;
}

module_init(unm_init_module);

static void __exit
unm_exit_module(void)
{
        int i;

        if (tx_user_packet_data != NULL) {
                kfree(tx_user_packet_data);
                tx_user_packet_data = NULL;
                tx_user_packet_length = 0;
        }
#ifdef UNM_IDS_ENABLED
        unm_nic_send_unregister_cmd();
#endif
        /*
         * Wait for some time to allow the dma to drain, if any.
         */
        mdelay(5);
#ifdef UNM_IDS_ENABLED
        unm_nic_unregister_all();
        unregister_chrdev(unm_nic_fs_major, UNM_MAP_DEV);
#endif
        pci_unregister_driver(&unm_driver);
	if (unm_nic_probe_err == 0) {
		if (unm_proc_dir_entry != NULL) {
			remove_proc_entry(unm_proc_dir_entry->name, proc_net);
			unm_proc_dir_entry = NULL;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		destroy_workqueue(unm_workq);
#endif
		for (i=0; i<NR_CPUS; i++) {
			if (cmd_desc_array[i])
				kmem_cache_free(cmd_desc_cache, cmd_desc_array[i]);
			if (cmd_buff_array[i])
				kmem_cache_free(cmd_buff_cache, cmd_buff_array[i]);
		}
		kmem_cache_destroy(cmd_desc_cache);
		kmem_cache_destroy(cmd_buff_cache);
#ifdef BOUNCE_BUFFERS
		unm_free_bounce_buffers();
#endif
        }
}

void nx_nic_get_tnic_version_number(struct net_device *netdev,
                        nic_version_t *version)
{
        struct unm_adapter_s   *adapter = (struct unm_adapter_s*)netdev->priv;

	version->major = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_TCP_FW_VERSION_MAJOR_ADDR));
	version->minor = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_TCP_FW_VERSION_MINOR_ADDR));
	version->sub = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_TCP_FW_VERSION_SUB_ADDR));
}

void nx_nic_get_nic_version_number(struct net_device *netdev,
                        nic_version_t *version)
{
        struct unm_adapter_s   *adapter = (struct unm_adapter_s*)netdev->priv;

	version->major = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_FW_VERSION_MAJOR));
	version->minor = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_FW_VERSION_MINOR));
	version->sub = UNM_NIC_PCI_READ_32(CRB_NORMALIZE(adapter,
			UNM_FW_VERSION_SUB));
}


#ifndef __VMKERNEL_MODULE__
EXPORT_SYMBOL(nx_nic_is_netxen_device);
EXPORT_SYMBOL(nx_nic_register_tnic);
EXPORT_SYMBOL(nx_nic_unregister_tnic);
EXPORT_SYMBOL(nx_nic_get_tnic_adapter);
EXPORT_SYMBOL(nx_nic_get_device_port);
EXPORT_SYMBOL(nx_nic_get_device_ring_ctx);
EXPORT_SYMBOL(nx_nic_send_tnic_msg);
EXPORT_SYMBOL(nx_nic_get_pcidev);
EXPORT_SYMBOL(nx_nic_cmp_adapter_id);
EXPORT_SYMBOL(nx_nic_get_tnic_version_number);
EXPORT_SYMBOL(nx_nic_get_nic_version_number);
#endif

module_exit(unm_exit_module);
